clear all, 
close all, 
clc
set(0,'DefaultFigureWindowStyle','normal')

run('../../Bibliotecas/parametrosFiguras.m')

%%
s = tf('s')
Ts = 0.05

G = 4/(s+2)^2*exp(-10*Ts*s);


%%

Gz = c2d(G,Ts,'zoh');
num = Gz.num{1};
den = Gz.den{1};
na = size(den,2)-1;
nb = size(num,2)-2;
dd = Gz.inputdelay;


M = ceil(10/Ts)
M=80;

Gcoef = step(G,Ts:Ts:10);
Gqcoef = step(G,0:Ts:10);

%% parâmetros de ajuste do DMC

N1 = 11;
N2 = 40;
Nu = 15;

delta = 1;
lambda = 1;


%% montando as matrizes do DMC

G = zeros(N2,Nu);
G(:,1) = Gcoef(1:N2,1);


for i=2:Nu
    G(i:end,i) = G(1:end-(i-1),1);    
end

G = G(N1:end,:);

Qy = delta*eye(N2-N1+1);
Qu = lambda*eye(Nu);

Kdmc = inv(G'*Qy*G+Qu)*G'*Qy

Kdmc1 = Kdmc(1,:);

Ylivre = ones(M,1)*0; %%% 0 é o valor inicial da saída do sistema em regime permanente


%% inicialização vetores
duAnt = 0;
uAnt = 0;

nin = M+1;
nit = round(10/Ts) + nin;

entradas = 0*ones(nit,1);
du = zeros(nit,1);

saidas = 0*ones(nit,1);

perts = zeros(nit,1);
perts(nin+round(4/Ts):end) = 10;

refs = 0*ones(nit,1);
refs(nin+round(0.5/Ts):nit) = 50;

erro = zeros(nit,1);

%% simulação sem blocagem

for i = nin:nit-N2-1
    %% modelo processo, não mexer
    saidas(i) = -den(2:end)*saidas(i-1:-1:i-na) + num*(entradas(i-dd:-1:i-nb-dd-1) + perts(i-dd:-1:i-dd-nb-1));
    
    erro(i) = refs(i)-saidas(i);
    
    %% Controlador
    
    Ylivre = Ylivre + Gcoef(1:M,1)*duAnt;
    
    eta = saidas(i)-Ylivre(1); %%% erro de predição
    
    Ylivre(1:end-1) = Ylivre(2:end); %%% deslocamento do vetor
    
    %%% resposta livre
    f = Ylivre(N1:N2,1)+eta;
    
    %%% referências
    R = ones(N2-N1+1,1)*refs(i);
    
    %%% calculo do incremento de controle ótimo    
    duAtual = Kdmc1*(R-f);
    
    %%% calculo da ação de controle real
    uAtual = duAtual + uAnt;
    
    %%% aplicar no sistema
    entradas(i) = uAtual;
    
    %%% atualização
    
    uAnt = uAtual;
    duAnt = duAtual;
    
    
end

saidasSemBlocagem = saidas;
entradasSemBlocagem = entradas;
duSemBlocagem = du;

%% DMC com blocagem uniforme
nbc = 4;
a = (Nu-1)/(nbc-1)
b = -a+1

Nubloc = a*(1:nbc)+b
Nubloc = round(Nubloc)

%% montando as matrizes do DMC com blocagem uniforme

G = zeros(N2,Nu);
G(:,1) = Gcoef(1:N2,1);


for i=2:Nu
    G(i:end,i) = G(1:end-(i-1),1);    
end

G = G(N1:end,Nubloc);

Qy = delta*eye(N2-N1+1);
Qu = lambda*eye(size(Nubloc,2));

Kdmc = inv(G'*Qy*G+Qu)*G'*Qy

Kdmc1 = Kdmc(1,:);

Ylivre = ones(M,1)*0; %%% 0 é o valor inicial da saída do sistema em regime permanente


%% inicialização vetores
duAnt = 0;
uAnt = 0;

entradas = 0*ones(nit,1);
du = zeros(nit,1);
saidas = 0*ones(nit,1);
erro = zeros(nit,1);

%% simulação com blocagem uniforme

for i = nin:nit-N2-1
    %% modelo processo, não mexer
    saidas(i) = -den(2:end)*saidas(i-1:-1:i-na) + num*(entradas(i-dd:-1:i-nb-dd-1) + perts(i-dd:-1:i-dd-nb-1));
    
    erro(i) = refs(i)-saidas(i);
    
    %% Controlador
    
    Ylivre = Ylivre + Gcoef(1:M,1)*duAnt;
    
    eta = saidas(i)-Ylivre(1); %%% erro de predição
    
    Ylivre(1:end-1) = Ylivre(2:end); %%% deslocamento do vetor
    
    %%% resposta livre
    f = Ylivre(N1:N2,1)+eta;
    
    %%% referências
    R = ones(N2-N1+1,1)*refs(i);
    
    %%% calculo do incremento de controle ótimo    
    duAtual = Kdmc1*(R-f);
    
    %%% calculo da ação de controle real
    uAtual = duAtual + uAnt;
    
    %%% aplicar no sistema
    entradas(i) = uAtual;
    
    %%% atualização
    
    uAnt = uAtual;
    duAnt = duAtual;
    
    
end

saidasBlocUniforme = saidas;
entradasBlocUniforme = entradas;
duBlocUniforme = du;


%% DMC com blocagem exponencial
nbc = 4;
b = Nu^(1/(nbc-1))
a = 1/b

Nubloc = a*b.^(1:nbc)
Nubloc = round(Nubloc)

%% montando as matrizes do DMC com blocagem uniforme

G = zeros(N2,Nu);
G(:,1) = Gcoef(1:N2,1);


for i=2:Nu
    G(i:end,i) = G(1:end-(i-1),1);    
end

G = G(N1:end,Nubloc);

Qy = delta*eye(N2-N1+1);
Qu = lambda*eye(size(Nubloc,2));

Kdmc = inv(G'*Qy*G+Qu)*G'*Qy

Kdmc1 = Kdmc(1,:);

Ylivre = ones(M,1)*0; %%% 0 é o valor inicial da saída do sistema em regime permanente


%% inicialização vetores
duAnt = 0;
uAnt = 0;

entradas = 0*ones(nit,1);
du = zeros(nit,1);
saidas = 0*ones(nit,1);
erro = zeros(nit,1);

%% simulação com blocagem exponencial

for i = nin:nit-N2-1
    %% modelo processo, não mexer
    saidas(i) = -den(2:end)*saidas(i-1:-1:i-na) + num*(entradas(i-dd:-1:i-nb-dd-1) + perts(i-dd:-1:i-dd-nb-1));
    
    erro(i) = refs(i)-saidas(i);
    
    %% Controlador
    
    Ylivre = Ylivre + Gcoef(1:M,1)*duAnt;
    
    eta = saidas(i)-Ylivre(1); %%% erro de predição
    
    Ylivre(1:end-1) = Ylivre(2:end); %%% deslocamento do vetor
    
    %%% resposta livre
    f = Ylivre(N1:N2,1)+eta;
    
    %%% referências
    R = ones(N2-N1+1,1)*refs(i);
    
    %%% calculo do incremento de controle ótimo    
    duAtual = Kdmc1*(R-f);
    
    %%% calculo da ação de controle real
    uAtual = duAtual + uAnt;
    
    %%% aplicar no sistema
    entradas(i) = uAtual;
    
    %%% atualização
    
    uAnt = uAtual;
    duAnt = duAtual;
    
    
end

saidasBlocExp = saidas;
entradasBlocExp = entradas;
duBlocExp = du;

%% plots
t = ((nin:nit-N2-1)-nin)*Ts;

cores = gray(5);
cores = cores(1:end-1,:);


hf=figure
h=subplot(2,1,1)
plot(t,saidasSemBlocagem(nin:nit-N2-1),'LineWidth',tamlinha,'Color',cores(1,:))
hold on
plot(t,saidasBlocUniforme(nin:nit-N2-1),'-.','LineWidth',tamlinha,'Color',cores(2,:))
plot(t,saidasBlocExp(nin:nit-N2-1),':','LineWidth',tamlinha,'Color',cores(3,:))
plot(t,refs(nin:nit-N2-1),'--','LineWidth',tamlinha,'Color',cores(4,:))

hl = legend('S/ Bloc.','Bloc. Unif.','Bloc. Exp.','Referência','Location','SouthEast')
ylabel('Temperatura (^oC)', 'FontSize', tamletra);
% axis([0 60 0 350]);
grid on
set(h, 'FontSize', tamletra);

h=subplot(2,1,2)
plot(t,entradasSemBlocagem(nin:nit-N2-1),'LineWidth',tamlinha,'Color',cores(1,:))
hold on
plot(t,entradasBlocUniforme(nin:nit-N2-1),'-.','LineWidth',tamlinha,'Color',cores(2,:))
plot(t,entradasBlocExp(nin:nit-N2-1),':','LineWidth',tamlinha,'Color',cores(1,:))
ylabel('Potência (kW)', 'FontSize', tamletra);
xlabel('Tempo (minutos)', 'FontSize', tamletra);
% axis([0 60 0 450]);
grid on
set(h, 'FontSize', tamletra);


hf.Position = tamfigura;
hl.Position = [0.7077 0.5411 0.2054 0.1242];

print('blocagem_comparacao','-depsc')
