clear all, 
close all, 
clc
% set(0,'DefaultFigureWindowStyle','normal')

run('../Bibliotecas/parametrosFiguras.m')
%%
s = tf('s')
Ts = 0.1

G = 4/(s+1)^2;

%%

Gz = c2d(G,Ts,'zoh');
num = Gz.num{1};
den = Gz.den{1};
na = size(den,2)-1;
nb = size(num,2)-2;
dd = Gz.inputdelay;


M = ceil(10/Ts)
M=100;

Gcoef = step(G,Ts:Ts:10);
Gqcoef = step(G,0:Ts:10);

%% parâmetros de ajuste

N1 = 1;
N2 = 30;
N = N2-N1+1;
Nu = 5;

pinf = 10;
pint = .01;
psup = 10;

lambda = 1;
delta = 1;

%% montando as matrizes do DMC

G = zeros(N2,Nu);
G(:,1) = Gcoef(1:N2,1);


for i=2:Nu
    G(i:end,i) = G(1:end-(i-1),1);    
end

G = G(N1:end,:);

Qu = lambda*eye(Nu);

Ylivre = ones(M,1)*0; %%% 0 é o valor inicial da saída do sistema em regime permanente


%% inicialização vetores
nin = M+1;
nit = round(25/Ts) + nin;

entradas = zeros(nit,1);
du = zeros(nit,1);

saidas = 0*ones(nit,1);

perts = zeros(nit,1);
perts(nin+round(2/Ts):end) = -3;
perts(nin+round(15/Ts):end) = 0;

Ta=20; %T ambiente
refs = 0*ones(nit,1);
refs(nin+round(2/Ts):nit) = 0;
refs(nin+round(8/Ts):nit) = 0;
% refs(nin+850:end) = 4;

bandasup = 2*ones(nit,1);
bandasup(nin+round(12/Ts):end) = 1;
bandainf = -3*ones(nit,1);
bandainf(nin+round(12/Ts):end) = -1;


erro = zeros(nit,1);

%% Caso 1
for i = nin:nit
    %% modelo processo, não mexer
    saidas(i) = -den(2:end)*saidas(i-1:-1:i-na) + num*(entradas(i-dd:-1:i-nb-dd-1) + perts(i-dd:-1:i-dd-nb-1));
    
    erro(i) = refs(i)-saidas(i);
    
    %% Controlador
    Ylivre = Ylivre + Gcoef(1:M,1)*du(i-1);
    
    eta = saidas(i)-Ylivre(1); %%% erro de predição
    
    Ylivre(1:end-1) = Ylivre(2:end); %%% deslocamento do vetor
    
    %%% resposta livre
    f = Ylivre(N1:N2,1)+eta;
    
    %%% cálculo das ponderações
    Qy = zeros(N);
    for j=1:N
        if(f(j)>=bandasup(i))
            Qy(j,j) = psup;
        elseif(f(j)<=bandainf(i))
            Qy(j,j) = pinf;
        else
            Qy(j,j) = pint;
        end
    end
    
    %%% referências
    R = ones(N2-N1+1,1)*refs(i);
    %R = refs(i+N1:i+N2); %%% uso de referencias futuras
    
    %%% calculo do incremento de controle ótimo    
    X = inv(G'*Qy*G+Qu)*G'*Qy*(R-f);
    du(i) = X(1);    
    
    %%% aplicar no sistema
    entradas(i) = entradas(i-1)+du(i);    
    
end

saidasDMC = saidas;
entradasDMC = entradas;
duDMC = du;

figure
subplot(2,1,1)
plot(saidas)
hold on
plot(refs)
subplot(2,1,2)
plot(entradas)


%% Caso 2
Qy = delta*eye(N);

Kdmc = inv(G'*Qy*G+Qu)*G'*Qy;
Kdmc1 = Kdmc(1,:);


entradas = zeros(nit,1);
du = zeros(nit,1);
saidas = 0*ones(nit,1);
Ylivre = zeros(M,1); %%% 0 é o valor inicial da saída do sistema em regime permanente

for i = nin:nit
    %% modelo processo, não mexer
    saidas(i) = -den(2:end)*saidas(i-1:-1:i-na) + num*(entradas(i-dd:-1:i-nb-dd-1) + perts(i-dd:-1:i-dd-nb-1));
    
    %% Controlador
    Ylivre = Ylivre + Gcoef(1:M,1)*du(i-1);
    
    eta = saidas(i)-Ylivre(1); %%% erro de predição
    
    Ylivre(1:end-1) = Ylivre(2:end); %%% deslocamento do vetor
    
    %%% resposta livre
    f = Ylivre(N1:N2,1)+eta;
    
    %%% referências
    R = ones(N,1)*refs(i);

    du(i) = Kdmc1*(R-f);
    
    %%% calculo da ação de controle real
    entradas(i) = entradas(i-1)+du(i);
        
    
end

saidasDMC2 = saidas;
entradasDMC2 = entradas;
duDMC2 = du;

figure
subplot(2,1,1)
plot(saidas)
hold on
plot(refs)
subplot(2,1,2)
plot(entradas)


%% plots
ind = nin:nit;
t = (ind-nin)*Ts;

cores = gray(5);
cores = cores(1:end-1,:);

hf = figure
h=subplot(2,1,1)
plot(t,saidasDMC(ind)+52,'LineWidth',tamlinha,'Color',cores(1,:))
hold on
plot(t,saidasDMC2(ind)+52,'-.','LineWidth',tamlinha,'Color',cores(2,:))

plot(t,refs(ind)+52,':','LineWidth',tamlinha,'Color',cores(4,:))
plot(t,bandasup(ind)+52,'--','LineWidth',tamlinha,'Color',cores(4,:))
plot(t,bandainf(ind)+52,'--','LineWidth',tamlinha,'Color',cores(4,:))

ylabel('Nível (%)', 'FontSize', tamletra);
% ylim([50 60]);
hl = legend('Caso 1','Caso 2','Referência','Limites')
% hl.Position = [0.6685 0.6029 0.2375 0.1107];

grid on


set(h, 'FontSize', tamletra);

h=subplot(2,1,2)
plot(t,entradasDMC(ind)+30,'LineWidth',tamlinha,'Color',cores(1,:))
hold on
plot(t,entradasDMC2(ind)+30,'-.','LineWidth',tamlinha,'Color',cores(2,:))

ylabel('Manipulada (%)', 'FontSize', tamletra);
xlabel('Tempo (minutos)', 'FontSize', tamletra);
% ylim([20 50]);
set(h, 'FontSize', tamletra);

grid on

% h=subplot(3,1,3)
% plot(t,erro(1:nit-N2-1),'LineWidth',tamlinha)
% title('Erro', 'FontSize', tamletra);
% xlabel('tempo', 'FontSize', tamletra);
% set(h, 'FontSize', tamletra);

hf.Position = tamfigura;
hl.Position = [0.7077 0.5411 0.2054 0.1242];


print('ex_pond_variavel','-depsc')
% print('ex_rest_monotonica_pb','-deps')



