clear all, 
close all, 
clc
% set(0,'DefaultFigureWindowStyle','normal')

run('../Bibliotecas/parametrosFiguras.m')
%%
s = tf('s')
Ts = 0.1

G = 2*(-0.75*s+1)/(1.5*s+1)^2;

ksi = 0.5;
wn = 3/ksi/4.8*1.5;
G = 2*(-0.75*s+1)*wn^2/(s^2+2*ksi*wn*s+wn^2);

% figure
% step(G)

%%

Gz = c2d(G,Ts,'zoh');
num = Gz.num{1};
den = Gz.den{1};
na = size(den,2)-1;
nb = size(num,2)-2;
dd = Gz.inputdelay;


M = ceil(10/Ts)
M=150;

Gcoef = step(G,Ts:Ts:2*M*Ts);

%% parâmetros de ajuste

N1 = 1;
N2 = 20;
Nu = 5;

delta = 1;
lambda = .1;
lambdaM = 1000;


%% montando as matrizes do DMC

G = zeros(N2,Nu);
G(:,1) = Gcoef(1:N2,1);


for i=2:Nu
    G(i:end,i) = G(1:end-(i-1),1);    
end

G = G(N1:end,:);

Qy = delta*eye(N2-N1+1);
Qu = lambda*eye(Nu);

Kdmc = inv(G'*Qy*G+Qu)*G'*Qy

Kdmc1 = Kdmc(1,:);

Ylivre = zeros(M,1); %%% 0 é o valor inicial da saída do sistema em regime permanente

%%% ajuste das matrizes para o caso com restrição.
Hot = 2*(G'*Qy*G+Qu);
bot = -2*G'*Qy;

Hot = blkdiag(Hot, 2*lambdaM);

LB = [-inf*ones(Nu,1);0];
UB = [inf*ones(Nu,1);inf];


%% inicialização vetores
duAnt = 0;
uAnt = 0;

nin = M+1;
nit = round(25/Ts) + nin;

entradas = zeros(nit,1);
du = zeros(nit,1);

saidas = 0*ones(nit,1);

perts = zeros(nit,1);
perts(nin+round(15/Ts):end) = -1;

Ta=20; %T ambiente
refs = 0*ones(nit,1);
refs(nin+round(2/Ts):nit) = 5;
refs(nin+round(8/Ts):nit) = 2;
% refs(nin+850:end) = 4;


erro = zeros(nit,1);

%% simulação sem filtro de referência
for i = nin:nit
    %% modelo processo, não mexer
    saidas(i) = -den(2:end)*saidas(i-1:-1:i-na) + num*(entradas(i-dd:-1:i-nb-dd-1) + perts(i-dd:-1:i-dd-nb-1));
    
    erro(i) = refs(i)-saidas(i);
    
    %% Controlador
    Ylivre = Ylivre + Gcoef(1:M,1)*duAnt;
    
    eta = saidas(i)-Ylivre(1); %%% erro de predição
    
    Ylivre(1:end-1) = Ylivre(2:end); %%% deslocamento do vetor
    
    %%% resposta livre
    f = Ylivre(N1:N2,1)+eta;
    
    %%% referências
    R = ones(N2-N1+1,1)*refs(i);
    %R = refs(i+N1:i+N2); %%% uso de referencias futuras
    
    %%% calculo do incremento de controle ótimo    
    duAtual = Kdmc1*(R-f);
    
    %%% calculo da ação de controle real
    uAtual = duAtual + uAnt;
    
    %%% aplicar no sistema
    entradas(i) = uAtual;
    
    %%% atualização
    
    uAnt = uAtual;
    duAnt = duAtual;
    
    
end

saidasDMC = saidas;
entradasDMC = entradas;
duDMC = du;

figure
subplot(2,1,1)
plot(saidas)
hold on
plot(refs)
subplot(2,1,2)
plot(entradas)


%% Caso DMC com restrições
entradas = zeros(nit,1);
du = zeros(nit,1);
saidas = 0*ones(nit,1);
Ylivre = zeros(M,1); %%% 0 é o valor inicial da saída do sistema em regime permanente

%% simulação sem filtro de referência
for i = nin:nit
    %% modelo processo, não mexer
    saidas(i) = -den(2:end)*saidas(i-1:-1:i-na) + num*(entradas(i-dd:-1:i-nb-dd-1) + perts(i-dd:-1:i-dd-nb-1));
    
    %% Controlador
    Ylivre = Ylivre + Gcoef(1:M,1)*du(i-1);
    
    eta = saidas(i)-Ylivre(1); %%% erro de predição
    
    Ylivre(1:end-1) = Ylivre(2:end); %%% deslocamento do vetor
    
    %%% resposta livre
    f = Ylivre(N1:N2,1)+eta;
    
    %%% referências
    R = ones(N2-N1+1,1)*refs(i);
    %R = refs(i+N1:i+N2); %%% uso de referencias futuras
    
    %%% restrições
    if(refs(i)==saidas(i))
        X = quadprog(Hot,[bot*(R-f); 0]);
    elseif(refs(i)>saidas(i))
        Arest = [-G -ones(N2-N1+1,1)];
        X = quadprog(Hot,[bot*(R-f); 0], Arest,-saidas(i)+f,[],[],LB,UB);
    else
        Arest = [G -ones(N2-N1+1,1)];
        X = quadprog(Hot,[bot*(R-f); 0],  Arest,saidas(i)-f,[],[],LB,UB);
    end
    %%% calculo do incremento de controle ótimo    

    du(i) = X(1);
    xi(i) = X(end);
    
    %%% calculo da ação de controle real
    entradas(i) = entradas(i-1)+du(i);
        
    
end

saidasDMCrest = saidas;
entradasDMCrest = entradas;
duDMCrest = du;

figure
subplot(2,1,1)
plot(saidas)
hold on
plot(refs)
subplot(2,1,2)
plot(entradas)

%% Caso com restrição monotônica
%%% ajuste das matrizes para o caso com restrição.
G1 = [zeros(1,Nu);
      G(1:end-1,:)];
Arest0 = G1-G;

Hot = 2*(G'*Qy*G+Qu);
bot = -2*G'*Qy;

Hot = blkdiag(Hot, 2*lambdaM);
LB = [-inf*ones(Nu,1);0];
UB = [inf*ones(Nu,1);inf];


%% simulação sem filtro de referência
entradas = zeros(nit,1);
du = zeros(nit,1);
saidas = 0*ones(nit,1);
Ylivre = zeros(M,1); %%% 0 é o valor inicial da saída do sistema em regime permanente

for i = nin:nit
    %% modelo processo, não mexer
    saidas(i) = -den(2:end)*saidas(i-1:-1:i-na) + num*(entradas(i-dd:-1:i-nb-dd-1) + perts(i-dd:-1:i-dd-nb-1));
    
    %% Controlador
    Ylivre = Ylivre + Gcoef(1:M,1)*du(i-1);
    
    eta = saidas(i)-Ylivre(1); %%% erro de predição
    
    Ylivre(1:end-1) = Ylivre(2:end); %%% deslocamento do vetor
    
    %%% resposta livre
    f = Ylivre(N1:N2,1)+eta;
    
    %%% referências
    R = ones(N2-N1+1,1)*refs(i);
    %R = refs(i+N1:i+N2); %%% uso de referencias futuras
    
    %%% restrições
    f1 = [saidas(i);f(1:end-1)];

    %%% calculo do incremento de controle ótimo
    if(refs(i)>=saidas(i))
        Arest = [Arest0, -ones(size(Arest0,1),1)];
        X = quadprog(Hot,[bot*(R-f);0], Arest,f-f1,[],[],LB,UB);
    else
        Arest = [-Arest0, -ones(size(Arest0,1),1)];
        X = quadprog(Hot,[bot*(R-f);0],Arest,f1-f,[],[],LB,UB);
    end
    
    du(i) = X(1);
    
    %%% calculo da ação de controle real
    entradas(i) = entradas(i-1)+du(i);
        
    
end

saidasDMCmon = saidas;
entradasDMCmon = entradas;
duDMCmon = du;

figure
subplot(2,1,1)
plot(saidas)
hold on
plot(refs)
subplot(2,1,2)
plot(entradas)

%% plots
ind = nin:nit;
t = (ind-nin)*Ts;

cores = gray(5);
cores = cores(1:end-1,:);


hf = figure
h=subplot(2,1,1)
plot(t,saidasDMC(ind),'LineWidth',tamlinha,'Color',cores(1,:))
hold on
plot(t,saidasDMCrest(ind),'--','LineWidth',tamlinha,'Color',cores(2,:))
plot(t,saidasDMCmon(ind),'-.','LineWidth',tamlinha,'Color',cores(3,:))

plot(t,refs(ind),':','LineWidth',tamlinha,'Color',cores(4,:))


ylabel('Controlada', 'FontSize', tamletra);
% ylim([50 60]);
hl = legend('Caso 1','Caso 2','Caso 3','Referência')
% hl.Position = [0.6685 0.6029 0.2375 0.1107];

grid on


set(h, 'FontSize', tamletra);

h=subplot(2,1,2)
plot(t,entradasDMC(ind),'LineWidth',tamlinha,'Color',cores(1,:))
hold on
plot(t,entradasDMCrest(ind),'--','LineWidth',tamlinha,'Color',cores(2,:))
plot(t,entradasDMCmon(ind),'-.','LineWidth',tamlinha,'Color',cores(3,:))

ylabel('Manipulada', 'FontSize', tamletra);
xlabel('Tempo (segundos)', 'FontSize', tamletra);
% ylim([20 50]);
set(h, 'FontSize', tamletra);

grid on

% h=subplot(3,1,3)
% plot(t,erro(1:nit-N2-1),'LineWidth',tamlinha)
% title('Erro', 'FontSize', tamletra);
% xlabel('tempo', 'FontSize', tamletra);
% set(h, 'FontSize', tamletra);

hf.Position = tamfigura;
hl.Position = [0.7077 0.4621 0.2054 0.2371];


print('ex_rest_inversa_comp','-depsc')



