clear all
% close all
clc

% MIMO DMC (pode ser usado com SISO, para isso, definir 1 ent e 1 saida)


m = 1; % número de entradas da planta - manipuladas
n = 1; %número de saidas da planta - controladas
mq = 2; % nmero de pertubaes

s = tf('s');
% Definir os modelos que relacionam {saida x,entrada y};
G(1,1) = -5/(50*s+1)*exp(-5*s);
%G(1,2) = -18.9/(21*s+1)*exp(-3*s);
%G(2,1) = 6.6/(10.9*s+1)*exp(-7*s);
%G(2,2) = -19.4/(14.4*s+1)*exp(-3*s);

% Definir os modelos que relacionam {saida x,perturbação z};
Gq(1,1) = 10/(50*s+1);
Gq(1,2) = 1/(50*s+1);


Ts = 2.0; %período de amostragem, OBS: manter atrasos múltiplos inteiros do Ts

%%% discretização do processo
Gz = c2d(G,Ts,'zoh');
Gzq = c2d(Gq,Ts,'zoh'); % perts de entrada


atrasoEntrada = totaldelay(Gz);
	
if(mq>0)
    atrasoPert = totaldelay(Gzq);
end

%% PARAMETROS DE AJUSTE DO CONTROLADOR

Nu = [5]; %horizontes de controle - dimensão 1 x m
N1 = [6]; % inicio dos horizontes de predição - incluir atraso - dimensão 1 x n
N2 = [26]; % fim dos horizontes de predição - dimensão 1 x n
Ny = N2-N1+1; %horizonte de predição - não editar
Nq = [0 0]; % % horizonte de predição da perturbação - dimensão 1 x mq

usarFF = true;

Nss = [100]; %horizonte de regime permanente - dimensão 1 x n - apenas par ao DMC

delta = [1]./Ny; %ponderação nos erros - dimensão 1 x n
lambda = [2]./Nu; %ponderação nas ações de controle - dimensão 1 x m       

alpha = [0]; %ajuste do filtro de referencia

umax = [10];
umin = [-10];
dumax = [1];
dumin = -dumax;

Qy = diag(repelem(delta,1,Ny)); %montagem da matriz de ponderação do erro
Qu = diag(repelem(lambda,1,Nu)); %montagem da matriz de ponderação da ação de controle

%% definição do cenário de simulação
nit =250; %tempo de simulação
nin = 10; %número da iteração inicial
nit = nit+nin;


refs = zeros(n,nit+max(N2));
perts = zeros(mq,nit+max(N2));

refs(1,nin+floor(10/Ts):end) = 10;

perts(1,nin+floor(110/Ts):end) = -1;
perts(2,nin+floor(200/Ts):end) = 1;


%% Controlador DMC - MIMO

%%% obtenção dos modelos
modDegrauU = {};
modDegrauQ = {}; 
%%%%%%%%%
for i = 1:n
    for j = 1:m
        modDegrauU{i,j} = step(Gz(i,j),Ts:Ts:Nss(i)*Ts);
    end
    for j = 1:mq
        modDegrauQ{i,j} = step(Gzq(i,j),0:Ts:(Nss(i)-1)*Ts);
    end
end


%% Monta as matrizes offline que definem o problema MIMO DMC
Gm = {};
Gmq = {};
for k=1:n 
    for j=1:m
        for i=1:Nu(j)
            Gm{k,j}(i:Ny(k),i) = modDegrauU{k,j}(N1(k):N2(k)-i+1);
        end
    end
    
    for j=1:mq
        for i=1:Nq(j)
            Gmq{k,j}(i:Ny(k),i) = modDegrauQ{k,j}(N1(k):N2(k)-i+1);
        end
    end
end
G = cell2mat(Gm);
Gq = cell2mat(Gmq);

Hoti = 2*(G'*Qy*G+Qu);
boti1 = -2*G'*Qy; 

Kdmc = (G'*Qy*G+Qu)\G'*Qy;

Kdmc1 = [];
for i=1:m
    Kdmc1 = [Kdmc1; Kdmc(sum(Nu(1:i-1))+1,:)];
end

%%% preparação das matrizes de restrições
LB = repelem(dumin,Nu');
UB = repelem(dumax,Nu');
Arest = [];
for i=1:m
    Arest = blkdiag(Arest,tril(ones(Nu(i))));
end
Arest = [Arest;-Arest];


%% inicializacao dos estados e variaveis da simulação
% - obtencao dos numerados e denominadores
numR = cell(n,m);
denR = cell(n,m);
numDR = cell(n,mq);
denDR = cell(n,mq);


for i=1:n
	% obtencao das respostas ao degrau das entradas
	for j=1:m
		% obtencao dos numeradores e denominadores para o simulador
        [numR{i}{j}, denR{i}{j}] = tfdata(Gz(i,j),'v');
	end
    
	% obtencao das respostas ao degrau das perts
    for j=1:mq
        [numDR{i}{j},denDR{i}{j}]=tfdata(Gzq(i,j),'v');
	end
end


estadosEntr = cell(n,m);
for i=1:n
    for j=1:m
		estadosEntr{i}{j} = zeros(1,nit);
    end
end

estadosPert = cell(n,mq);
for i=1:n
	for j=1:mq
		estadosPert{i}{j}= zeros(1,nit);
	end    
end

saidas = zeros(n,nit); % vetor da saída
entradas = zeros(m,nit); % vetor do sinal de controle
du = zeros(m,nit); % vetor dos incrementos de controle
dq = zeros(mq,nit); % vetor de incrementos das perturbações

rfant = refs(:,nin);

fma = {};
for i=1:n
        fma{i} = zeros(Nss(i),1);
end

for k = nin:nit
    %% -- simulador, não alterar
    %%% Simulador MIMO que utiliza estados
    for i=1:n
        aux = 0;
        for j=1:m
            sizeN = size(numR{i}{j},2);
            sizeD = size(denR{i}{j},2);

            str1 = numR{i}{j}*entradas(j,k-atrasoEntrada(i,j):-1:k+1-sizeN-atrasoEntrada(i,j))';
            str2 = -denR{i}{j}(1,2:end)*estadosEntr{i}{j}(k-1:-1:k-sizeD+1)';
            estadosEntr{i}{j}(k) = str1+str2;

            aux = aux + estadosEntr{i}{j}(k);
        end

        for j=1:mq
            sizeN = size(numDR{i}{j}, 2);
            sizeD = size(denDR{i}{j}, 2);

            str1 = numDR{i}{j}*perts(j,k-atrasoPert(i,j):-1:k+1-sizeN-atrasoPert(i,j))';
            str2 = -denDR{i}{j}(1,2:end)*estadosPert{i}{j}(k-1:-1:k-sizeD+1)';
            estadosPert{i}{j}(k) = str1+str2;

            aux = aux + estadosPert{i}{j}(k);
        end
        saidas(i,k) = aux ;%;+ 0.01*randn(1,1);
    end
    %% -- Controlador DMC 
    dq(:,k) = perts(:,k)-perts(:,k-1);
    
    dqf = [];
    for i=1:mq
        dqf = [dqf;(perts(i,k+1:k+Nq(i))-perts(i,k:k+Nq(i)-1))'];
    end
    
    %%% referencias
    %%% filtro de referência
    rf = alpha'.*rfant + (1-alpha').*refs(:,k);
    rfant = rf;
    
    R = repelem(rf,Ny');
    
    %%% calculo da resposta livre
    ep = zeros(n,1); % erros de predição
    
    f = []; %vetor de resposta livre corrigida
    
    %%%% atualização da resposta livre
    for i = 1:n
        for j = 1:m
            fma{i} = fma{i} + modDegrauU{i,j}*du(j,k-1);
        end
        
        if(usarFF)
            for j=1:mq
                fma{i} = fma{i} + modDegrauQ{i,j}*dq(j,k);
            end
        end
        
        ep(i) = saidas(i,k)-fma{i}(1);
        
        fma{i} = [fma{i}(2:Nss(i)); fma{i}(Nss(i))];
        f = [f; fma{i}(N1(i):N2(i)) + ones(Ny(i),1)*ep(i)];
    end
    
    if(~isempty(Gq) && usarFF) %%% termo das perturbações futuras
        f = f+Gq*dqf;
    end
    
    %% Resolve o problema de otimização
    
    boti = boti1*(R(:)-f);
    
    brest = repelem(umax'-entradas(:,k-1),Nu');
    brest = [brest;
             repelem(entradas(:,k-1)-umin',Nu')];
    X = quadprog(Hoti,boti,Arest,brest,[],[],LB,UB);
    
    for i=1:m
        du(i,k) = X(sum(Nu(1:i-1))+1,1);
    end
    

%     du(:,k) = Kdmc1*(R-f);
    entradas(:,k) = entradas(:,k-1)+du(:,k);

end


figure
subplot(2,1,1)
plot(saidas')
subplot(2,1,2)
plot(entradas')

saidasDMC = saidas;
entradasDMC = entradas;
duDMC = du;


%% Gera gráficos


