clear all
close all
clc

run('../../Bibliotecas/parametrosFiguras.m')


%% modelo não-linear e seus parâmetros
global k1 k2 Vc Qf
k1 = 10;
k2 = 10;
Vc = 1;
Qf = 1/30;


Cfbar = 3.288
Cobar = 1.316



%% modelo linearizado

s = tf('s');
% Definir os modelos que relacionam {saida x,entrada y};
G = 3.433/(103.1*s-1);

Ts = 5; %período de amostragem, OBS: manter atrasos múltiplos inteiros do Ts

%%% discretização do processo
Gz = c2d(G,Ts,'zoh');

%% PARAMETROS DE AJUSTE DO CONTROLADOR
Nu = [5]; %horizontes de controle - dimensão 1 x m
N1 = [1]; % inicio dos horizontes de predição - incluir atraso - dimensão 1 x n
N2 = [30]; % fim dos horizontes de predição - dimensão 1 x n
Ny = N2-N1+1; %horizonte de predição - não editar

Nss = [40]; %horizonte de regime permanente - dimensão 1 x n - apenas para o DMC
Nf = 25; % horizonte de regime permanente para o GDMC

delta = 1; %ponderação nos erros - dimensão 1 x n
lambda = 1; %ponderação nas ações de controle - dimensão 1 x m       

umax = [0.05]+Cfbar;
umin = [-0.05]+Cfbar;
dumax = 0.025;
dumin = -dumax;
ymax = 0.12+Cobar;
ymin = -0.05+Cobar;

Qy = delta*eye(Ny); %montagem da matriz de ponderação do erro
Qu = lambda*eye(Nu); %montagem da matriz de ponderação da ação de controle

%% definição do cenário de simulação
nit =round(800/Ts); %tempo de simulação
nin = max(Nss,Nf)+10; %número da iteração inicial
nit = nit+nin;


refs = Cobar*ones(1,nit+max(N2));
perts = zeros(1,nit+max(N2));

refs(1,nin+round(50/Ts):end) = -0.05+Cobar;
refs(1,nin+round(200/Ts):end) = 0.1+Cobar;
refs(1,nin+round(400/Ts):end) = 0.05+Cobar;
refs(1,nin+round(600/Ts):end) =+Cobar;


perts(1,nin+round(200/Ts):end) = 0;


%% Obtenção da resposta ao degrau a partir da FT linearizada
modDegrauU = step(Gz,Ts:Ts:2*Nss*Ts);


%% Monta as matrizes offline que definem o problema DMC
G = [];
for i=1:Nu
    G(i:Ny,i) = modDegrauU(N1:N2-i+1);
end


Hoti = 2*(G'*Qy*G+Qu);
boti1 = -2*G'*Qy; 

Kdmc = (G'*Qy*G+Qu)\G'*Qy; % utilizado apenas no caso irrestrito

%%% preparação das matrizes de restrições
LB = repelem(dumin,Nu');
UB = repelem(dumax,Nu');

Arest = tril(ones(Nu));
Arest = [Arest;-Arest];

Arest = [Arest;
         G;
         -G];


%% cálculo dos filtros dos erros de predição (SISO) para o GDMC
betaf = 0.7; % formato do filtro (az+b)/(z-betaf)^2

z = tf('z',Ts);
F = tf(0,1,Ts);
nf = 2;

pz = pole(Gz); % obtem o polo indesejado (instável) de malha aberta

for i=N1(1):N2(1)
    %%% monta sistema para obtenção dos parâmetros do filtro
    %%% primeira equação (z^i - F_i(z) = 0 -> para z=pz
    %%% segunda equação F_i(1) = 0 -> força ganho unitário
    indf = i-N1(1)+1;
    Af = [pz 1;
          1 1];
    bf = [pz^i*(pz-betaf)^2;
          (1-betaf)^2];
    X = Af\bf;
    F(indf,1) = (X(1)*z+X(2))/(z-betaf)^2;
    %%% teste da condição
%     pz^i-(X(1)*pz+X(2))/(pz-betaf)^2

    %%% armazena coeficientes gtil
    modDegrauUF{i} = filter(F(indf,1).num{1},F(indf,1).den{1},modDegrauU);

end


%%% calcula a matriz H para o cálculo da resposta livre no caso GDMC
H1 = [];
H2 = [];

for i=N1(1):N2(1)
    H1 = [H1;modDegrauU(i+1:i+Nf)'];
    H2 = [H2;modDegrauUF{i}(1:Nf)'];
    
end
H = H1-H2


%%% calcula a matriz H para o cálculo da resposta livre no caso DMC
H1dmc = [];
H2dmc = [];

for i=N1(1):N2(1)
    H1dmc = [H1dmc;modDegrauU(i+1:i+Nss)'];
    H2dmc = [H2dmc;modDegrauU(1:Nss)'];
    
end
Hdmc = H1dmc-H2dmc



%% Simulação do caso GDMC
%%% inicializacao dos estados e variaveis da simulação
saidas = Cobar*ones(1,nit); % vetor da saída
entradas = Cfbar*ones(1,nit); % vetor do sinal de controle
% Du = zeros(1,nit); % vetor dos sinais de controle em torno do ponto de operação
du = zeros(1,nit); % vetor dos incrementos de controle
yfilt = Cobar*ones(Ny(1),nit);


for k = nin:nit
    %% -- simulador
    saidas(k) = simModelo(saidas(k-1),entradas(k-1)+perts(k-1),Ts);
    
    %% -- Controlador GDMC 
    %%% referencias
    R = repelem(refs(k),Ny');
    
    %%% calculo da resposta livre
    for i=1:Ny(1)
        yfilt(i,k) = -F(i,1).den{1}(2:end)*yfilt(i,k-1:-1:k-nf)' + F(i,1).num{1}*saidas(1,k:-1:k-nf)';
    end
    
    f = H*du(1,k-1:-1:k-Nf)' + yfilt(:,k);
    
    %% Resolve o problema de otimização
    boti = boti1*(R(:)-f);
    
    brest = [repelem(umax'-entradas(:,k-1),Nu)';
             repelem(entradas(:,k-1)-umin',Nu)';
             ymax-f;
             -ymin+f];
    X = quadprog(Hoti,boti,Arest,brest,[],[],LB,UB);
    
    du(k) = X(1);
%     du(:,k) = Kdmc1*(R-f);
    entradas(:,k) = entradas(:,k-1)+du(:,k);

end


figure
subplot(2,1,1)
plot(saidas')
hold on
plot(refs)

subplot(2,1,2)
plot(entradas')

saidasGDMC = saidas;
entradasGDMC = entradas;
duGDMC = du;

%% Simulação do caso DMC
%%% inicializacao dos estados e variaveis da simulação
saidas = Cobar*ones(1,nit); % vetor da saída
entradas = Cfbar*ones(1,nit); % vetor do sinal de controle
% Du = zeros(1,nit); % vetor dos sinais de controle em torno do ponto de operação
du = zeros(1,nit); % vetor dos incrementos de controle


for k = nin:nit
    %% -- simulador
    saidas(k) = simModelo(saidas(k-1),entradas(k-1)+perts(k-1),Ts);
    
    %% -- Controlador GDMC 
    %%% referencias
    R = repelem(refs(k),Ny');
    
    %%% calculo da resposta livre
   
    f = Hdmc*du(1,k-1:-1:k-Nss)' + saidas(:,k);
    
    %% Resolve o problema de otimização
    boti = boti1*(R(:)-f);
    
    brest = [repelem(umax'-entradas(:,k-1),Nu)';
             repelem(entradas(:,k-1)-umin',Nu)';
             ymax-f;
             -ymin+f];
    X = quadprog(Hoti,boti,Arest,brest,[],[],LB,UB);
    
    %%% caso dê infactível, remover a restrição na saída
    if(isempty(X))
        brest = [repelem(umax'-entradas(:,k-1),Nu)';
                 repelem(entradas(:,k-1)-umin',Nu)';
                ];
        X = quadprog(Hoti,boti,Arest(1:2*Nu,:),brest,[],[],LB,UB);
    end
        
    
    du(k) = X(1);
%     du(:,k) = Kdmc1*(R-f);
    entradas(:,k) = entradas(:,k-1)+du(:,k);

end


figure
subplot(2,1,1)
plot(saidas')
hold on
plot(refs)

subplot(2,1,2)
plot(entradas')

saidasDMC = saidas;
entradasDMC = entradas;
duDMC = du;


%% Gera gráficos

t = (0:nit-nin-1)*Ts;
cores = gray(4);
cores = cores(1:end-1,:);


hf = figure
h=subplot(3,1,1)
plot(t,saidasGDMC(1,nin+1:nit),'LineWidth',tamlinha,'Color',cores(1,:))
hold on
plot(t,saidasDMC(1,nin+1:nit),'-.','LineWidth',tamlinha,'Color',cores(2,:))
plot(t,refs(1,nin+1:nit),'--','LineWidth',tamlinha,'Color',cores(3,:))
ylim([1.2 1.5])
hl = legend('GDMC','DMC','Referência','Location','NorthEast')
% hl.Position = [0.6785 0.7018 0.2368 0.1136];
ylabel('C_o (mol/L)','FontSize', tamletra)
set(h, 'FontSize', tamletra);
grid on

h = subplot(3,1,2)
plot(t,entradasGDMC(1,nin+1:nit),'LineWidth',tamlinha,'Color',cores(1,:))
hold on
plot(t,entradasDMC(1,nin+1:nit),'-.','LineWidth',tamlinha,'Color',cores(2,:))
h.YTick = [3.2400 3.2800 3.3200]

ylabel('C_f (mol/L)','FontSize', tamletra)
grid on
set(h, 'FontSize', tamletra);

h = subplot(3,1,3)
plot(t,duGDMC(1,nin+1:nit),'LineWidth',tamlinha,'Color',cores(1,:))
hold on
plot(t,duDMC(1,nin+1:nit),'-.','LineWidth',tamlinha,'Color',cores(2,:))

ylabel('\Delta u','FontSize', tamletra)
xlabel('Tempo (s)','FontSize', tamletra)
grid on
set(h, 'FontSize', tamletra);

% h=subplot(3,1,3)
% plot(t,erro(1:nit-N2-1),'LineWidth',tamlinha,'Color',cores(1,:))
% title('Erro', 'FontSize', tamletra);
% xlabel('tempo', 'FontSize', tamletra);
% set(h, 'FontSize', tamletra);


hf.Position = tamfigura;
hl.Position = [0.7559 0.5691 0.2054 0.1806];

ax = get(gcf,'children');
ind = find(isgraphics(ax,'Legend'));
set(gcf,'children',ax([ind:end,1:ind-1]))

print('simEstudoInstavel_2.eps','-depsc')
% print('simEstudoInstavel_2_pb.eps','-deps')

