clear all
close all
clc

run('../../Bibliotecas/parametrosFiguras.m')


%% modelo não-linear e seus parâmetros
global k1 k2 Vc Qf
k1 = 10;
k2 = 10;
Vc = 1;
Qf = 1/30;


Cfbar = 3.288
Cobar = 1.316



%% modelo linearizado

s = tf('s');
% Definir os modelos que relacionam {saida x,entrada y};
G = 3.433/(103.1*s-1);

Ts = 5; %período de amostragem, OBS: manter atrasos múltiplos inteiros do Ts

%%% discretização do processo
Gz = c2d(G,Ts,'zoh');

%% PARAMETROS DE AJUSTE DO CONTROLADOR
Nu = [5]; %horizontes de controle - dimensão 1 x m
N1 = [1]; % inicio dos horizontes de predição - incluir atraso - dimensão 1 x n
N2 = [30]; % fim dos horizontes de predição - dimensão 1 x n
Ny = N2-N1+1; %horizonte de predição - não editar

Nss = [40]; %horizonte de regime permanente - dimensão 1 x n - apenas para o DMC
Nf = 25; % horizonte de regime permanente para o GDMC

delta = 1; %ponderação nos erros - dimensão 1 x n
lambda = 1; %ponderação nas ações de controle - dimensão 1 x m       

umax = [0.05]+Cfbar;
umin = [-0.05]+Cfbar;
dumax = 0.025;
dumin = -dumax;
ymax = 0.12+Cobar;
ymin = -0.05+Cobar;

Qy = delta*eye(Ny); %montagem da matriz de ponderação do erro
Qu = lambda*eye(Nu); %montagem da matriz de ponderação da ação de controle

%% definição do cenário de simulação
nit =round(400/Ts); %tempo de simulação
nin = max(Nss,Nf)+10; %número da iteração inicial
nit = nit+nin;


refs = Cobar*ones(1,nit+max(N2));
perts = zeros(1,nit+max(N2));

refs(1,nin+round(50/Ts):end) = 0.1+Cobar;

perts(1,nin+round(200/Ts):end) = -.03;


%% Obtenção da resposta ao degrau a partir da FT linearizada
modDegrauU = step(Gz,Ts:Ts:2*Nss*Ts);


%% Monta as matrizes offline que definem o problema DMC
G = [];
for i=1:Nu
    G(i:Ny,i) = modDegrauU(N1:N2-i+1);
end


Hoti = 2*(G'*Qy*G+Qu);
boti1 = -2*G'*Qy; 

Kdmc = (G'*Qy*G+Qu)\G'*Qy; % utilizado apenas no caso irrestrito

%%% preparação das matrizes de restrições
LB = repelem(dumin,Nu');
UB = repelem(dumax,Nu');

Arest = tril(ones(Nu));
Arest = [Arest;-Arest];

Arest = [Arest;
         G;
         -G];


%% cálculo dos filtros dos erros de predição (SISO) para o GDMC
betaf = 0.5; % formato do filtro (az+b)/(z-betaf)^2

sim_3

saidas1 = saidas;
entradas1 = entradas;
du1 = du;

%%%
betaf = 0.6; % formato do filtro (az+b)/(z-betaf)^2

sim_3

saidas2 = saidas;
entradas2 = entradas;
du2 = du;

%%%
betaf = 0.7; % formato do filtro (az+b)/(z-betaf)^2

sim_3

saidas3 = saidas;
entradas3 = entradas;
du3 = du;

%%%
betaf = 0.8; % formato do filtro (az+b)/(z-betaf)^2

sim_3

saidas4 = saidas;
entradas4 = entradas;
du4 = du;



%% Gera gráficos

t = (0:nit-nin-1)*Ts;

numPlots = 4;
ramp = linspace(0, 0.75, numPlots);
grayColors = [ramp; ramp; ramp]';

hf = figure
h=subplot(2,1,1)
plot(t,saidas1(1,nin+1:nit),'LineWidth',tamlinha,'Color',grayColors(1,:))
hold on
plot(t,saidas2(1,nin+1:nit),'LineWidth',tamlinha,'Color',grayColors(2,:))
plot(t,saidas3(1,nin+1:nit),'LineWidth',tamlinha,'Color',grayColors(3,:))
plot(t,saidas4(1,nin+1:nit),'LineWidth',tamlinha,'Color',grayColors(4,:))

plot(t,refs(1,nin+1:nit),'--','LineWidth',tamlinha,'Color','k')
ylim([1.3 1.43])
hl = legend('z_f=0,5','z_f=0,6','z_f=0,7','z_f=0,8','Referência','Location','NorthEast')
% hl.Position = [0.6785 0.7018 0.2368 0.1136];
ylabel('C_o (mol/L)','FontSize', tamletra)
set(h, 'FontSize', tamletra);
grid on

% set(gca, 'ColorOrder', colormap(gray(5+1)))


h = subplot(2,1,2)
plot(t,entradas1(1,nin+1:nit),'LineWidth',tamlinha,'Color',grayColors(1,:))
hold on
plot(t,entradas2(1,nin+1:nit),'LineWidth',tamlinha,'Color',grayColors(2,:))
plot(t,entradas3(1,nin+1:nit),'LineWidth',tamlinha,'Color',grayColors(3,:))
plot(t,entradas4(1,nin+1:nit),'LineWidth',tamlinha,'Color',grayColors(4,:))

% h.YTick = [3.2400 3.2800 3.3200]

ylabel('C_f (mol/L)','FontSize', tamletra)
grid on
set(h, 'FontSize', tamletra);

% h = subplot(3,1,3)
% plot(t,du1(1,nin+1:nit),'LineWidth',tamlinha)
% hold on
% plot(t,du2(1,nin+1:nit),'-.','LineWidth',tamlinha)
% plot(t,du3(1,nin+1:nit),'-.','LineWidth',tamlinha)
% plot(t,du4(1,nin+1:nit),'-.','LineWidth',tamlinha)
% 
% 
% ylabel('\Delta u','FontSize', tamletra)
xlabel('Tempo (s)','FontSize', tamletra)
grid on
set(h, 'FontSize', tamletra);

% h=subplot(3,1,3)
% plot(t,erro(1:nit-N2-1),'LineWidth',tamlinha)
% title('Erro', 'FontSize', tamletra);
% xlabel('tempo', 'FontSize', tamletra);
% set(h, 'FontSize', tamletra);

% set(gca, 'ColorOrder', colormap(gray(4+1)))


hf.Position = tamfigura;
hl.Position = [0.7523 0.4288 0.2054 0.3710];

ax = get(gcf,'children');
ind = find(isgraphics(ax,'Legend'));
set(gcf,'children',ax([ind:end,1:ind-1]))

print('simEstudoInstavel_3.eps','-depsc')
% print('simEstudoInstavel_3_pb.eps','-deps')

