clear all
close all
clc

run('../../Bibliotecas/parametrosFiguras.m')

%% modelo
modDegrau = [0.2,0.6,1.4,1.8,2,2,2,2,2,2,2]';

Ts = 1;
Gz = tf([0 0.2 0.4 0.8 0.4 0.2],[1],Ts,'Variable','z^-1')
na = size(Gz.den{1},2)-1

figure
step(Gz)

%% PARAMETROS DE AJUSTE DO CONTROLADOR
Nu = [2]; %horizontes de controle - dimensão 1 x m
N1 = [1]; % inicio dos horizontes de predição - incluir atraso - dimensão 1 x n
N2 = [3]; % fim dos horizontes de predição - dimensão 1 x n
Ny = N2-N1+1; %horizonte de predição - não editar

Nss = [5]; %horizonte de regime permanente - dimensão 1 x n - apenas para o DMC

delta = 1; %ponderação nos erros - dimensão 1 x n
lambda = 0.5; %ponderação nas ações de controle - dimensão 1 x m       


Qy = delta*eye(Ny); %montagem da matriz de ponderação do erro
Qu = lambda*eye(Nu); %montagem da matriz de ponderação da ação de controle

%% definição do cenário de simulação
nit =round(40/Ts); %tempo de simulação
nin = max(Nss)+10; %número da iteração inicial
nit = nit+nin;


refs = zeros(1,nit+max(N2));
perts = zeros(1,nit+max(N2));

refs(1,nin+round(4/Ts):end) = 1;

perts(1,nin+round(25/Ts):end) = 2;
dq = perts(1,1:end)-[0 perts(1,1:end-1)];

%% Obtenção da resposta ao degrau a partir da FT linearizada
modDegrauU = step(Gz,Ts:Ts:2*Nss*Ts);
modDegrauQ = step(Gz,0:Ts:2*Nss*Ts);

%% Monta as matrizes offline que definem o problema DMC
G = [];
for i=1:Nu
    G(i:Ny,i) = modDegrauU(N1:N2-i+1);
end


Hoti = 2*(G'*Qy*G+Qu);
boti1 = -2*G'*Qy; 

Kdmc = (G'*Qy*G+Qu)\G'*Qy; % utilizado apenas no caso irrestrito
Kdmc1 = Kdmc(1,:);

%%% calcula a matriz H para o cálculo da resposta livre no caso DMC
H1dmc = [];
H2dmc = [];

H1qdmc = [];
H2qdmc = [];

for i=N1(1):N2(1)
    H1dmc = [H1dmc;modDegrauU(i+1:i+Nss)'];
    H2dmc = [H2dmc;modDegrauU(1:Nss)'];

    H1qdmc = [H1qdmc;modDegrauQ(i+1:i+Nss+1)'];
    H2qdmc = [H2qdmc;modDegrauQ(1:Nss+1)'];    
    
end
H = H1dmc-H2dmc
Hq = H1qdmc-H2qdmc


%%% controlador equivalente
f0 = sum(Kdmc1);
Hu = conv([1 -1],[1 Kdmc1*H])
Huq = Kdmc1*Hq;

Cr = tf(f0,Hu,Ts,'Variable','z^-1')
Cff = -tf(Huq,[1 Kdmc1*H],Ts,'Variable','z^-1')

mf = feedback(Cr*Gz,1)

%%% controlador PID aproximado
Cpid = tf([0.7457, -0.8722, +0.305],conv([1 -1],[1 -0.1624]),Ts)
% Fpid = tf(sum(Cpid.num{1})*[1 0 0],Cpid.num{1},Ts)
% Fpid = tf((1-0.3848)*[1 0],[1 -0.3848],Ts)
% Fpid = tf(sum(Cpid.num{1})*[1 -0.4 0]./(1-0.4),Cpid.num{1},Ts)
Fpid = tf(0.65582*[1 -0.9 0.265],[1 -1.17 0.409],Ts)
Fpid = Fpid/dcgain(Fpid)

Cffpid = -tf([0.7251,0.5322,0.4166],[1,0.3706,0.4397],Ts)

mfpid = feedback(Cpid*Gz,1)

figure
step(mf,mfpid,mfpid*Fpid)
legend('1','2','3')

figure
bodemag(mf,mfpid,mfpid*Fpid)

% figure
% step(mf,mfpid*Ftool)
% legend('1','2')

%% Simulação do caso DMC

%%% inicializacao dos estados e variaveis da simulação
saidas = zeros(1,nit); % vetor da saída
entradas = zeros(1,nit); % vetor do sinal de controle
du = zeros(1,nit); % vetor dos incrementos de controle




for k = nin:nit
    %% -- simulador
    saidas(k) = -Gz.den{1}(2:end)*saidas(k-1:-1:k-na)' + Gz.num{1}*(entradas(k:-1:k-na)'+perts(1,k:-1:k-na)');
    
    %% -- Controlador GDMC 
    %%% referencias
    R = refs(k)*ones(Ny,1);
    
    f = H*du(1,k-1:-1:k-Nss)'  + saidas(k) + Hq*dq(1,k:-1:k-Nss)';
    
    %% Resolve o problema de otimização
    du(:,k) = Kdmc1*(R-f);
    entradas(:,k) = entradas(:,k-1)+du(:,k);

end


figure
subplot(2,1,1)
plot(saidas')
hold on
plot(refs)
% ylim([0 3])

subplot(2,1,2)
plot(entradas')

saidasDMC = saidas;
entradasDMC = entradas;
duDMC = du;

%% Simulação do caso PID

%%% inicializacao dos estados e variaveis da simulação
saidas = zeros(1,nit); % vetor da saída
rf = zeros(1,nit);
erro = zeros(1,nit);
entradas = zeros(1,nit); % vetor do sinal de controle
du = zeros(1,nit); % vetor dos incrementos de controle
ua = zeros(1,nit);
uf = zeros(1,nit);



for k = nin:nit
    %% -- simulador
    saidas(k) = -Gz.den{1}(2:end)*saidas(k-1:-1:k-na)' + Gz.num{1}*(entradas(k:-1:k-na)'+perts(1,k:-1:k-na)');
    
    %% -- Controlador PID
    rf(k) = -Fpid.den{1}(2:end)*rf(k-1:-1:k-2)' + Fpid.num{1}*refs(k:-1:k-2)';
    
    erro(k) = rf(k) - saidas(k);
    
    %%% realimentacao
    uf(k) = -Cpid.den{1}(2:end)*uf(k-1:-1:k-2)'+ Cpid.num{1}*erro(k:-1:k-2)';
    
    %%% antecipativo
    ua(k) = -Cffpid.den{1}(2:end)*ua(k-1:-1:k-2)'+ Cffpid.num{1}*perts(k:-1:k-2)';

    entradas(k) = ua(k) + uf(k);
    du(k) = entradas(k)-entradas(k-1);
end


figure
subplot(2,1,1)
plot(saidas')
hold on
plot(refs)
% ylim([0 3])

subplot(2,1,2)
plot(entradas')

saidasPID = saidas;
entradasPID = entradas;
duPID = du;


%% características frequenciais dos controladores
[mag1,pha1,w1] = bode(Cr);
[mag2,pha2] = bode(Cpid,w1);
[mag3,pha3,w2] = bode(Cff);
[mag4,pha4] = bode(Cffpid,w2);


%% Gera gráficos
t = (0:nit-nin-1)*Ts;

cores = gray(4);
cores = cores(1:end-1,:);



hf = figure
h=subplot(2,1,1)
plot(t,saidasDMC(1,nin+1:nit),'LineWidth',tamlinha,'Color',cores(1,:))
hold on
plot(t,saidasPID(1,nin+1:nit),'-.','LineWidth',tamlinha,'Color',cores(2,:))
plot(t,refs(1,nin+1:nit),'--','LineWidth',tamlinha,'Color',cores(3,:))
% ylim([0 1.6])
% h.YTick = [0 0.5 1 1.5];
hl = legend('DMC','PID','Referência','Location','NorthEast')
% hl.Position = [0.6785 0.7018 0.2368 0.1136];
ylabel('Controlada','FontSize', tamletra)
set(h, 'FontSize', tamletra);
grid on


h = subplot(2,1,2)
plot(t,entradasDMC(1,nin+1:nit),'LineWidth',tamlinha,'Color',cores(1,:))
hold on
plot(t,entradasPID(1,nin+1:nit),'-.','LineWidth',tamlinha,'Color',cores(2,:))
% h.YTick = [-2 -1 0 1 2]
% ylim([-2.5 2])

ylabel('Manipulada','FontSize', tamletra)
grid on
xlabel('Tempo (segundos)','FontSize', tamletra)

set(h, 'FontSize', tamletra);

hf.Position = tamfigura;
hl.Position = [0.7470 0.4659 0.2054 0.1806];

ax = get(gcf,'children');
ind = find(isgraphics(ax,'Legend'));
set(gcf,'children',ax([ind:end,1:ind-1]))

print('exemplo_ceq_mao.eps','-depsc')
print('exemplo_ceq_mao_pb.eps','-deps')

%%
hf = figure
h = subplot(2,1,1)
semilogx(w1,20*log10(mag1(:)),'LineWidth',tamlinha,'Color',cores(1,:))
hold on
semilogx(w1,20*log10(mag2(:)),'--','LineWidth',tamlinha,'Color',cores(2,:))
% semilogx(w,20*log10(mag3(:)),'-.','LineWidth',tamlinha,'Color',cores(1,:))
semilogx([2*pi/Ts/2, 2*pi/Ts/2],[-100 100],'k') 
set(h,'Fontsize',tamletra)
% ylim([-10 15])
xlim([min(w1) max(w1)])
% h.YTick = [0 0.5 1 1.5];
hl = legend('DMC','PID','Location','NorthEast')
% hl.Position = [0.6785 0.7018 0.2368 0.1136];
% ylabel('Controlada','FontSize', tamletra)
grid on


ylabel('Magnitude (dB)','FontSize', tamletra)
grid on

h = subplot(2,1,2)
semilogx(w1,pha1(:),'LineWidth',tamlinha,'Color',cores(1,:))
hold on
semilogx(w1,pha2(:),'--','LineWidth',tamlinha,'Color',cores(2,:))
% semilogx(w,20*log10(mag3(:)),'-.','LineWidth',tamlinha,'Color',cores(1,:))
semilogx([2*pi/Ts/2, 2*pi/Ts/2],[-100 100],'k') 
grid on
xlim([min(w1) max(w1)])
ylabel('Fase (graus)', 'FontSize',tamletra)

xlabel('Frequência (rad/s)','FontSize', tamletra)
set(h,'Fontsize',tamletra)
% set(gcf, 'FontSize', tamletra);

hf.Position = tamfigura;
% hl.Position = [0.1527 0.1769 0.2054 0.1806];

ax = get(gcf,'children');
ind = find(isgraphics(ax,'Legend'));
set(gcf,'children',ax([ind:end,1:ind-1]))

print('exemplo_ceq_mao_ceq.eps','-depsc')
print('exemplo_ceq_mao_ceq_pb.eps','-deps')


hf = figure
h = subplot(2,1,1)
semilogx(w2,20*log10(mag3(:)),'LineWidth',tamlinha,'Color',cores(1,:))
hold on
semilogx(w2,20*log10(mag4(:)),'--','LineWidth',tamlinha,'Color',cores(2,:))
% semilogx(w,20*log10(mag3(:)),'-.','LineWidth',tamlinha,'Color',cores(1,:))
semilogx([2*pi/Ts/2, 2*pi/Ts/2],[-100 100],'k') 

% ylim([-10 15])
xlim([10^-1 max(w1)])
% h.YTick = [0 0.5 1 1.5];
hl = legend('DMC','PID','Location','SouthWest')
% hl.Position = [0.6785 0.7018 0.2368 0.1136];
% ylabel('Controlada','FontSize', tamletra)
grid on



ylabel('Magnitude (dB)','FontSize', tamletra)
grid on
set(h,'Fontsize',tamletra)

h = subplot(2,1,2)
semilogx(w2,pha3(:),'LineWidth',tamlinha,'Color',cores(1,:))
hold on
semilogx(w2,pha4(:),'--','LineWidth',tamlinha,'Color',cores(2,:))
% semilogx(w,20*log10(mag3(:)),'-.','LineWidth',tamlinha,'Color',cores(1,:))
semilogx([2*pi/Ts/2, 2*pi/Ts/2],[-100 100],'k') 
grid on
xlim([10^-1 max(w1)])
ylabel('Fase (graus)', 'FontSize',tamletra)

xlabel('Frequência (rad/s)','FontSize', tamletra)

% set(gcf, 'FontSize', tamletra);
set(h,'Fontsize',tamletra)

hf.Position = tamfigura;
% hl.Position = [0.1527 0.1769 0.2054 0.1806];

ax = get(gcf,'children');
ind = find(isgraphics(ax,'Legend'));
set(gcf,'children',ax([ind:end,1:ind-1]))

print('exemplo_ceq_mao_cff.eps','-depsc')
print('exemplo_ceq_mao_cff_pb.eps','-deps')
