clear all
close all
clc

run('../../Bibliotecas/parametrosFiguras.m')


%% modelo linearizado

Ts = 1;
z = tf('z',1);
%%% discretização do processo
Gz = 0.1/(z-1.1)/z^2
na = size(Gz.den{1},2)-1

Gzr = 0.1/(z-1.1)/z^4;
nar = size(Gzr.den{1},2)-1


%% PARAMETROS DE AJUSTE DO CONTROLADOR
Nu = [5]; %horizontes de controle - dimensão 1 x m
N1 = [3]; % inicio dos horizontes de predição - incluir atraso - dimensão 1 x n
N2 = [25]; % fim dos horizontes de predição - dimensão 1 x n
Ny = N2-N1+1; %horizonte de predição - não editar

Nss = [80]; %horizonte de regime permanente - dimensão 1 x n - apenas para o DMC
Nf = 80; % horizonte de regime permanente para o GDMC

delta = 1; %ponderação nos erros - dimensão 1 x n
lambda = 1; %ponderação nas ações de controle - dimensão 1 x m       


Qy = delta*eye(Ny); %montagem da matriz de ponderação do erro
Qu = lambda*eye(Nu); %montagem da matriz de ponderação da ação de controle

%% definição do cenário de simulação
nit =round(200/Ts); %tempo de simulação
nin = max(Nss,Nf)+10; %número da iteração inicial
nit = nit+nin;


refs = zeros(1,nit+max(N2));
perts = zeros(1,nit+max(N2));

refs(1,nin+round(4/Ts):end) = 0.5;

perts(1,nin+round(50/Ts):end) = 0.2;


%% Obtenção da resposta ao degrau a partir da FT linearizada
modDegrauU = step(Gz,Ts:Ts:2*Nss*Ts);

%% Monta as matrizes offline que definem o problema DMC
G = [];
for i=1:Nu
    G(i:Ny,i) = modDegrauU(N1:N2-i+1);
end


Hoti = 2*(G'*Qy*G+Qu);
boti1 = -2*G'*Qy; 

Kdmc = (G'*Qy*G+Qu)\G'*Qy; % utilizado apenas no caso irrestrito
Kdmc1 = Kdmc(1,:);


%% cálculo dos filtros dos erros de predição (SISO) para o GDMC
betaf = 0.7; % formato do filtro (az^2+bz)/(z-betaf)^2

z = tf('z',Ts);
F = tf(0,1,Ts);
nf = 2;

pz1 = 1.1; % polo da FT da entrada

for i=N1(1):N2(1)
    %%% monta sistema para obtenção dos parâmetros do filtro
    %%% primeira equação (z^i - F_i(z) = 0 -> para z=pz1
    %%% segunda equação (z^i - F_i(z) = 0 -> para z=pz2
    %%% terceira equação F_i(1) = 0 -> força ganho unitário
    indf = i-N1(1)+1;
    Af = [pz1^2 pz1;
          1 1];
    bf = [pz1^i*(pz1-betaf)^2;
          (1-betaf)^2];
    X = Af\bf;
    F(indf,1) = (X(1)*z^2+X(2)*z)/(z-betaf)^2;
    %%% teste da condição
%     pz^i-(X(1)*pz+X(2))/(pz-betaf)^2

    %%% armazena coeficientes gtil
    modDegrauUF{i} = filter(F(indf,1).num{1},F(indf,1).den{1},modDegrauU);

end


%%% calcula a matriz H para o cálculo da resposta livre no caso GDMC
H1 = [];
H2 = [];

for i=N1(1):N2(1)
    H1 = [H1;modDegrauU(i+1:i+Nf)'];
    H2 = [H2;modDegrauUF{i}(1:Nf)'];
    
    
end
H = H1-H2


num = 0

for i = 1:size(Kdmc1,2)
    num = num + Kdmc1(i)*F(i).num{1};
end

Fke1 = tf(num,F(1).den{1},Ts)


denc = conv([1 -1],[1 Kdmc1*H]);
numc = [1 zeros(1,size(denc,2)-1)];


Cr = Fke1*tf(numc,denc,Ts)

mf1 = feedback(Cr*Gz,1)

figure
bodemag(mf1)


%% Simulação do caso GDMC1 nominal

%%% inicializacao dos estados e variaveis da simulação
saidas = zeros(1,nit); % vetor da saída
entradas = zeros(1,nit); % vetor do sinal de controle
du = zeros(1,nit); % vetor dos incrementos de controle
yfilt = zeros(Ny(1),nit);




for k = nin:nit
    %% -- simulador
    saidas(k) = -Gz.den{1}(2:end)*saidas(k-1:-1:k-na)' + Gz.num{1}*(entradas(k:-1:k-na)'+perts(1,k:-1:k-na)');
    
    %% -- Controlador GDMC 
    %%% referencias
    R = refs(k)*ones(Ny,1);
    
    %%% calculo da resposta livre
    for i=1:Ny(1)
        yfilt(i,k) = -F(i,1).den{1}(2:end)*yfilt(i,k-1:-1:k-nf)' + F(i,1).num{1}*saidas(1,k:-1:k-nf)';
    end
    
    f = H*du(1,k-1:-1:k-Nf)' + yfilt(:,k);
    
    %% Resolve o problema de otimização
    du(:,k) = Kdmc1*(R-f);
    entradas(:,k) = entradas(:,k-1)+du(:,k);

end


figure
subplot(2,1,1)
plot(saidas')
hold on
plot(refs)

subplot(2,1,2)
plot(entradas')

saidasGDMC = saidas;
entradasGDMC = entradas;
duGDMC = du;

%% Simulação do caso GDMC1 real

%%% inicializacao dos estados e variaveis da simulação
saidas = zeros(1,nit); % vetor da saída
entradas = zeros(1,nit); % vetor do sinal de controle
du = zeros(1,nit); % vetor dos incrementos de controle
yfilt = zeros(Ny(1),nit);




for k = nin:nit
    %% -- simulador
%     saidas(k) = -Gzr.den{1}(2:end)*saidas(k-1:-1:k-nar)' + Gzr.num{1}*(entradas(k:-1:k-nar)'+perts(1,k:-1:k-nar)');
    saidas(k) = -Gzr.den{1}(2:end)*saidas(k-1:-1:k-nar)' + Gzr.num{1}*(entradas(k:-1:k-nar)');
    
    %% -- Controlador GDMC 
    %%% referencias
    R = refs(k)*ones(Ny,1);
    
    %%% calculo da resposta livre
    for i=1:Ny(1)
        yfilt(i,k) = -F(i,1).den{1}(2:end)*yfilt(i,k-1:-1:k-nf)' + F(i,1).num{1}*saidas(1,k:-1:k-nf)';
    end
    
    f = H*du(1,k-1:-1:k-Nf)' + yfilt(:,k);
    
    %% Resolve o problema de otimização
    du(:,k) = Kdmc1*(R-f);
    entradas(:,k) = entradas(:,k-1)+du(:,k);

end


figure
subplot(2,1,1)
plot(saidas')
hold on
plot(refs)

subplot(2,1,2)
plot(entradas')

saidasGDMCr = saidas;
entradasGDMCr = entradas;
duGDMCr = du;


%% cálculo dos filtros dos erros de predição (SISO) para o GDMC
betaf = 0.92; % formato do filtro (az^2+bz)/(z-betaf)^2

z = tf('z',Ts);
F = tf(0,1,Ts);
nf = 2;

pz1 = 1.1; % polo da FT da entrada

for i=N1(1):N2(1)
    %%% monta sistema para obtenção dos parâmetros do filtro
    %%% primeira equação (z^i - F_i(z) = 0 -> para z=pz1
    %%% segunda equação (z^i - F_i(z) = 0 -> para z=pz2
    %%% terceira equação F_i(1) = 0 -> força ganho unitário
    indf = i-N1(1)+1;
    Af = [pz1^2 pz1;
          1 1];
    bf = [pz1^i*(pz1-betaf)^2;
          (1-betaf)^2];
    X = Af\bf;
    F(indf,1) = (X(1)*z^2+X(2)*z)/(z-betaf)^2;
    %%% teste da condição
%     pz^i-(X(1)*pz+X(2))/(pz-betaf)^2

    %%% armazena coeficientes gtil
    modDegrauUF{i} = filter(F(indf,1).num{1},F(indf,1).den{1},modDegrauU);

end


%%% calcula a matriz H para o cálculo da resposta livre no caso GDMC
H1 = [];
H2 = [];

for i=N1(1):N2(1)
    H1 = [H1;modDegrauU(i+1:i+Nf)'];
    H2 = [H2;modDegrauUF{i}(1:Nf)'];
    
    
end
H = H1-H2

num = 0

for i = 1:size(Kdmc1,2)
    num = num + Kdmc1(i)*F(i).num{1};
end

Fke2 = tf(num,F(1).den{1},Ts)


denc = conv([1 -1],[1 Kdmc1*H]);
numc = [1 zeros(1,size(denc,2)-1)];


Cr = Fke2*tf(numc,denc,Ts)

mf2 = feedback(Cr*Gz,1)

figure
bodemag(mf2)



%% Simulação do caso GDMC2 nominal

%%% inicializacao dos estados e variaveis da simulação
saidas = zeros(1,nit); % vetor da saída
entradas = zeros(1,nit); % vetor do sinal de controle
du = zeros(1,nit); % vetor dos incrementos de controle
yfilt = zeros(Ny(1),nit);




for k = nin:nit
    %% -- simulador
    saidas(k) = -Gz.den{1}(2:end)*saidas(k-1:-1:k-na)' + Gz.num{1}*(entradas(k:-1:k-na)'+perts(1,k:-1:k-na)');
    
    %% -- Controlador GDMC 
    %%% referencias
    R = refs(k)*ones(Ny,1);
    
    %%% calculo da resposta livre
    for i=1:Ny(1)
        yfilt(i,k) = -F(i,1).den{1}(2:end)*yfilt(i,k-1:-1:k-nf)' + F(i,1).num{1}*saidas(1,k:-1:k-nf)';
    end
    
    f = H*du(1,k-1:-1:k-Nf)' + yfilt(:,k);
    
    %% Resolve o problema de otimização
    du(:,k) = Kdmc1*(R-f);
    entradas(:,k) = entradas(:,k-1)+du(:,k);

end


figure
subplot(2,1,1)
plot(saidas')
hold on
plot(refs)

subplot(2,1,2)
plot(entradas')

saidasGDMC2 = saidas;
entradasGDMC2 = entradas;
duGDMC2 = du;

%% Simulação do caso GDMC2 real

%%% inicializacao dos estados e variaveis da simulação
saidas = zeros(1,nit); % vetor da saída
entradas = zeros(1,nit); % vetor do sinal de controle
du = zeros(1,nit); % vetor dos incrementos de controle
yfilt = zeros(Ny(1),nit);




for k = nin:nit
    %% -- simulador
%     saidas(k) = -Gzr.den{1}(2:end)*saidas(k-1:-1:k-nar)' + Gzr.num{1}*(entradas(k:-1:k-nar)'+perts(1,k:-1:k-nar)');
    saidas(k) = -Gzr.den{1}(2:end)*saidas(k-1:-1:k-nar)' + Gzr.num{1}*(entradas(k:-1:k-nar)');
    
    %% -- Controlador GDMC 
    %%% referencias
    R = refs(k)*ones(Ny,1);
    
    %%% calculo da resposta livre
    for i=1:Ny(1)
        yfilt(i,k) = -F(i,1).den{1}(2:end)*yfilt(i,k-1:-1:k-nf)' + F(i,1).num{1}*saidas(1,k:-1:k-nf)';
    end
    
    f = H*du(1,k-1:-1:k-Nf)' + yfilt(:,k);
    
    %% Resolve o problema de otimização
    du(:,k) = Kdmc1*(R-f);
    entradas(:,k) = entradas(:,k-1)+du(:,k);

end


figure
subplot(2,1,1)
plot(saidas')
hold on
plot(refs)

subplot(2,1,2)
plot(entradas')

saidasGDMCr2 = saidas;
entradasGDMCr2 = entradas;
duGDMCr2 = du;

%% cálculo dos filtros dos erros de predição (SISO) para o GDMC3
betaf = 0.9; % formato do filtro (az^3+bz^2)/(z-betaf)^3

z = tf('z',Ts);
F = tf(0,1,Ts);
nf = 3;

pz1 = 1.1; % polo da FT da entrada

for i=N1(1):N2(1)
    %%% monta sistema para obtenção dos parâmetros do filtro
    %%% primeira equação (z^i - F_i(z) = 0 -> para z=pz1
    %%% segunda equação (z^i - F_i(z) = 0 -> para z=pz2
    %%% terceira equação F_i(1) = 0 -> força ganho unitário
    indf = i-N1(1)+1;
    Af = [pz1^3 pz1^2;
          1 1];
    bf = [pz1^i*(pz1-betaf)^3;
          (1-betaf)^3];
    X = Af\bf;
    F(indf,1) = (X(1)*z^3+X(2)*z^2)/(z-betaf)^3;
    %%% teste da condição
%     pz^i-(X(1)*pz+X(2))/(pz-betaf)^2

    %%% armazena coeficientes gtil
    modDegrauUF{i} = filter(F(indf,1).num{1},F(indf,1).den{1},modDegrauU);

end


%%% calcula a matriz H para o cálculo da resposta livre no caso GDMC
H1 = [];
H2 = [];

for i=N1(1):N2(1)
    H1 = [H1;modDegrauU(i+1:i+Nf)'];
    H2 = [H2;modDegrauUF{i}(1:Nf)'];
    
    
end
H = H1-H2

num = 0

for i = 1:size(Kdmc1,2)
    num = num + Kdmc1(i)*F(i).num{1};
end

Fke3 = tf(num,F(1).den{1},Ts)


denc = conv([1 -1],[1 Kdmc1*H]);
numc = [1 zeros(1,size(denc,2)-1)];


Cr = Fke3*tf(numc,denc,Ts)

mf3 = feedback(Cr*Gz,1)

figure
bodemag(mf3)

%% Simulação do caso GDMC3 nominal

%%% inicializacao dos estados e variaveis da simulação
saidas = zeros(1,nit); % vetor da saída
entradas = zeros(1,nit); % vetor do sinal de controle
du = zeros(1,nit); % vetor dos incrementos de controle
yfilt = zeros(Ny(1),nit);




for k = nin:nit
    %% -- simulador
    saidas(k) = -Gz.den{1}(2:end)*saidas(k-1:-1:k-na)' + Gz.num{1}*(entradas(k:-1:k-na)'+perts(1,k:-1:k-na)');
    
    %% -- Controlador GDMC 
    %%% referencias
    R = refs(k)*ones(Ny,1);
    
    %%% calculo da resposta livre
    for i=1:Ny(1)
        yfilt(i,k) = -F(i,1).den{1}(2:end)*yfilt(i,k-1:-1:k-nf)' + F(i,1).num{1}*saidas(1,k:-1:k-nf)';
    end
    
    f = H*du(1,k-1:-1:k-Nf)' + yfilt(:,k);
    
    %% Resolve o problema de otimização
    du(:,k) = Kdmc1*(R-f);
    entradas(:,k) = entradas(:,k-1)+du(:,k);

end


figure
subplot(2,1,1)
plot(saidas')
hold on
plot(refs)

subplot(2,1,2)
plot(entradas')

saidasGDMC3 = saidas;
entradasGDMC3 = entradas;
duGDMC3 = du;

%% Simulação do caso GDMC3 real

%%% inicializacao dos estados e variaveis da simulação
saidas = zeros(1,nit); % vetor da saída
entradas = zeros(1,nit); % vetor do sinal de controle
du = zeros(1,nit); % vetor dos incrementos de controle
yfilt = zeros(Ny(1),nit);




for k = nin:nit
    %% -- simulador
%     saidas(k) = -Gzr.den{1}(2:end)*saidas(k-1:-1:k-nar)' + Gzr.num{1}*(entradas(k:-1:k-nar)'+perts(1,k:-1:k-nar)');
    saidas(k) = -Gzr.den{1}(2:end)*saidas(k-1:-1:k-nar)' + Gzr.num{1}*(entradas(k:-1:k-nar)');
    
    %% -- Controlador GDMC 
    %%% referencias
    R = refs(k)*ones(Ny,1);
    
    %%% calculo da resposta livre
    for i=1:Ny(1)
        yfilt(i,k) = -F(i,1).den{1}(2:end)*yfilt(i,k-1:-1:k-nf)' + F(i,1).num{1}*saidas(1,k:-1:k-nf)';
    end
    
    f = H*du(1,k-1:-1:k-Nf)' + yfilt(:,k);
    
    %% Resolve o problema de otimização
    du(:,k) = Kdmc1*(R-f);
    entradas(:,k) = entradas(:,k-1)+du(:,k);

end


figure
subplot(2,1,1)
plot(saidas')
hold on
plot(refs)

subplot(2,1,2)
plot(entradas')

saidasGDMCr3 = saidas;
entradasGDMCr3 = entradas;
duGDMCr3 = du;


%% calculo do índice de robustez
dP = Gzr/Gz - 1

figure
bodemag(dP,1/mf1,1/mf2,1/mf3)

w = logspace(-4,log10(2*pi/2/Ts),200);

[mag1,pha1] = bode(dP,w);
[mag2,pha2] = bode(1/mf1,w);
[mag3,pha3] = bode(1/mf2,w);
[mag4,pha4] = bode(1/mf3,w);



%% Gera gráficos
t = (0:nit-nin-1)*Ts;
cores = gray(4);
cores = cores(1:end-1,:);



hf = figure
h=subplot(2,1,1)
plot(t,saidasGDMC(1,nin+1:nit),'LineWidth',tamlinha,'Color',cores(1,:))
hold on
plot(t,saidasGDMC2(1,nin+1:nit),'-.','LineWidth',tamlinha,'Color',cores(2,:))
plot(t,refs(1,nin+1:nit),'--','LineWidth',tamlinha,'Color',cores(3,:))
% ylim([0 1.6])
% h.YTick = [0 0.5 1 1.5];
hl = legend('GDMC1','GDMC2','Referência','Location','NorthEast')
% hl.Position = [0.6785 0.7018 0.2368 0.1136];
ylabel('Controlada','FontSize', tamletra)
set(h, 'FontSize', tamletra);
grid on

xlim([0 150])

h = subplot(2,1,2)
plot(t,entradasGDMC(1,nin+1:nit),'LineWidth',tamlinha,'Color',cores(1,:))
hold on
plot(t,entradasGDMC2(1,nin+1:nit),'-.','LineWidth',tamlinha,'Color',cores(2,:))
% h.YTick = [-2 -1 0 1 2]
% ylim([-2.5 2])

ylabel('Manipulada','FontSize', tamletra)
grid on
xlabel('Tempo (amostras)','FontSize', tamletra)

set(h, 'FontSize', tamletra);

xlim([0 150])
hf.Position = tamfigura;
hl.Position = [0.7470 0.4659 0.2054 0.1806];

ax = get(gcf,'children');
ind = find(isgraphics(ax,'Legend'));
set(gcf,'children',ax([ind:end,1:ind-1]))

print('exemplo_gdmc_robustez.eps','-depsc')
print('exemplo_gdmc_robustez_pb.eps','-deps')

%%

hf = figure
h=subplot(2,1,1)
plot(t,saidasGDMCr(1,nin+1:nit),'LineWidth',tamlinha,'Color',cores(1,:))
hold on
plot(t,saidasGDMCr2(1,nin+1:nit),'-.','LineWidth',tamlinha,'Color',cores(2,:))
plot(t,refs(1,nin+1:nit),'--','LineWidth',tamlinha,'Color',cores(3,:))
ylim([0 1])
% h.YTick = [0 0.5 1 1.5];
hl = legend('GDMC1','GDMC2','Referência','Location','NorthEast')
% hl.Position = [0.6785 0.7018 0.2368 0.1136];
ylabel('Controlada','FontSize', tamletra)
set(h, 'FontSize', tamletra);
grid on

h = subplot(2,1,2)
plot(t,entradasGDMCr(1,nin+1:nit),'LineWidth',tamlinha,'Color',cores(1,:))
hold on
plot(t,entradasGDMCr2(1,nin+1:nit),'-.','LineWidth',tamlinha,'Color',cores(2,:))
% h.YTick = [-2 -1 0 1 2]
ylim([-1.1 0.7])

ylabel('Manipulada','FontSize', tamletra)
grid on
xlabel('Tempo (amostras)','FontSize', tamletra)

set(h, 'FontSize', tamletra);


hf.Position = tamfigura;
hl.Position = [0.7470 0.4659 0.2054 0.1806];

ax = get(gcf,'children');
ind = find(isgraphics(ax,'Legend'));
set(gcf,'children',ax([ind:end,1:ind-1]))

print('exemplo_gdmc_robustez2.eps','-depsc')
print('exemplo_gdmc_robustez2_pb.eps','-deps')


%%
hf = figure
semilogx(w,20*log10(mag1(:)),'LineWidth',tamlinha,'Color',cores(1,:))
hold on
semilogx(w,20*log10(mag2(:)),'--','LineWidth',tamlinha,'Color',cores(2,:))
semilogx(w,20*log10(mag3(:)),'-.','LineWidth',tamlinha,'Color',cores(3,:))
semilogx([2*pi/Ts/2, 2*pi/Ts/2],[-100 100],'k') 

ylim([-20 15])
% h.YTick = [0 0.5 1 1.5];
hl = legend('\delta G(z)','GDMC1','GDMC2','Location','NorthEast')
% hl.Position = [0.6785 0.7018 0.2368 0.1136];
% ylabel('Controlada','FontSize', tamletra)
grid on


ylabel('Magnitude (dB)','FontSize', tamletra)
grid on
xlabel('Frequência (rad/s)','FontSize', tamletra)

% set(gcf, 'FontSize', tamletra);

hf.Position = tamfigura;
hl.Position = [0.1527 0.1769 0.2054 0.1806];

ax = get(gcf,'children');
ind = find(isgraphics(ax,'Legend'));
set(gcf,'children',ax([ind:end,1:ind-1]))

print('exemplo_gdmc_robustez3.eps','-depsc')
print('exemplo_gdmc_robustez3_pb.eps','-deps')

