clear all
close all
clc

run('../../Bibliotecas/parametrosFiguras.m')


%% modelo linearizado

Ts = 1;
z = tf('z',1);
%%% discretização do processo
Gz = 0.1/(z-1.1)/z^2
Gqz = 0.15/(z-0.95)

na = size(Gz.den{1},2)-1

%% PARAMETROS DE AJUSTE DO CONTROLADOR
Nu = [5]; %horizontes de controle - dimensão 1 x m
N1 = [3]; % inicio dos horizontes de predição - incluir atraso - dimensão 1 x n
N2 = [25]; % fim dos horizontes de predição - dimensão 1 x n
Ny = N2-N1+1; %horizonte de predição - não editar

Nss = [80]; %horizonte de regime permanente - dimensão 1 x n - apenas para o DMC
Nf = 40; % horizonte de regime permanente para o GDMC

delta = 1; %ponderação nos erros - dimensão 1 x n
lambda = 1; %ponderação nas ações de controle - dimensão 1 x m       


Qy = delta*eye(Ny); %montagem da matriz de ponderação do erro
Qu = lambda*eye(Nu); %montagem da matriz de ponderação da ação de controle

%% definição do cenário de simulação
nit =round(250/Ts); %tempo de simulação
nin = max(Nss,Nf)+10; %número da iteração inicial
nit = nit+nin;


refs = zeros(1,nit+max(N2));
perts = zeros(1,nit+max(N2));

refs(1,nin+round(4/Ts):end) = 0.5;

perts(1,nin+round(100/Ts):end) = 0.2;
perts(2,nin+round(200/Ts):end) = 0.1;


%% Obtenção da resposta ao degrau a partir da FT linearizada
modDegrauU = step(Gz,Ts:Ts:2*Nss*Ts);
modDegrauQ{1} = step(Gz,0:Ts:2*Nss*Ts);
modDegrauQ{2} = step(Gqz,0:Ts:2*Nss*Ts);


%% Monta as matrizes offline que definem o problema DMC
G = [];
for i=1:Nu
    G(i:Ny,i) = modDegrauU(N1:N2-i+1);
end


Hoti = 2*(G'*Qy*G+Qu);
boti1 = -2*G'*Qy; 

Kdmc = (G'*Qy*G+Qu)\G'*Qy; % utilizado apenas no caso irrestrito
Kdmc1 = Kdmc(1,:);


%% cálculo dos filtros dos erros de predição (SISO) para o GDMC
betaf = 0.85; % formato do filtro (az+b)/(z-betaf)^2

z = tf('z',Ts);
F = tf(0,1,Ts);
nf = 3;

pz1 = 1.1; % polo da FT da entrada
pz2 = 0.95; % polo da FT da perturbação

for i=N1(1):N2(1)
    %%% monta sistema para obtenção dos parâmetros do filtro
    %%% primeira equação (z^i - F_i(z) = 0 -> para z=pz1
    %%% segunda equação (z^i - F_i(z) = 0 -> para z=pz2
    %%% terceira equação F_i(1) = 0 -> força ganho unitário
    indf = i-N1(1)+1;
    Af = [pz1^2 pz1 1;
          pz2^2 pz2 1
          1 1 1];
    bf = [pz1^i*(pz1-betaf)^3;
          pz2^i*(pz2-betaf)^3;
          (1-betaf)^3];
    X = Af\bf;
    F(indf,1) = (X(1)*z^2+X(2)*z+X(3))/(z-betaf)^3;
    %%% teste da condição
%     pz^i-(X(1)*pz+X(2))/(pz-betaf)^2

    %%% armazena coeficientes gtil
    modDegrauUF{i} = filter(F(indf,1).num{1},F(indf,1).den{1},modDegrauU);
    modDegrauQF{i,1} = filter(F(indf,1).num{1},F(indf,1).den{1},modDegrauQ{1});
    modDegrauQF{i,2} = filter(F(indf,1).num{1},F(indf,1).den{1},modDegrauQ{2});

end


%%% calcula a matriz H para o cálculo da resposta livre no caso GDMC
H1 = [];
H2 = [];

H1q1 = [];
H2q1 = [];

H1q2 = [];
H2q2 = [];


for i=N1(1):N2(1)
    H1 = [H1;modDegrauU(i+1:i+Nf)'];
    H2 = [H2;modDegrauUF{i}(1:Nf)'];
    
    H1q1 = [H1q1;modDegrauQ{1}(i+1:i+Nf+1)'];
    H2q1 = [H2q1;modDegrauQF{i,1}(1:Nf+1)'];
    
    H1q2 = [H1q2;modDegrauQ{2}(i+1:i+Nf+1)'];
    H2q2 = [H2q2;modDegrauQF{i,2}(1:Nf+1)'];
    
end
H = H1-H2
Hq1 = H1q1-H2q1
Hq2 = H1q2-H2q2





%% Simulação do caso GDMC
%%% obtenção do modelo para simulação no formato
%%% A(z)y(k) = B(z)(u(k)+q(k)) + Bq(z)(p(z))
Az = conv(Gz.den{1},Gqz.den{1})
Bz = conv(Gz.num{1},Gqz.den{1})
Bqz = conv(Gqz.num{1},Gz.den{1})

na1 = size(Az,2)-1


%%% inicializacao dos estados e variaveis da simulação
saidas = zeros(1,nit); % vetor da saída
entradas = zeros(1,nit); % vetor do sinal de controle
du = zeros(1,nit); % vetor dos incrementos de controle
yfilt = zeros(Ny(1),nit);




for k = nin:nit
    %% -- simulador
    saidas(k) = -Az(2:end)*saidas(k-1:-1:k-na1)' + Bz*(entradas(k:-1:k-na1)'+perts(1,k:-1:k-na1)')...
                +Bqz*perts(2,k:-1:k-na1)';
    
    %% -- Controlador GDMC 
    %%% referencias
    R = refs(k)*ones(Ny,1);
    
    %%% calculo da resposta livre
    for i=1:Ny(1)
        yfilt(i,k) = -F(i,1).den{1}(2:end)*yfilt(i,k-1:-1:k-nf)' + F(i,1).num{1}*saidas(1,k:-1:k-nf)';
    end
    
    f = H*du(1,k-1:-1:k-Nf)' + yfilt(:,k);
    
    %% Resolve o problema de otimização
    du(:,k) = Kdmc1*(R-f);
    entradas(:,k) = entradas(:,k-1)+du(:,k);

end


figure
subplot(2,1,1)
plot(saidas')
hold on
plot(refs)

subplot(2,1,2)
plot(entradas')

saidasGDMC = saidas;
entradasGDMC = entradas;
duGDMC = du;

%% Simulação do caso GDMC com feed-forward
%%% inicializacao dos estados e variaveis da simulação
saidas = zeros(1,nit); % vetor da saída
entradas = zeros(1,nit); % vetor do sinal de controle
du = zeros(1,nit); % vetor dos incrementos de controle
dq = perts - [zeros(2,1), perts(:,1:end-1)];
yfilt = zeros(Ny(1),nit);




for k = nin:nit
    %% -- simulador
    saidas(k) = -Az(2:end)*saidas(k-1:-1:k-na1)' + Bz*(entradas(k:-1:k-na1)'+perts(1,k:-1:k-na1)')...
                +Bqz*perts(2,k:-1:k-na1)';
    
    %% -- Controlador GDMC 
    %%% referencias
    R = refs(k)*ones(Ny,1);
    
    %%% calculo da resposta livre
    for i=1:Ny(1)
        yfilt(i,k) = -F(i,1).den{1}(2:end)*yfilt(i,k-1:-1:k-nf)' + F(i,1).num{1}*saidas(1,k:-1:k-nf)';
    end
    
    f = H*du(1,k-1:-1:k-Nf)' + yfilt(:,k) + Hq1*dq(1,k:-1:k-Nf)';% + Hq2*dq(2,k:-1:k-Nf)';
    
    %% Resolve o problema de otimização
    du(:,k) = Kdmc1*(R-f);
    entradas(:,k) = entradas(:,k-1)+du(:,k);

end


figure
subplot(2,1,1)
plot(saidas')
hold on
plot(refs)

subplot(2,1,2)
plot(entradas')

saidasGDMC2 = saidas;
entradasGDMC2 = entradas;
duGDMC2 = du;

%% Gera gráficos
t = (0:nit-nin-1)*Ts;

hf = figure
h=subplot(2,1,1)
plot(t,saidasGDMC(1,nin+1:nit),'LineWidth',tamlinha)
hold on
plot(t,saidasGDMC2(1,nin+1:nit),'-.','LineWidth',tamlinha)
plot(t,refs(1,nin+1:nit),'--','LineWidth',tamlinha)
% ylim([0 1.6])
% h.YTick = [0 0.5 1 1.5];
hl = legend('S/ FF','C/ FF','Referência','Location','NorthEast')
% hl.Position = [0.6785 0.7018 0.2368 0.1136];
ylabel('Controlada','FontSize', tamletra)
set(h, 'FontSize', tamletra);
grid on

h = subplot(2,1,2)
plot(t,entradasGDMC(1,nin+1:nit),'LineWidth',tamlinha)
hold on
plot(t,entradasGDMC2(1,nin+1:nit),'-.','LineWidth',tamlinha)
% h.YTick = [-2 -1 0 1 2]
% ylim([-2.5 2])

ylabel('Manipulada','FontSize', tamletra)
grid on
xlabel('Tempo (amostras)','FontSize', tamletra)

set(h, 'FontSize', tamletra);

% h = subplot(3,1,3)
% plot(t,duGDMC(1,nin+1:nit),'LineWidth',tamlinha)
% hold on
% plot(t,duDMC(1,nin+1:nit),'-.','LineWidth',tamlinha)
% 
% ylabel('\Delta u','FontSize', tamletra)
% grid on
% set(h, 'FontSize', tamletra);

% h=subplot(3,1,3)
% plot(t,erro(1:nit-N2-1),'LineWidth',tamlinha)
% title('Erro', 'FontSize', tamletra);
% xlabel('tempo', 'FontSize', tamletra);
% set(h, 'FontSize', tamletra);


hf.Position = tamfigura;
hl.Position = [0.7470 0.4659 0.2054 0.1806];

ax = get(gcf,'children');
ind = find(isgraphics(ax,'Legend'));
set(gcf,'children',ax([ind:end,1:ind-1]))

print('exemplo_gdmc_ff.eps','-depsc')
print('exemplo_gdmc_ff_pb.eps','-deps')

