clear all
close all
clc

run('../../Bibliotecas/parametrosFiguras.m')


%% modelo linearizado

Ts = 1;
z = tf('z',1);
%%% discretização do processo
Gz = 0.05/(z-0.95)*0.1/(z-1.1)/z^2

na = size(Gz.den{1},2)-1

%% PARAMETROS DE AJUSTE DO CONTROLADOR
Nu = [5]; %horizontes de controle - dimensão 1 x m
N1 = [3]; % inicio dos horizontes de predição - incluir atraso - dimensão 1 x n
N2 = [25]; % fim dos horizontes de predição - dimensão 1 x n
Ny = N2-N1+1; %horizonte de predição - não editar

Nss = [200]; %horizonte de regime permanente - dimensão 1 x n - apenas para o DMC

delta = 1; %ponderação nos erros - dimensão 1 x n
lambda = 1; %ponderação nas ações de controle - dimensão 1 x m       


Qy = delta*eye(Ny); %montagem da matriz de ponderação do erro
Qu = lambda*eye(Nu); %montagem da matriz de ponderação da ação de controle

%% definição do cenário de simulação
nit =round(250/Ts); %tempo de simulação
nin = Nss+10; %número da iteração inicial
nit = nit+nin;


refs = zeros(1,nit+max(N2));
perts = zeros(1,nit+max(N2));

refs(1,nin+round(4/Ts):end) = 1;

perts(1,nin+round(100/Ts):end) = 0.5;


%% Obtenção da resposta ao degrau a partir da FT linearizada
modDegrauU = step(Gz,Ts:Ts:2*Nss*Ts);


%% Monta as matrizes offline que definem o problema DMC
G = [];
for i=1:Nu
    G(i:Ny,i) = modDegrauU(N1:N2-i+1);
end


Hoti = 2*(G'*Qy*G+Qu);
boti1 = -2*G'*Qy; 

Kdmc = (G'*Qy*G+Qu)\G'*Qy; % utilizado apenas no caso irrestrito
Kdmc1 = Kdmc(1,:);


%% cálculo dos filtros dos erros de predição (SISO) para o GDMC
betaf = 0.8; % formato do filtro (az+b)/(z-betaf)^2
Nf = 120
z = tf('z',Ts);
F = tf(0,1,Ts);
nf = 2;

pz = 1.1; % obtem o polo indesejado (instável) de malha aberta

for i=N1(1):N2(1)
    %%% monta sistema para obtenção dos parâmetros do filtro
    %%% primeira equação (z^i - F_i(z) = 0 -> para z=pz
    %%% segunda equação F_i(1) = 0 -> força ganho unitário
    indf = i-N1(1)+1;
    Af = [pz^2 pz;
          1 1];
    bf = [pz^i*(pz-betaf)^2;
          (1-betaf)^2];
    X = Af\bf;
    F(indf,1) = (X(1)*z^2+X(2)*z)/(z-betaf)^2;
    %%% teste da condição
%     pz^i-(X(1)*pz+X(2))/(pz-betaf)^2

    %%% armazena coeficientes gtil
    modDegrauUF{i} = filter(F(indf,1).num{1},F(indf,1).den{1},modDegrauU);

end


%%% calcula a matriz H para o cálculo da resposta livre no caso GDMC
H1 = [];
H2 = [];

for i=N1(1):N2(1)
    H1 = [H1;modDegrauU(i+1:i+Nf)'];
    H2 = [H2;modDegrauUF{i}(1:Nf)'];
    
end
H = H1-H2
Hx2 = H;

%% Simulação do caso GDMC
%%% inicializacao dos estados e variaveis da simulação
saidas = zeros(1,nit); % vetor da saída
entradas = zeros(1,nit); % vetor do sinal de controle
du = zeros(1,nit); % vetor dos incrementos de controle
yfilt = zeros(Ny(1),nit);


for k = nin:nit
    %% -- simulador
    saidas(k) = -Gz.den{1}(2:end)*saidas(k-1:-1:k-na)' + Gz.num{1}*(entradas(k:-1:k-na)'+perts(k:-1:k-na)');
    
    %% -- Controlador GDMC 
    %%% referencias
    R = refs(k)*ones(Ny,1);
    
    %%% calculo da resposta livre
    for i=1:Ny(1)
        yfilt(i,k) = -F(i,1).den{1}(2:end)*yfilt(i,k-1:-1:k-nf)' + F(i,1).num{1}*saidas(1,k:-1:k-nf)';
    end
    
    f = H*du(1,k-1:-1:k-Nf)' + yfilt(:,k);
    
    %% Resolve o problema de otimização
    du(:,k) = Kdmc1*(R-f);
    entradas(:,k) = entradas(:,k-1)+du(:,k);

end


figure
subplot(2,1,1)
plot(saidas')
hold on
plot(refs)

subplot(2,1,2)
plot(entradas')

saidasGDMC2 = saidas;
entradasGDMC2 = entradas;
duGDMC2 = du;


%% cálculo dos filtros dos erros de predição (SISO) para o GDMC
betaf = 0.9; % formato do filtro (az+b)/(z-betaf)^2
Nf = 70;
z = tf('z',Ts);
F = tf(0,1,Ts);
nf = 2;

pz1 = 1.1; % obtem o polo indesejado (instável) de malha aberta
pz2 = 0.95;

for i=N1(1):N2(1)
    %%% monta sistema para obtenção dos parâmetros do filtro
    %%% primeira equação (z^i - F_i(z) = 0 -> para z=pz
    %%% segunda equação F_i(1) = 0 -> força ganho unitário
    indf = i-N1(1)+1;
    Af = [pz1^2 pz1 1;
          pz2^2 pz2 1;
          1 1 1];
    bf = [pz1^i*(pz1-betaf)^2;
          pz2^i*(pz2-betaf)^2;
          (1-betaf)^2];
    X = Af\bf;
    F(indf,1) = (X(1)*z^2+X(2)*z+X(3))/(z-betaf)^2;
    %%% teste da condição
%     pz^i-(X(1)*pz+X(2))/(pz-betaf)^2

    %%% armazena coeficientes gtil
    modDegrauUF{i} = filter(F(indf,1).num{1},F(indf,1).den{1},modDegrauU);

end


%%% calcula a matriz H para o cálculo da resposta livre no caso GDMC
H1 = [];
H2 = [];

for i=N1(1):N2(1)
    H1 = [H1;modDegrauU(i+1:i+Nf)'];
    H2 = [H2;modDegrauUF{i}(1:Nf)'];
    
end
H = H1-H2
Hx1 = H;

%% Simulação do caso GDMC
%%% inicializacao dos estados e variaveis da simulação
saidas = zeros(1,nit); % vetor da saída
entradas = zeros(1,nit); % vetor do sinal de controle
du = zeros(1,nit); % vetor dos incrementos de controle
yfilt = zeros(Ny(1),nit);


for k = nin:nit
    %% -- simulador
    saidas(k) = -Gz.den{1}(2:end)*saidas(k-1:-1:k-na)' + Gz.num{1}*(entradas(k:-1:k-na)'+perts(k:-1:k-na)');
    
    %% -- Controlador GDMC 
    %%% referencias
    R = refs(k)*ones(Ny,1);
    
    %%% calculo da resposta livre
    for i=1:Ny(1)
        yfilt(i,k) = -F(i,1).den{1}(2:end)*yfilt(i,k-1:-1:k-nf)' + F(i,1).num{1}*saidas(1,k:-1:k-nf)';
    end
    
    f = H*du(1,k-1:-1:k-Nf)' + yfilt(:,k);
    
    %% Resolve o problema de otimização
    du(:,k) = Kdmc1*(R-f);
    entradas(:,k) = entradas(:,k-1)+du(:,k);

end


figure
subplot(2,1,1)
plot(saidas')
hold on
plot(refs)

subplot(2,1,2)
plot(entradas')

saidasGDMC1 = saidas;
entradasGDMC1 = entradas;
duGDMC1 = du;





%% Gera gráficos
t = (0:nit-nin-1)*Ts;

cores = gray(4);
cores = cores(1:end-1,:);


hf = figure
h=subplot(2,1,1)
plot(t,saidasGDMC1(1,nin+1:nit),'LineWidth',tamlinha,'Color',cores(1,:))
hold on
plot(t,saidasGDMC2(1,nin+1:nit),'-.','LineWidth',tamlinha,'Color',cores(2,:))
plot(t,refs(1,nin+1:nit),'--','LineWidth',tamlinha,'Color',cores(3,:))
% ylim([0 1.6])
h.YTick = [0 0.5 1 1.5 2];
hl = legend('GDMC1','GDMC2','Referência','Location','NorthEast')
% hl.Position = [0.6785 0.7018 0.2368 0.1136];
ylabel('Controlada','FontSize', tamletra)
set(h, 'FontSize', tamletra);
grid on

h = subplot(2,1,2)
plot(t,entradasGDMC1(1,nin+1:nit),'LineWidth',tamlinha,'Color',cores(1,:))
hold on
plot(t,entradasGDMC2(1,nin+1:nit),'-.','LineWidth',tamlinha,'Color',cores(2,:))
% h.YTick = [-2 -1 0 1 2]
% ylim([-2.5 2])

ylabel('Manipulada','FontSize', tamletra)
grid on
xlabel('Tempo (amostras)','FontSize', tamletra)

set(h, 'FontSize', tamletra);


hf.Position = tamfigura;
hl.Position = [0.7470 0.4659 0.2054 0.1806];

ax = get(gcf,'children');
ind = find(isgraphics(ax,'Legend'));
set(gcf,'children',ax([ind:end,1:ind-1]))

print('exemplo_gdmc_instavel_lento.eps','-depsc')
% print('exemplo_gdmc_dmc_pb.eps','-deps')

