clear all
close all
clc

% MIMO DMC (pode ser usado com SISO, para isso, definir 1 ent e 1 saida)


m = 1; % número de entradas da planta - manipuladas
n = 1; %número de saidas da planta - controladas
mq = 1; % número de pertubações

s = tf('s');
% Definir os modelos que relacionam {saida x,entrada y};
G = 1/(-s+1);

Gq = G;

Ts = .1; %período de amostragem, OBS: manter atrasos múltiplos inteiros do Ts

%%% discretização do processo
Gz = c2d(G,Ts,'zoh');
Gzq = c2d(Gq,Ts,'zoh'); % perts de entrada


atrasoEntrada = totaldelay(Gz);
	
if(mq>0)
    atrasoPert = totaldelay(Gzq);
end

%% PARAMETROS DE AJUSTE DO CONTROLADOR

Nu = [5]; %horizontes de controle - dimensão 1 x m
N1 = [1]; % inicio dos horizontes de predição - incluir atraso - dimensão 1 x n
N2 = [30]; % fim dos horizontes de predição - dimensão 1 x n
Ny = N2-N1+1; %horizonte de predição - não editar
Nq = 0;

Nss = [45]; %horizonte de regime permanente - dimensão 1 x n - apenas par ao DMC

delta = [1]./Ny; %ponderação nos erros - dimensão 1 x n
lambda = [1]./Nu; %ponderação nas ações de controle - dimensão 1 x m       

umax = [100];
umin = [-100];
dumax = [100];
dumin = -dumax;

Qy = diag(repelem(delta,1,Ny)); %montagem da matriz de ponderação do erro
Qu = diag(repelem(lambda,1,Nu)); %montagem da matriz de ponderação da ação de controle

%% definição do cenário de simulação
nit =250; %tempo de simulação
nin = Nss+10; %número da iteração inicial
nit = nit+nin;


refs = zeros(n,nit+max(N2));
perts = zeros(mq,nit+max(N2));

refs(1,nin+10:end) = 1;

perts(1,nin+100:end) = -.1;


%% Controlador DMC - MIMO

%%% obtenção dos modelos
modDegrauU = {};
modDegrauQ = {}; 
%%%%%%%%%
for i = 1:n
    for j = 1:m
        modDegrauU{i,j} = step(Gz(i,j),Ts:Ts:2*Nss(i)*Ts);
    end
    for j = 1:mq
        modDegrauQ{i,j} = step(Gzq(i,j),0:Ts:2*(Nss(i)-1)*Ts);
    end
end


%% Monta as matrizes offline que definem o problema MIMO DMC
Gm = {};
Gmq = {};
for k=1:n 
    for j=1:m
        for i=1:Nu(j)
            Gm{k,j}(i:Ny(k),i) = modDegrauU{k,j}(N1(k):N2(k)-i+1);
        end
    end
    
    for j=1:mq
        for i=1:Nq(j)
            Gmq{k,j}(i:Ny(k),i) = modDegrauQ{k,j}(N1(k):N2(k)-i+1);
        end
    end
end
G = cell2mat(Gm);
Gq = cell2mat(Gmq);

Hoti = 2*(G'*Qy*G+Qu);
boti1 = -2*G'*Qy; 

Kdmc = (G'*Qy*G+Qu)\G'*Qy;

Kdmc1 = [];
for i=1:m
    Kdmc1 = [Kdmc1; Kdmc(sum(Nu(1:i-1))+1,:)];
end

%%% preparação das matrizes de restrições
LB = repelem(dumin,Nu');
UB = repelem(dumax,Nu');
Arest = [];
for i=1:m
    Arest = blkdiag(Arest,tril(ones(Nu(i))));
end
Arest = [Arest;-Arest];


%% cálculo dos filtros dos erros de predição (SISO)
betaf = 0.7; % formato do filtro (az+b)/(z-betaf)^2

z = tf('z',Ts);
F = tf(0,1,Ts);
nf = 2;

pz = pole(Gz);

for i=N1(1):N2(1)
    Af = [pz 1;
          1 1];
    bf = [pz^i*(pz-betaf)^2;
          (1-betaf)^2];
    X = Af\bf;
    F(i,1) = (X(1)*z+X(2))/(z-betaf)^2;
    %%% teste da condição
%     pz^i-(X(1)*pz+X(2))/(pz-betaf)^2

    modDegrauUF{i} = filter(F(i,1).num{1},F(i,1).den{1},modDegrauU{1,1});

end


% Nm = 100;
% x = step(Gz,Ts:Ts:Ts*Nm)
% 
% y = filter(F(1,1).num{1},F(1,1).den{1},x)
% 
% y = step(Gz*F(1),0:Ts:Ts*(Nm-1))
% 
% 
% x-y
% 
% figure
% plot(x)
% hold on
% plot(y)

H = [];
H1 = [];
H2 = [];

for i=N1(1):N2(1)
    H1 = [H1;modDegrauU{1,1}(i+1:i+Nss)'];
    H2 = [H2;modDegrauUF{i}(1:Nss)'];
    
end
H = H1-H2



%% inicializacao dos estados e variaveis da simulação
% - obtencao dos numerados e denominadores
numR = cell(n,m);
denR = cell(n,m);
numDR = cell(n,mq);
denDR = cell(n,mq);


for i=1:n
	% obtencao das respostas ao degrau das entradas
	for j=1:m
		% obtencao dos numeradores e denominadores para o simulador
        [numR{i}{j}, denR{i}{j}] = tfdata(Gz(i,j),'v');
	end
    
	% obtencao das respostas ao degrau das perts
    for j=1:mq
        [numDR{i}{j},denDR{i}{j}]=tfdata(Gzq(i,j),'v');
	end
end


estadosEntr = cell(n,m);
for i=1:n
    for j=1:m
		estadosEntr{i}{j} = zeros(1,nit);
    end
end

estadosPert = cell(n,mq);
for i=1:n
	for j=1:mq
		estadosPert{i}{j}= zeros(1,nit);
	end    
end

saidas = zeros(n,nit); % vetor da saída
entradas = zeros(m,nit); % vetor do sinal de controle
du = zeros(m,nit); % vetor dos incrementos de controle
dq = zeros(mq,nit); % vetor de incrementos das perturbações
yfilt = zeros(Ny(1),nit);



for k = nin:nit
    %% -- simulador, não alterar
    %%% Simulador MIMO que utiliza estados
    for i=1:n
        aux = 0;
        for j=1:m
            sizeN = size(numR{i}{j},2);
            sizeD = size(denR{i}{j},2);

            str1 = numR{i}{j}*entradas(j,k-atrasoEntrada(i,j):-1:k+1-sizeN-atrasoEntrada(i,j))';
            str2 = -denR{i}{j}(1,2:end)*estadosEntr{i}{j}(k-1:-1:k-sizeD+1)';
            estadosEntr{i}{j}(k) = str1+str2;

            aux = aux + estadosEntr{i}{j}(k);
        end

        for j=1:mq
            sizeN = size(numDR{i}{j}, 2);
            sizeD = size(denDR{i}{j}, 2);

            str1 = numDR{i}{j}*perts(j,k-atrasoPert(i,j):-1:k+1-sizeN-atrasoPert(i,j))';
            str2 = -denDR{i}{j}(1,2:end)*estadosPert{i}{j}(k-1:-1:k-sizeD+1)';
            estadosPert{i}{j}(k) = str1+str2;

            aux = aux + estadosPert{i}{j}(k);
        end
        saidas(i,k) = aux ;%;+ 0.01*randn(1,1);
    end
    %% -- Controlador DMC 
    dq(:,k) = perts(:,k)-perts(:,k-1);
    
    dqf = [];
    for i=1:mq
        dqf = [dqf;(perts(i,k+1:k+Nq(i))-perts(i,k:k+Nq(i)-1))'];
    end
    
    %%% referencias
    %%% filtro de referÃªncia
    rf = refs(:,k);
    
    R = repelem(rf,Ny');
    
    %%% calculo da resposta livre
    for i=1:Ny(1)
        yfilt(i,k) = -F(i,1).den{1}(2:end)*yfilt(i,k-1:-1:k-nf)' + F(i,1).num{1}*saidas(1,k:-1:k-nf)';
    end
    
    f = H*du(1,k-1:-1:k-Nss)' + yfilt(:,k);
    
    %% Resolve o problema de otimização
    
    boti = boti1*(R(:)-f);
    
    brest = repelem(umax'-entradas(:,k-1),Nu');
    brest = [brest;
             repelem(entradas(:,k-1)-umin',Nu')];
    X = quadprog(Hoti,boti,Arest,brest,[],[],LB,UB);
    
    for i=1:m
        du(i,k) = X(sum(Nu(1:i-1))+1,1);
    end
    

%     du(:,k) = Kdmc1*(R-f);
    entradas(:,k) = entradas(:,k-1)+du(:,k);

end


figure
subplot(2,1,1)
plot(saidas')
hold on
plot(refs)

subplot(2,1,2)
plot(entradas')

saidasDMC = saidas;
entradasDMC = entradas;
duDMC = du;


%% Gera gráficos


