%%% simulação GPC e DMC para o caso com restrições, sem filtro de
%%% referência e com ação antecipativa
clear all
close all
clc


addpath('../../Bibliotecas/')
run('../../Bibliotecas/parametrosFiguras.m')
%% carrega ponto de operação
pontoOperacao

%% Sintonia do GPC e DMC

N1 = 1; %% horizonte de predição inicial
N2 = 30; %% horizonte predição
Ny = N2-N1+1;
Nu = 10; %% horizonte de controle
Nq = 0; %% horizonte de perturbação

lambda = 1; % ponderação do incremento de controle
delta = 1;  % ponderação dos erros futuros


umax = 0.54; % valor de controle máximo
umin = 0; % valor de controle mínimo
dumax = 0.05; % incremento de controle máximo
dumin = -dumax; % incremento de controle mínimo

%% carrega modelos
load('modeloFT')
load('modeloDegrau')

%% Modelo FT
hmax = 5; % nível maximo para normalização

A = Gz.den{1}
B = Gz.num{1}(2)
Bq = [0 Gzq.num{1}(2)]

d=0;
dq = 0;

nb = size(B,2)-1; % ordem do polinômio B
na = size(A,2)-1; % ordem do polinômio A
nbq = size(Bq,2)-1; % ordem do polinômio Bq

%% montagem das matrizes
Btil = conv(B,[zeros(1,d) 1]); % incorporação do atraso no numerador B
Bqtil = conv(Bq,[zeros(1,dq) 1]); % incorporação do atraso no numerador Bq


G = zeros(Ny,Ny); % matriz dinâmica G
H = zeros(Ny,nb+d);
Gq = zeros(Ny,Ny-1); 
Hq = zeros(Ny,nbq+dq);

[E,F] = diofantina(conv(A,[1 -1]),N1,N2);

for i=N1:N2
    EjB = conv(E(i-N1+1,1:i),Btil);
    EjBq = conv(E(i-N1+1,1:i),Bqtil);
    
    G(i-N1+1,1:i) = EjB(i:-1:1);    
    H(i-N1+1,:) = EjB(i+1:end);

    Gq(i-N1+1,1:i) = EjBq(i:-1:1);    
    Hq(i-N1+1,:) = EjBq(i+1:end);
    
end
G = G(:,1:Nu);
Gq = Gq(:,1:Nq);


G,F,H,Gq,Hq

%% obtenção do ganho irrestrito
Qu = lambda*eye(Nu); % matriz de ponderação dos incrementos de controle
Qe = delta*eye(Ny);  % matriz de ponderaão dos erros futuros

Kgpc = (G'*Qe*G+Qu)\G'*Qe;

Kgpc1 = Kgpc(1,:);

%%% matrizes de restrição e do problema de otimização
Hotigpc = 2*(G'*Qe*G+Qu);

Aoti = [tril(ones(Nu));
        -tril(ones(Nu))];
LB = dumin*ones(Nu,1);
UB = dumax*ones(Nu,1);

%% vetores de simulação
nin = 5; % iteração inicial (para inicialização correta)
nit = nin+150; % número de iterações da simulação

refs = h0+zeros(1,nit+Ny); % vetor das referências
perts = as0+zeros(1,nit+Ny); % vetor da perturbação

refs(nin+10:end) = h0*1.1;
perts(1,nin+75:end) = as0*1.5;


saidasGPC = h0+zeros(1,nit); % vetor da saída
entradasGPC = ae0+zeros(1,nit); % vetor do sinal de controle
duGPC = zeros(1,nit); % vetor dos incrementos de controle


for i=nin+1:nit
    %%% simulação do processo
    saidasGPC(1,i) = simModelo(entradasGPC(1,i-1),perts(1,i-1),saidasGPC(1,i-1));
              
    %%% vetor de referências
    R = (refs(1,i)/hmax)*ones(N2-N1+1,1);
    
    %%% cálculo da resposta livre;
    f = F*(saidasGPC(1,i:-1:i-na)'/hmax);
    
    %%% adição da parcela dos incrementos de controle passados na
    %%% resposta livre
    if(~isempty(H))
        f = f+ H*duGPC(1,i-1:-1:i-nb-d)';
    end
    
    %%% adição da parcela do incremento da perturbação passada na
    %%% resposta livre
    deltaQ = perts(1,i:-1:i-nbq-dq+1)'-perts(1,i-1:-1:i-nbq-dq)';
    if(~isempty(Hq))
        f = f+ Hq*deltaQ;
    end
    
    %%% cálculo do incremento de controle ótimo
    fotigpc = 2*G'*Qe*(f-R);
    botigpc = [(umax-entradasGPC(1,i-1))*ones(Nu,1);
               (entradasGPC(1,i-1)-umin)*ones(Nu,1)];
    
    duOti = quadprog(Hotigpc,fotigpc,Aoti,botigpc,[],[],LB,UB);
    duGPC(1,i) = duOti(1);
    
    %%% cálculo do sinal de controle ótimo
    entradasGPC(1,i) = duGPC(1,i)+entradasGPC(1,i-1);
    
              
end

%% Controlador DMC
%%% montando as matrizes do DMC
M = size(Gstep,1); %% horizonte de modelo
Gstepq = [0;Gstepq];

Gdmc = zeros(N2,Nu);
Gdmc(:,1) = Gstep(1:N2,1);

for i=2:Nu
    Gdmc(i:end,i) = Gdmc(1:end-(i-1),1);    
end

Gdmc = Gdmc(N1:end,:);


Kdmc = inv(Gdmc'*Qe*Gdmc+Qu)*Gdmc'*Qe;

Kdmc1 = Kdmc(1,:);

Ylivre = ones(M,1)*h0; %%% 0 é o valor inicial da saída do sistema em regime permanente

%%% matrizes de restrição e do problema de otimização
Hotidmc = 2*(G'*Qe*G+Qu);
%%% as outras matrizes do problema de otimização e restrição são iguais às
%%% do GPC


%% Simulação DMC
saidasDMC = h0+zeros(1,nit); % vetor da saída
entradasDMC = ae0+zeros(1,nit); % vetor do sinal de controle
duDMC = zeros(1,nit); % vetor dos incrementos de controle


for i=nin+1:nit
    %%% simulação do processo
    saidasDMC(1,i) = simModelo(entradasDMC(1,i-1),perts(1,i-1),saidasDMC(1,i-1));
              
    %% Controlador DMC
    dq = perts(i)-perts(i-1);
    
    Ylivre = Ylivre + Gstep(1:M,1)*duDMC(1,i-1)+ Gstepq(1:M,1)*dq;
    
    eta = (saidasDMC(i)/hmax)-Ylivre(1); %%% erro de predição
    
    Ylivre(1:end-1) = Ylivre(2:end); %%% deslocamento do vetor
    
    %%% resposta livre
    f = Ylivre(N1:N2,1)+eta;
    
    %%% referências
    R = ones(N2-N1+1,1)*(refs(1,i)/hmax);
    
    %%% calculo do incremento de controle ótimo    
    fotidmc = 2*G'*Qe*(f-R);
    botidmc = [(umax-entradasDMC(1,i-1))*ones(Nu,1);
               (entradasDMC(1,i-1)-umin)*ones(Nu,1)];
    
    duOti = quadprog(Hotidmc,fotidmc,Aoti,botidmc,[],[],LB,UB);
    duDMC(1,i) = duOti(1);
        
    %%% calculo da ação de controle real
    entradasDMC(i) = entradasDMC(1,i-1)+duDMC(1,i);    
              
end


%% Geração das figuras
cores = gray(4);
cores = cores(1:end-1,:);

hf = figure
h=subplot(3,1,1)
plot(saidasGPC(1,nin+1:nit),'LineWidth',tamlinha,'Color',cores(1,:))
hold on
plot(saidasDMC(1,nin+1:nit),'-.','LineWidth',tamlinha,'Color',cores(2,:))
plot(refs(1,nin+1:nit),'--','LineWidth',tamlinha,'Color',cores(3,:))
% ylim([0 2])
hl = legend('GPC','DMC','Referência','Location','NorthEast')
hl.Position = [0.7028 0.6412 0.2368 0.1662];
ylabel(['Controlada',newline,'(m)'],'FontSize', tamletra)
set(h, 'FontSize', tamletra);
grid on

h = subplot(3,1,2)
plot(entradasGPC(1,nin+1:nit),'LineWidth',tamlinha,'Color',cores(1,:))
hold on
plot(entradasDMC(1,nin+1:nit),'-.','LineWidth',tamlinha,'Color',cores(2,:))

ylabel('Manipulada','FontSize', tamletra)
grid on
set(h, 'FontSize', tamletra);

h = subplot(3,1,3)
plot(duGPC(1,nin+1:nit),'LineWidth',tamlinha,'Color',cores(1,:))
hold on
plot(duDMC(1,nin+1:nit),'-.','LineWidth',tamlinha,'Color',cores(2,:))

ylabel('\Delta u','FontSize', tamletra)
xlabel('Tempo (s)','FontSize', tamletra)
grid on
set(h, 'FontSize', tamletra);

% h=subplot(3,1,3)
% plot(t,erro(1:nit-N2-1),'LineWidth',tamlinha,'Color',cores(1,:))
% title('Erro', 'FontSize', tamletra);
% xlabel('tempo', 'FontSize', tamletra);
% set(h, 'FontSize', tamletra);

hf.Position = tamfigura;
hl.Position = [0.6792 0.6361 0.2071 0.1806];
print('simDMC_GPC_ff_rest_normalizado','-depsc')




