%%% Exemplo comparação GPC com PID
clear all
close all
clc

addpath('../../Bibliotecas/')
run('../../Bibliotecas/parametrosFiguras.m')

%% Modelo da planta

gz = c2d(tf(1,[2 2 1]),.2);

% A = [1 -1.8006 0.8187];
% B = [0.0093 0.0087];

B = gz.num{1}(2:end)
Bq = gz.num{1}
A = gz.den{1}



d=0;
dq = 0;

nb = size(B,2)-1; % ordem do polinômio B
na = size(A,2)-1; % ordem do polinômio A
nbq = size(Bq,2)-1;

%% Sintonia do GPC

N1 = 1; %% horizonte de predição inicial
N2 = 15; %% horizonte predição
Ny = N2-N1+1;
Nu = 10; %% horizonte de controle

lambda = 1; % ponderação do incremento de controle
delta = 1;  % ponderação dos erros futuros

af = 0; % constante de tempo do filtro de referência

umax = 1.2; % valor de controle máximo
umin = -1.2; % valor de controle mínimo

%% montagem das matrizes
Btil = conv(B,[zeros(1,d) 1]); % incorporação do atraso no numerador B


G = zeros(Ny,Ny); % matriz dinâmica G
H = zeros(Ny,nb+d);

[E,F] = diofantina(conv(A,[1 -1]),N1,N2);

for i=N1:N2
    EjB = conv(E(i-N1+1,1:i),Btil);
%     EjBq = conv(E(i-N1+1,1:i),Bqtil);
    
    G(i-N1+1,1:i) = EjB(i:-1:1);    
    H(i-N1+1,:) = EjB(i+1:end);

    
end
G = G(:,1:Nu);



G,F,H

%%% matrizes de restrições
T = tril(ones(Nu));
Aoti = [T;
        -T];

%% obtenção do ganho irrestrito
Qu = lambda*eye(Nu); % matriz de ponderação dos incrementos de controle
Qe = delta*eye(Ny);  % matriz de ponderaão dos erros futuros

Kmpc = (G'*Qe*G+Qu)\G'*Qe

Kmpc1 = Kmpc(1,:)

Hoti = 2*(G'*Qe*G+Qu);

%% obtenção das funções de transferências

delta = tf([1 -1],[1 0],-1)
den1 = tf([1 Kmpc1*H],[1 zeros(1,size(Kmpc1*H,2))],-1)
C1num = tf(sum(Kmpc1),1,-1)
C2num = tf(Kmpc1*F,[1 zeros(1,size(Kmpc1*F,2)-1)],-1)

%%% controlador de realimentação
Cr = minreal(C2num/den1/delta)

%%% filtro de referência
Fr = minreal(C1num/C2num)


%%% Função de transferência de malha fechada com e sem filtro de referÊncia
Hr = feedback(Cr*gz,1)

Hrf = minreal(Hr*Fr)

%%% função de transferência de malha fechada para a perturbação
Hq = minreal(feedback(1,Cr*gz)*gz)

figure
subplot(2,1,1)
step(Hr,Hrf)
legend('S/ Filtro','C/ Filtro','Location','SouthEast')
title('Resposta para mudanças de referência')

subplot(2,1,2)
step(Hq)
title('Resposta para perturbações')


%% vetores de simulação para o caso GPC
nin = 10; % iteração inicial (para inicialização correta)
nit = nin+100; % número de iterações da simulação

refs = zeros(1,nit); % vetor das referências
saidas = zeros(1,nit); % vetor da saída
perts = zeros(1,nit); % vetor da perturbação
entradas = zeros(1,nit); % vetor do sinal de controle
du = zeros(1,nit); % vetor dos incrementos de controle

refs(nin+10:end) = 1;
perts(1,nin+50:end) = 2;

rfant = 0;

for i=nin+1:nit
    %%% simulação do processo
    saidas(1,i) = -A(2:end)*saidas(1,i-1:-1:i-na)' ...
                  +B*entradas(1,i-d-1:-1:i-1-nb-d)' ...
                  +Bq*perts(1,i-dq:-1:i-nbq-dq)';
              
    %%% vetor de referências
    rf = af*rfant + (1-af)*refs(i);
    rfant = rf;
    R = rf*ones(Ny,1);
    
    %%% cálculo da resposta livre;
    f = F*saidas(1,i:-1:i-na)';
    
    
    if(~isempty(H))
        f = f+ H*du(1,i-1:-1:i-nb-d)'; % parcela dos incrementos de controle
    end
    
    %%% cálculo das matrizes do problema de otimização
    foti = 2*G'*Qe*(f-R);
    
    boti = [(umax-entradas(1,i-1))*ones(Nu,1);
            (entradas(1,i-1)-umin)*ones(Nu,1)];
    
    
    %%% cálculo do incremento de controle ótimo
    duOti = quadprog(Hoti,foti,Aoti,boti);
    
    du(1,i) = duOti(1);
    
    %%% cálculo do sinal de controle ótimo
    entradas(1,i) = du(1,i)+entradas(1,i-1);
    
              
end

%% vetores de simulação para o caso PID
saidasPID = zeros(1,nit); % vetor da saída
entradasPID = zeros(1,nit); % vetor do sinal de controle
duPID = zeros(1,nit); % vetor dos incrementos de controle

reffiltPID = zeros(1,nit);
nfr = size(Fr.den{1},2)-1;

ncr = size(Cr.den{1},2)-1;

for i=nin+1:nit
    %%% simulação do processo
    saidasPID(1,i) = -A(2:end)*saidasPID(1,i-1:-1:i-na)' ...
                  +B*sat(entradasPID(1,i-d-1:-1:i-1-nb-d),umin,umax)' ...
                  +Bq*perts(1,i-dq:-1:i-nbq-dq)';
              
    %%% cálculo da referência filtrada
    reffiltPID(1,i) = -Fr.den{1}(2:end)*reffiltPID(1,i-1:-1:i-nfr)' ...
                      +Fr.num{1}*refs(1,i:-1:i-nfr)';

    
    %%% cálculo do sinal de controle do PID
    entradasPID(1,i) = -Cr.den{1}(2:end)*entradasPID(1,i-1:-1:i-ncr)'...
                       +Cr.num{1}*(reffiltPID(1,i:-1:i-ncr)-saidasPID(1,i:-1:i-ncr))';
                                    
    %%% cálculo do incremento de controle ótimo
    duPID(1,i) = entradasPID(1,i)-entradasPID(1,i-1);
    
end


%% Geração das figuras
cores = gray(4);
cores = cores(1:end-1,:);

hf = figure
h=subplot(3,1,1)
plot(saidas(1,nin+1:nit),'LineWidth',tamlinha,'Color',cores(1,:))
hold on
plot(saidasPID(1,nin+1:nit),'-.','LineWidth',tamlinha,'Color',cores(2,:))
plot(refs(1,nin+1:nit),'--','LineWidth',tamlinha,'Color',cores(3,:))
% ylim([0 2])
hl = legend('GPC', 'PID s/ AW','Referência','Location','NorthEast')
% hl.Position = [0.6921    0.6355    0.2439    0.1662];
ylabel('Controlada','FontSize', tamletra)
set(h, 'FontSize', tamletra);
grid on

h = subplot(3,1,2)
plot(entradas(1,nin+1:nit),'LineWidth',tamlinha,'Color',cores(1,:))
hold on
plot(entradasPID(1,nin+1:nit),'-.','LineWidth',tamlinha,'Color',cores(2,:))
ylabel('Manipulada','FontSize', tamletra)
grid on
set(h, 'FontSize', tamletra);

h = subplot(3,1,3)
plot(du(1,nin+1:nit),'LineWidth',tamlinha,'Color',cores(1,:))
hold on
plot(duPID(1,nin+1:nit),'-.','LineWidth',tamlinha,'Color',cores(2,:))
ylabel('\Delta u','FontSize', tamletra)
xlabel('Tempo (amostras)','FontSize', tamletra)
grid on
set(h, 'FontSize', tamletra);


hf.Position = tamfigura;
hl.Position = [0.6774 0.6909 0.2071 0.1242]; 
print('exemploGPC_PID_1c','-depsc')
print('exemploGPC_PID_1c_pb','-deps')





