%%% Exemplo duplo integrador do capítulo GPC SISO
%%% cálculo do ganho irrestrito
%%% cálculo das funções de transferência equivalentes.

clear all
% close all
clc


addpath('../../Bibliotecas/')

%% Modelo da planta
A = [1 -0.9 0.8 -0.7];
B = [0.3];
Bq = [0 B];
C = 1;
% C = poly([0.4 0.3 0.2]);

d=8;
dq = d;

nb = size(B,2)-1; % ordem do polinômio B
na = size(A,2)-1; % ordem do polinômio A
nbq = size(Bq,2)-1; % ordem do polinômio Bq
nC = size(C,2)-1; % ordem do polinômio C

%% Sintonia do GPC

N1 = 1; %% horizonte de predição inicial
N2 = 50; %% horizonte predição
Ny = N2-N1+1;
Nu = 10; %% horizonte de controle

lambda = 1; % ponderação do incremento de controle
delta = 1;  % ponderação dos erros futuros

af = 0; % polo do filtro de referência


%% montagem das matrizes
Btil = conv(B,[zeros(1,d) 1]); % incorporação do atraso no numerador B
Bqtil = conv(Bq,[zeros(1,dq) 1]); % incorporação do atraso no numerador Bq


G = zeros(Ny,Ny); % matriz dinâmica G
H = zeros(Ny,nb+d);
Fbar = zeros(Ny,max(nC-1,na)+1);
nF = size(Fbar,2);

[E,F] = diofantinaC(conv(A,[1 -1]),C,1,N2);

Ebar = cell(1,N2);

for j=1:N2
    Ebar{j} = E(j,1:j);
    Fbar(j,:) = F(j,:);
    
    for i=1:nC
        if(j-i>0)
            Ebar{j} = Ebar{j} - C(1,i+1)*conv([zeros(1,i),1],Ebar{j-i});
            
            Fbar(j,:) = Fbar(j,:) - C(1,i+1)*Fbar(j-i,:);
        else
            temp = zeros(1,size(Fbar,2));
            temp(1,i-j+1) = C(1,i+1);
            
            Fbar(j,:) = Fbar(j,:) - temp;
        end
    end

    EbarjB = conv(Ebar{j},Btil);
    
    G(j,1:j) = EbarjB(j:-1:1);
    H(j,:) = EbarjB(j+1:end);
    
end

G = G(N1:N2,1:Nu);
H = H(N1:N2,:);
Fbar = Fbar(N1:N2,:);

G,Fbar,H

%% obtenção do ganho irrestrito
Qu = lambda*eye(Nu); % matriz de ponderação dos incrementos de controle
Qe = delta*eye(Ny);  % matriz de ponderaão dos erros futuros

Kmpc = (G'*Qe*G+Qu)\G'*Qe;

Kmpc1 = Kmpc(1,:);

%% vetores de simulação
nin = 15; % iteração inicial (para inicialização correta)
nit = nin+100; % número de iterações da simulação

refs = zeros(1,nit); % vetor das referências
saidas = zeros(1,nit); % vetor da saída
perts = zeros(1,nit); % vetor da perturbação
entradas = zeros(1,nit); % vetor do sinal de controle
du = zeros(1,nit); % vetor dos incrementos de controle

refs(nin+10:end) = 1;
perts(1,nin+50:end) = .1;

rfant = 0;

for i=nin+1:nit
    %%% simulação do processo
    saidas(1,i) = -A(2:end)*saidas(1,i-1:-1:i-na)' ...
                  +B*entradas(1,i-d-1:-1:i-1-nb-d)' ...
                  +Bq*perts(1,i-dq:-1:i-nbq-dq)';
              
    %%% vetor de referências
    rf = af*rfant + (1-af)*refs(i);
    rfant = rf;
    
    R = rf*ones(Ny,1);
    
    %%% cálculo da resposta livre;
    f = Fbar*saidas(1,i:-1:i-nF+1)';
    
    
    if(~isempty(H))
        f = f+ H*du(1,i-1:-1:i-nb-d)'; % parcela dos incrementos de controle
    end
    
    %%% cálculo do incremento de controle ótimo
    du(1,i) = Kmpc1*(R-f);
    
    %%% cálculo do sinal de controle ótimo
    entradas(1,i) = du(1,i)+entradas(1,i-1);
    
              
end

%% Geração das figuras
tamletra=14;
tamnum=14;

figure
h=subplot(3,1,1)
plot(saidas(1,nin+1:nit),'LineWidth',2)
hold on
plot(refs(1,nin+1:nit),'--','LineWidth',2)
% ylim([0 2])
hl = legend('Saída','Referência','Location','NorthEast')
hl.Position = [0.6785 0.7018 0.2368 0.1136];
ylabel('Saída','FontSize', tamletra)
set(h, 'FontSize', tamnum);
grid on

h = subplot(3,1,2)
plot(entradas(1,nin+1:nit),'LineWidth',2)
ylabel('Entrada','FontSize', tamletra)
grid on
set(h, 'FontSize', tamnum);

h = subplot(3,1,3)
plot(du(1,nin+1:nit),'LineWidth',2)
ylabel('\Delta u','FontSize', tamletra)
xlabel('Tempo (amostras)','FontSize', tamletra)
grid on
set(h, 'FontSize', tamnum);

% h=subplot(3,1,3)
% plot(t,erro(1:nit-N2-1),'LineWidth',2)
% title('Erro', 'FontSize', tamletra);
% xlabel('tempo', 'FontSize', tamletra);
% set(h, 'FontSize', tamnum);



% print('exemploAcelPerturbacao_poli_c','-depsc')






