%%% Exemplo duplo integrador do capítulo GPC SISO
%%% cálculo do ganho irrestrito
%%% cálculo das funções de transferência equivalentes.

clear all
close all
clc

addpath('../../Bibliotecas/')
run('parametrosFiguras.m')

%% Modelo da planta
A = conv([1 -1],[1 -1]);
B = 0.2;
Bq = [0.1 0.2];

d=0;
dq = 0;

nb = size(B,2)-1; % ordem do polinômio B
na = size(A,2)-1; % ordem do polinômio A
nbq = size(Bq,2)-1; % ordem do polinômio Bq


%% Sintonia do GPC

N1 = 1; %% horizonte de predição inicial
N2 = 30; %% horizonte predição
Ny = N2-N1+1;
Nu = 3; %% horizonte de controle
Nq = 0; %% horizonte de perturbação

lambda = 1; % ponderação do incremento de controle
delta = 1;  % ponderação dos erros futuros



%% montagem das matrizes
Btil = conv(B,[zeros(1,d) 1]); % incorporação do atraso no numerador B
Bqtil = conv(Bq,[zeros(1,dq) 1]); % incorporação do atraso no numerador Bq


G = zeros(Ny,Ny); % matriz dinâmica G
H = zeros(Ny,nb+d);
Gq = zeros(Ny,Ny-1); 
Hq = zeros(Ny,nbq+dq);

[E,F] = diofantina(conv(A,[1 -1]),N1,N2);

for i=N1:N2
    EjB = conv(E(i-N1+1,1:i),Btil);
    EjBq = conv(E(i-N1+1,1:i),Bqtil);
    
    G(i-N1+1,1:i) = EjB(i:-1:1);    
    H(i-N1+1,:) = EjB(i+1:end);

    Gq(i-N1+1,1:i) = EjBq(i:-1:1);    
    Hq(i-N1+1,:) = EjBq(i+1:end);
    
end
G = G(:,1:Nu);
Gq = Gq(:,1:Nq);


G,F,H,Gq,Hq

%% obtenção do ganho irrestrito
Qu = lambda*eye(Nu); % matriz de ponderação dos incrementos de controle
Qe = delta*eye(Ny);  % matriz de ponderaão dos erros futuros

Kmpc = (G'*Qe*G+Qu)\G'*Qe;

Kmpc1 = Kmpc(1,:);

%% vetores de simulação
nin = 5; % iteração inicial (para inicialização correta)
nit = nin+100; % número de iterações da simulação

refs = zeros(1,nit+Ny); % vetor das referências
saidas = zeros(1,nit); % vetor da saída
perts = zeros(1,nit+Ny); % vetor da perturbação
entradas = zeros(1,nit); % vetor do sinal de controle
du = zeros(1,nit); % vetor dos incrementos de controle

refs(nin+10:end) = 1;
perts(1,nin+50:end) = 0.5;

for i=nin+1:nit
    %%% simulação do processo
    saidas(1,i) = -A(2:end)*saidas(1,i-1:-1:i-na)' ...
                  +B*entradas(1,i-d-1:-1:i-1-nb-d)' ...
                  +Bq*perts(1,i-dq:-1:i-nbq-dq)';
              
    %%% vetor de referências (caso futuro e conhecido)
    R = refs(1,i+N1:i+N2)';
    
    %%% cálculo da resposta livre;
    f = F*saidas(1,i:-1:i-na)';
    
    %%% adição da parcela dos incrementos de controle passados na
    %%% resposta livre
    if(~isempty(H))
        f = f+ H*du(1,i-1:-1:i-nb-d)';
    end
    
    %%% adição da parcela do incremento da perturbação passada na
    %%% resposta livre
    deltaQ = perts(1,i:-1:i-nbq-dq+1)'-perts(1,i-1:-1:i-nbq-dq)';
    if(~isempty(Hq))
        f = f+ Hq*deltaQ;
    end
    
    %%% cálculo do incremento de controle ótimo
    du(1,i) = Kmpc1*(R-f);
    
    %%% cálculo do sinal de controle ótimo
    entradas(1,i) = du(1,i)+entradas(1,i-1);
    
              
end

%% Geração das figuras
cores = gray(4);
cores = cores(1:end-1,:);



hf = figure
h=subplot(3,1,1)
plot(saidas(1,nin+1:nit),'LineWidth',tamlinha,'Color',cores(1,:))
hold on
plot(refs(1,nin+1:nit),'--','LineWidth',tamlinha,'Color',cores(2,:))
ylim([0 1.5])
hl = legend('Controlada','Referência','Location','NorthEast')
% hl.Position = [0.6785 0.7018 0.2368 0.1136];
ylabel('Controlada','FontSize', tamletra)
set(h, 'FontSize', tamletra);
grid on

h = subplot(3,1,2)
plot(entradas(1,nin+1:nit),'LineWidth',tamlinha,'Color',cores(1,:))
ylabel('Manipulada','FontSize', tamletra)
grid on
set(h, 'FontSize', tamletra);

h = subplot(3,1,3)
plot(du(1,nin+1:nit),'LineWidth',tamlinha,'Color',cores(1,:))
ylabel('\Delta u','FontSize', tamletra)
xlabel('Tempo (amostras)','FontSize', tamletra)
grid on
set(h, 'FontSize', tamletra);

% h=subplot(3,1,3)
% plot(t,erro(1:nit-N2-1),'LineWidth',tamlinha,'Color',cores(1,:))
% title('Erro', 'FontSize', tamletra);
% xlabel('tempo', 'FontSize', tamletra);
% set(h, 'FontSize', tamletra);

hf.Position = tamfigura;
hl.Position = [0.6952 0.6683 0.2054 0.1242];
print('exemploDuploIntegrador_completo_sint2','-depsc')
% print('exemploDuploIntegrador_completo_sint2_pb','-deps')






