%%% Exemplo duplo integrador do capítulo GPC SISO
%%% cálculo do ganho irrestrito
%%% cálculo das funções de transferência equivalentes.

clear all
% close all
clc


addpath('../../Bibliotecas/')
run('../../Bibliotecas/parametrosFiguras.m')
%% Modelo da planta
A = [1 -1.905, 0.905];
num = [0.09675 0.09358]; 
B = 4*num;
Bq = -[0 num];

d=0;
dq = d;

nb = size(B,2)-1; % ordem do polinômio B
na = size(A,2)-1; % ordem do polinômio A
nbq = size(Bq,2)-1; % ordem do polinômio Bq


%% Sintonia do GPC1

N1 = 1; %% horizonte de predição inicial
N2 = 30; %% horizonte predição
Ny = N2-N1+1;
Nu = 5; %% horizonte de controle
Nq = 0; %% horizonte de perturbação

lambda1 = 1; % ponderação do incremento de controle para o GPC1
lambda2 = 0; % ponderação do incremento de controle para o GPC2
delta = 1;  % ponderação dos erros futuros

af = 0; % polo do filtro de referência

Btil = conv(B,[zeros(1,d) 1]); % incorporação do atraso no numerador B
Bqtil = conv(Bq,[zeros(1,dq) 1]); % incorporação do atraso no numerador Bq


G = zeros(Ny,Ny); % matriz dinâmica G
H = zeros(Ny,nb+d);
Gq = zeros(Ny,Ny-1); 
Hq = zeros(Ny,nbq+dq);

[E,F] = diofantina(conv(A,[1 -1]),N1,N2);

for i=N1:N2
    EjB = conv(E(i-N1+1,1:i),Btil);
    EjBq = conv(E(i-N1+1,1:i),Bqtil);
    
    G(i-N1+1,1:i) = EjB(i:-1:1);    
    H(i-N1+1,:) = EjB(i+1:end);

    Gq(i-N1+1,1:i) = EjBq(i:-1:1);    
    Hq(i-N1+1,:) = EjBq(i+1:end);
    
end
G = G(:,1:Nu);
Gq = Gq(:,1:Nq);


G,F,H,Gq,Hq


%% obtenção do ganho irrestrito
Qu1 = lambda1*eye(Nu); % matriz de ponderação dos incrementos de controle
Qu2 = lambda2*eye(Nu); % matriz de ponderação dos incrementos de controle
Qe = delta*eye(Ny);  % matriz de ponderaão dos erros futuros

Kgpc1 = (G'*Qe*G+Qu1)\G'*Qe;
Kgpc11 = Kgpc1(1,:);

Kgpc2 = (G'*Qe*G+Qu2)\G'*Qe;
Kgpc21 = Kgpc2(1,:);


%% vetores de simulação
nin = 10; % iteração inicial (para inicialização correta)
nit = nin+30; % número de iterações da simulação

refs = zeros(1,nit); % vetor das referências
perts = zeros(1,nit); % vetor da perturbação

% refs(nin+40:end) = 30;
perts(1,nin+5:end) = 20;
% perts(1,nin+70:end) = 30;

%% simulação do GPC1
saidas1 = zeros(1,nit); % vetor da saída
entradas1 = zeros(1,nit); % vetor do sinal de controle
du1 = zeros(1,nit); % vetor dos incrementos de controle


rfant = 0;

for i=nin+1:nit
    %%% simulação do processo
    saidas1(1,i) = -A(2:end)*saidas1(1,i-1:-1:i-na)' ...
                  +B*entradas1(1,i-d-1:-1:i-1-nb-d)' ...
                  +Bq*perts(1,i-dq:-1:i-nbq-dq)';
              
    %%% vetor de referências
    rf = af*rfant + (1-af)*refs(i);
    rfant = rf;
    R = rf*ones(Ny,1);
    
    %%% cálculo da resposta livre;
    f = F*saidas1(1,i:-1:i-na)';
    
    
    if(~isempty(H))
        f = f+ H*du1(1,i-1:-1:i-nb-d)'; % parcela dos incrementos de controle
    end
    
    %%% adição da parcela do incremento da perturbação passada na
    %%% resposta livre
    deltaQ = perts(1,i:-1:i-nbq-dq+1)'-perts(1,i-1:-1:i-nbq-dq)';
    if(~isempty(Hq))
        f = f+ Hq*deltaQ;
    end

    
    %%% cálculo do incremento de controle ótimo
    du1(1,i) = Kgpc11*(R-f);
    
    %%% cálculo do sinal de controle ótimo
    entradas1(1,i) = du1(1,i)+entradas1(1,i-1);
    
              
end

%% simulação do GPC2
saidas2 = zeros(1,nit); % vetor da saída
entradas2 = zeros(1,nit); % vetor do sinal de controle
du2 = zeros(1,nit); % vetor dos incrementos de controle


rfant = 0;

for i=nin+1:nit
    %%% simulação do processo
    saidas2(1,i) = -A(2:end)*saidas2(1,i-1:-1:i-na)' ...
                  +B*entradas2(1,i-d-1:-1:i-1-nb-d)' ...
                  +Bq*perts(1,i-dq:-1:i-nbq-dq)';
              
    %%% vetor de referências
    rf = af*rfant + (1-af)*refs(i);
    rfant = rf;
    R = rf*ones(Ny,1);
    
    %%% cálculo da resposta livre;
    f = F*saidas2(1,i:-1:i-na)';
    
    
    if(~isempty(H))
        f = f+ H*du2(1,i-1:-1:i-nb-d)'; % parcela dos incrementos de controle
    end

    %%% adição da parcela do incremento da perturbação passada na
    %%% resposta livre
    deltaQ = perts(1,i:-1:i-nbq-dq+1)'-perts(1,i-1:-1:i-nbq-dq)';
    if(~isempty(Hq))
        f = f+ Hq*deltaQ;
    end    
    
    %%% cálculo do incremento de controle ótimo
    du2(1,i) = Kgpc21*(R-f);
    
    %%% cálculo do sinal de controle ótimo
    entradas2(1,i) = du2(1,i)+entradas2(1,i-1);
    
              
end


%% Geração das figuras
cores = gray(4);
cores = cores(1:end-1,:);

hf = figure
h=subplot(3,1,1)
plot(saidas1(1,nin+1:nit),'LineWidth',tamlinha,'Color',cores(1,:))
hold on
plot(saidas2(1,nin+1:nit),'-.','LineWidth',tamlinha,'Color',cores(2,:))
plot(refs(1,nin+1:nit),'--','LineWidth',tamlinha,'Color',cores(3,:))
% ylim([0 2])
hl = legend('GPC1','GPC2','Referência','Location','SouthEast')
% hl.Position = [0.6870    0.6663    0.2368    0.1662];
ylabel('Controlada','FontSize', tamletra)
set(h, 'FontSize', tamletra);
grid on

h = subplot(3,1,2)
plot(entradas1(1,nin+1:nit),'LineWidth',tamlinha,'Color',cores(1,:))
hold on
plot(entradas2(1,nin+1:nit),'-.','LineWidth',tamlinha,'Color',cores(2,:))

ylabel('Manipulada','FontSize', tamletra)
grid on
set(h, 'FontSize', tamletra);
% ylim([-5 5])

h = subplot(3,1,3)
plot(du1(1,nin+1:nit),'LineWidth',tamlinha,'Color',cores(1,:))
hold on
plot(du2(1,nin+1:nit),'-.','LineWidth',tamlinha,'Color',cores(2,:))

ylabel('\Delta u','FontSize', tamletra)
xlabel('Tempo (amostras)','FontSize', tamletra)
grid on
% ylim([-5 5])
set(h, 'FontSize', tamletra);

% h=subplot(3,1,3)
% plot(t,erro(1:nit-N2-1),'LineWidth',tamlinha,'Color',cores(1,:))
% title('Erro', 'FontSize', tamletra);
% xlabel('tempo', 'FontSize', tamletra);
% set(h, 'FontSize', tamletra);


hf.Position = tamfigura;
hl.Position = [0.6774 0.6909 0.2071 0.1242]; 
print('exemploMotorDC_ff','-depsc')
% print('exemploMotorDC_ff_pb','-deps')





