%%% Exemplo duplo integrador do capítulo GPC SISO
%%% cálculo do ganho irrestrito
%%% cálculo das funções de transferência equivalentes.

clear all
close all
clc


addpath('../../Bibliotecas/')
run('../../Bibliotecas/parametrosFiguras.m')
%% Modelo da planta
A = [1 -1.905, 0.905];
num = [0.09675 0.09358]; 
B = 4*num;
Bq = -[0 num];

d=0;
dq = d;

nb = size(B,2)-1; % ordem do polinômio B
na = size(A,2)-1; % ordem do polinômio A
nbq = size(Bq,2)-1; % ordem do polinômio Bq


%% Sintonia do GPC

N1 = 1; %% horizonte de predição inicial
N2 = 30; %% horizonte predição
Ny = N2-N1+1;
Nu = 5; %% horizonte de controle

lambda = 1; % ponderação do incremento de controle
delta = 1;  % ponderação dos erros futuros

af = 0; % polo do filtro de referência

Umax = 100;
Umin = 0;

%%% matrizes das restrições de sinal de controle e do problema de
%%% otimização
Arest = [tril(ones(Nu));
         -tril(ones(Nu))];
    


%% montagem das matrizes
Btil = conv(B,[zeros(1,d) 1]); % incorporação do atraso no numerador B
Bqtil = conv(Bq,[zeros(1,dq) 1]); % incorporação do atraso no numerador Bq


G = zeros(Ny,Ny); % matriz dinâmica G
H = zeros(Ny,nb+d);

[E,F] = diofantina(conv(A,[1 -1]),N1,N2);

for i=N1:N2
    EjB = conv(E(i-N1+1,1:i),Btil);
    
    G(i-N1+1,1:i) = EjB(i:-1:1);    
    H(i-N1+1,:) = EjB(i+1:end);
    
end
G = G(:,1:Nu);


G,F,H

%% obtenção do ganho irrestrito
Qu = lambda*eye(Nu); % matriz de ponderação dos incrementos de controle
Qe = delta*eye(Ny);  % matriz de ponderaão dos erros futuros

Kmpc = (G'*Qe*G+Qu)\G'*Qe;

Kmpc1 = Kmpc(1,:);

%%% matriz do problema de otimização com restrições
Hoti = 2*(G'*Qe*G+Qu);
foti = -2*G'*Qe;

%% vetores de simulação
nin = 10; % iteração inicial (para inicialização correta)
nit = nin+100; % número de iterações da simulação

refs = zeros(1,nit); % vetor das referências
saidas = zeros(1,nit); % vetor da saída
perts = zeros(1,nit); % vetor da perturbação
entradas = zeros(1,nit); % vetor do sinal de controle
du = zeros(1,nit); % vetor dos incrementos de controle

refs(nin+40:end) = 30;
perts(1,nin+5:end) = 20;
perts(1,nin+70:end) = 30;

rfant = 0;

for i=nin+1:nit
    %%% simulação do processo
    saidas(1,i) = -A(2:end)*saidas(1,i-1:-1:i-na)' ...
                  +B*entradas(1,i-d-1:-1:i-1-nb-d)' ...
                  +Bq*perts(1,i-dq:-1:i-nbq-dq)';
              
    %%% vetor de referências
    rf = af*rfant + (1-af)*refs(i);
    rfant = rf;
    R = rf*ones(Ny,1);
    
    %%% cálculo da resposta livre;
    f = F*saidas(1,i:-1:i-na)';
    
    
    if(~isempty(H))
        f = f+ H*du(1,i-1:-1:i-nb-d)'; % parcela dos incrementos de controle
    end
    
    %%% cálculo das matrizes de restrição
    brest = [(Umax-entradas(1,i-1))*ones(Nu,1);
             (-Umin+entradas(1,i-1))*ones(Nu,1)];
    
    
    %%% cálculo do incremento de controle ótimo
    duOti = quadprog(Hoti,foti*(R-f),Arest,brest);
%     duOti = quadprog(Hoti,foti*(R-f),[],[]);
    du(1,i) = duOti(1);
    
%     du(1,i) = Kmpc1*(R-f);
    
    %%% cálculo do sinal de controle ótimo
    entradas(1,i) = du(1,i)+entradas(1,i-1);
    
              
end
%%
saidasGPC1 = saidas;
duGPC1 = du;
entradasGPC1 = entradas;

figure
subplot(2,1,1)
plot(saidas)
subplot(2,1,2)
plot(entradas)




%% Geração das figuras
cores = gray(4);
cores = cores(1:end-1,:);

hf = figure
h=subplot(3,1,1)
plot(saidas(1,nin+1:nit),'LineWidth',tamlinha,'Color',cores(1,:))
hold on
plot(refs(1,nin+1:nit),'--','LineWidth',tamlinha,'Color',cores(2,:))
ylim([-15 35])
hl = legend('Controlada','Referência','Location','NorthEast')
% hl.Position = [0.6785 0.7018 0.2368 0.1136];
ylabel('Controlada','FontSize', tamletra)
set(h, 'FontSize', tamletra);
grid on

h = subplot(3,1,2)
plot(entradas(1,nin+1:nit),'LineWidth',tamlinha,'Color',cores(1,:))
ylabel('Manipulada','FontSize', tamletra)
grid on
set(h, 'FontSize', tamletra);
% ylim([-5 5])

h = subplot(3,1,3)
plot(du(1,nin+1:nit),'LineWidth',tamlinha,'Color',cores(1,:))
ylabel('\Delta u','FontSize', tamletra)
xlabel('Tempo (amostras)','FontSize', tamletra)
grid on
% ylim([-5 5])
set(h, 'FontSize', tamletra);

% h=subplot(3,1,3)
% plot(t,erro(1:nit-N2-1),'LineWidth',tamlinha,'Color',cores(1,:))
% title('Erro', 'FontSize', tamletra);
% xlabel('tempo', 'FontSize', tamletra);
% set(h, 'FontSize', tamletra);


hf.Position = tamfigura;
hl.Position = [0.6774 0.6909 0.2071 0.1242]; 
print('exemploMotorDC_rest_u','-depsc')
print('exemploMotorDC_rest_u_pb','-deps')





