%%% Exemplo duplo integrador do capítulo GPC SISO
%%% cálculo do ganho irrestrito
%%% cálculo das funções de transferência equivalentes.

clear all
close all
clc


addpath('../../Bibliotecas/')
run('parametrosFiguras.m')
%% Modelo da planta
A = [1 -0.9];
B = [0.3];
Bq = [0 B];
C = 1;
% C = [1 -0.7];

d=10;
dq = d;

nb = size(B,2)-1; % ordem do polinômio B
na = size(A,2)-1; % ordem do polinômio A
nbq = size(Bq,2)-1; % ordem do polinômio Bq


%% Sintonia do GPC 1

N1 = d+1; %% horizonte de predição inicial
N2 = d+20; %% horizonte predição
Ny = N2-N1+1;
Nu = 5; %% horizonte de controle

lambda = 1; % ponderação do incremento de controle
delta = 1;  % ponderação dos erros futuros

af = 0; % polo do filtro de referência


%% montagem das matrizes
Btil = conv(B,[zeros(1,d) 1]); % incorporação do atraso no numerador B
Bqtil = conv(Bq,[zeros(1,dq) 1]); % incorporação do atraso no numerador Bq


G = zeros(Ny,Ny); % matriz dinâmica G
H = [];

[E,F] = diofantinaC(conv(A,[1 -1]),C,N1,N2);

for i=N1:N2
    EjB = conv(E(i-N1+1,1:i),Btil);
    
    [M,N] = diofantinaC(C,EjB,i,i);
    
    G(i-N1+1,1:i) = M(end:-1:1);    
    H = [H;N];
    
end
G = G(:,1:Nu);

nF = size(F,2);
nC = size(C,2);
nH = size(H,2);

G,F,H

%% obtenção do ganho irrestrito
Qu = lambda*eye(Nu); % matriz de ponderação dos incrementos de controle
Qe = delta*eye(Ny);  % matriz de ponderaão dos erros futuros

Kmpc = (G'*Qe*G+Qu)\G'*Qe;

Kmpc1 = Kmpc(1,:);

%% vetores de simulação
nin = 15; % iteração inicial (para inicialização correta)
nit = nin+100; % número de iterações da simulação

refs = zeros(1,nit); % vetor das referências
saidas = zeros(1,nit); % vetor da saída
saidasC = zeros(1,nit);
perts = zeros(1,nit); % vetor da perturbação
entradas = zeros(1,nit); % vetor do sinal de controle
du = zeros(1,nit); % vetor dos incrementos de controle
duC = zeros(1,nit);

refs(nin+10:end) = 1;
perts(1,nin+50:end) = .2;

rfant = 0;

for i=nin+1:nit
    %%% simulação do processo
    saidas(1,i) = -A(2:end)*saidas(1,i-1:-1:i-na)' ...
                  +B*entradas(1,i-d-1:-1:i-1-nb-d)' ...
                  +Bq*perts(1,i-dq:-1:i-nbq-dq)';
              
    %%% vetor de referências
    rf = af*rfant + (1-af)*refs(i);
    rfant = rf;
    
    R = (rf)*ones(Ny,1);
    
    %%% cálculo da resposta livre;
    saidasC(1,i) = saidas(1,i) -C(2:end)*saidasC(1,i-1:-1:i-nC+1)'; 

    f = F*saidasC(1,i:-1:i-nF+1)';
    
    
    if(~isempty(H))
        f = f+ H*duC(1,i-1:-1:i-nb-d)'; % parcela dos incrementos de controle
    end
    
    %%% cálculo do incremento de controle ótimo
    du(1,i) = Kmpc1*(R-f);
    duC(1,i) = -C(2:end)*duC(1,i-1:-1:i-nC+1)' + du(1,i) ;
    
    %%% cálculo do sinal de controle ótimo
    entradas(1,i) = du(1,i)+entradas(1,i-1);
    
              
end

%%
saidasGPC1 = saidas;
duGPC1 = du;
entradasGPC1 = entradas;

figure
subplot(2,1,1)
plot(saidas)
subplot(2,1,2)
plot(entradas)

%% GPC 2
C = [1 -0.7];

%% montagem das matrizes
Btil = conv(B,[zeros(1,d) 1]); % incorporação do atraso no numerador B
Bqtil = conv(Bq,[zeros(1,dq) 1]); % incorporação do atraso no numerador Bq


G = zeros(Ny,Ny); % matriz dinâmica G
H = [];

[E,F] = diofantinaC(conv(A,[1 -1]),C,N1,N2);

for i=N1:N2
    EjB = conv(E(i-N1+1,1:i),Btil);
    
    [M,N] = diofantinaC(C,EjB,i,i);
    
    G(i-N1+1,1:i) = M(end:-1:1);    
    H = [H;N];
    
end
G = G(:,1:Nu);

nF = size(F,2);
nC = size(C,2);
nH = size(H,2);

G,F,H

%% vetores de simulação
saidas = zeros(1,nit); % vetor da saída
saidasC = zeros(1,nit);
entradas = zeros(1,nit); % vetor do sinal de controle
du = zeros(1,nit); % vetor dos incrementos de controle
duC = zeros(1,nit);


rfant = 0;

for i=nin+1:nit
    %%% simulação do processo
    saidas(1,i) = -A(2:end)*saidas(1,i-1:-1:i-na)' ...
                  +B*entradas(1,i-d-1:-1:i-1-nb-d)' ...
                  +Bq*perts(1,i-dq:-1:i-nbq-dq)';
              
    %%% vetor de referências
    rf = af*rfant + (1-af)*refs(i);
    rfant = rf;
    
    R = (rf)*ones(Ny,1);
    
    %%% cálculo da resposta livre;
    saidasC(1,i) = saidas(1,i) -C(2:end)*saidasC(1,i-1:-1:i-nC+1)'; 

    f = F*saidasC(1,i:-1:i-nF+1)';
    
    
    if(~isempty(H))
        f = f+ H*duC(1,i-1:-1:i-nb-d)'; % parcela dos incrementos de controle
    end
    
    %%% cálculo do incremento de controle ótimo
    du(1,i) = Kmpc1*(R-f);
    duC(1,i) = -C(2:end)*duC(1,i-1:-1:i-nC+1)' + du(1,i) ;
    
    %%% cálculo do sinal de controle ótimo
    entradas(1,i) = du(1,i)+entradas(1,i-1);
    
              
end

%%
saidasGPC2 = saidas;
duGPC2 = du;
entradasGPC2 = entradas;

figure
subplot(2,1,1)
plot(saidas)
subplot(2,1,2)
plot(entradas)



%% Geração das figuras
cores = gray(4);
cores = cores(1:end-1,:);


hf = figure 
h=subplot(3,1,1)
plot(saidasGPC1(1,nin+1:nit),'LineWidth',tamlinha,'Color',cores(1,:))
hold on

plot(saidasGPC2(1,nin+1:nit),'-.','LineWidth',tamlinha,'Color',cores(2,:))
plot(refs(1,nin+1:nit),'--','LineWidth',tamlinha,'Color',cores(3,:))
% ylim([0 2])
hl = legend('C(z^{-1})=1','C(z^{-1})\neq 1','Referência','Location','NorthEast')
% hl.Position = [0.6785 0.7018 0.2368 0.1136];
ylabel('Controlada','FontSize', tamletra)
set(h, 'FontSize', tamletra);
grid on

h = subplot(3,1,2)
plot(entradasGPC1(1,nin+1:nit),'LineWidth',tamlinha,'Color',cores(1,:))
hold on
plot(entradasGPC2(1,nin+1:nit),'-.','LineWidth',tamlinha,'Color',cores(2,:))

ylabel('Manipulada','FontSize', tamletra)
grid on
set(h, 'FontSize', tamletra);

h = subplot(3,1,3)
plot(duGPC1(1,nin+1:nit),'LineWidth',tamlinha,'Color',cores(1,:))
hold on
plot(duGPC2(1,nin+1:nit),'-.','LineWidth',tamlinha,'Color',cores(2,:))

ylabel('\Delta u','FontSize', tamletra)
xlabel('Tempo (amostras)','FontSize', tamletra)
grid on
set(h, 'FontSize', tamletra);

% h=subplot(3,1,3)
% plot(t,erro(1:nit-N2-1),'LineWidth',tamlinha,'Color',cores(1,:))
% title('Erro', 'FontSize', tamletra);
% xlabel('tempo', 'FontSize', tamletra);
% set(h, 'FontSize', tamletra);


hf.Position = tamfigura;
hl.Position = [0.6774 0.6909 0.2071 0.1242]; 
print('exemplo_poli_c_2_novo','-depsc')





