clear all
close all
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%exemplo 52 - MIMO DMC planta integradora 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
N_entradas = 2; % nmero de entradas da planta - manipuladas
N_saidas = 2; %nmero de saidas da planta - controladas
N_perturbacoes = 1;

s = tf('s');

% Escolha de modelo com mais atraso ou menos atraso

at=0.5; % deixa atraso do modelo original
at=0.1;  % divide atraso por 10

% escolha do tau aproxima modelo integrador
tau=3;

% Definir os modelos que relacionam {saida x,entrada y};
G{1,1} = 3.5*tau/(tau*s+1)*exp(-at*s);
G{1,2} = -1/(2*s+1)*exp(-at*5*s);
G{2,1} = -2/(1.5*s+1)*exp(-at*7*s);
G{2,2} = 1/(3.2*s+1)*exp(-at*5*s);

Gnom = G;
Gnom{1,1} = (3.5/(s))*exp(-at*s);
% Definir os modelos que relacionam {saida x,perturbao z};
Gp{1,1} = 3.5*tau/(tau*s+1)*exp(-at*s);
Gp{2,1} = -4.5/(2*s+1)*exp(-at*2*s);
Gpnom = Gp;
Gpnom{1,1} = (3.5/(s))*exp(-at*s);
Ts = 0.1; %perodo de amostragem


%%%Modelo FT para simular a planta integradora
for i = 1:N_saidas
    for j = 1:N_entradas
        sysd{i,j} = c2d(Gnom{i,j},Ts); 
    end
    for j = 1:N_perturbacoes
        syspd{i,j} = c2d(Gpnom{i,j},Ts); 
    end
end
for i = 1:N_saidas
    A{i,i} = [1];
    for j = 1:N_entradas
         A{i,i} = conv(A{i,i},sysd{i,j}.den{1});  
    end
end
T = [1 -1];

for i = 1:N_saidas
    for j = 1:N_entradas
        B{i,j} = deconv(conv(A{i,i},sysd{i,j}.num{1}),sysd{i,j}.den{1});  
        B{i,j} = B{i,j}(2:end);
    end
    for j = 1:N_perturbacoes
        Bp{i,j} = deconv(conv(A{i,i},syspd{i,j}.num{1}),syspd{i,j}.den{1});  
        Bp{i,j} = Bp{i,j}(2:end);
    end
    Atio{i,i} = conv(A{i,i},T);
end

dimB = size(B{1,1},2);
dimA = size(A{1,1},2);

Nrp = [(2*2+5)/Ts (3.2*5+7)/Ts]; %horizonte de regime permanente - dimenso 1 x N_saidas

%% PARAMETROS DE AJUSTE DO Cse ONTROLADOR

Nu = [5 5]; %e controle - dimenso 1 x N_entradas
N1 = [2/Ts 6/Ts]; % inicio dos horizontes de predio - incluir atraso - dimenso 1 x N_saidas
N2 = [40 80]; % fim dos horizontes de predio - dimenso 1 x N_saidas
N = N2-N1+1; %horizonte de predio - no editar

flag_ref_futura = false; %habilita referncias futuras conhecidas
%%%
flag_filtro_ref = true;
F_ref{1} = [3 0.1]; %filtro de referncia - [tau_polo tau_zero]
F_ref{2} = [1 2]; %filtro de referncia - [tau_polo tau_zero]
%%%    
pondY = [5/N(1) 1/N(2)]; %ponderao nos erros - dimenso 1 x N_saidas
pondU = [5/Nu(1) 1/Nu(2)]; %ponderao nas aes de controle - dimenso 1 x N_entradas       

%restrio no incremeto de controle
du_lim = [10 10 -10 -10]; %limites por ao de controle ex:2 [sup_1 sup_2 inf_1 inf_2]
R = [eye(sum(Nu));-eye(sum(Nu))];
r_inc = [repelem(du_lim(1:N_entradas),1,Nu)';repelem(-1*ones(1,N_entradas).*du_lim(N_entradas+1:end),1,Nu)'];
flag_restricao_u = 0; 

Tsim =1000; %tempo de simulao
degraus_ref{1} = [0 0.5]; %Amplitudes dos degraus no ensaio para cada sada
degraus_ref{2} = [0 0.2];
degraus_t{1} = [100 Tsim-100];
degraus_t{2} = [500 Tsim-500];

degraus_pert{1} = [0 0.001];
degraus_pert_t  = [700 Tsim-700];

usar_feed_forward = false;

%%%%%%%%%
for i = 1:N_saidas
    for j = 1:N_entradas
        g{i,j} = step(G{i,j},Ts:Ts:Nrp(i)*Ts);
    end
    for j = 1:N_perturbacoes
        gp{i,j} = step(Gp{i,j},0:Ts:(Nrp(i)-1)*Ts);
    end
    F_refd{i} = c2d((F_ref{i}(2)*s+1)/(F_ref{i}(1)*s+1),Ts);
end

Qy = diag(repelem(pondY,1,N)); %montagem da matriz de ponderao do erro
Qu = diag(repelem(pondU,1,Nu)); %montagem da matriz de ponderao da ao de controle

 
%% PARAMETROS DA SIMULAO
Tini = 4;
Tsim = Tsim + Tini;
for i = 1:N_saidas
    ref{i} = repelem(degraus_ref{i},1,[Tini N(i)+N1(i)]+degraus_t{i});%vetor de referncias - [valores dos degraus] - [repetio do respectivo valor]  
end 
for i = 1:N_perturbacoes
    pert{i} = repelem(degraus_pert{i},1,[Tini N(i)+N1(i)]+degraus_pert_t);%vetor de perturbaes - [valores dos degraus] - [repetio do respectivo valor]
end

 %% RODA SIMULAO
 %inicializa variaveis
for i = 1:N_saidas
     y{i}(1:Tsim) = 0;
     fma{i} = zeros(Nrp(i),1);
     ysim{i} = zeros(Nrp(i),1);
     ref_fil{i}(1:Tsim) = 0;
     if (flag_filtro_ref)
         for j = Tini:size(ref{i},2)
            ref_fil{i}(j) = -F_refd{i}.den{1}(2)*ref_fil{i}(j-1)+ F_refd{i}.num{1}(1)*ref{i}(j)+ F_refd{i}.num{1}(2)*ref{i}(j-1);
         end
     end
end
for i = 1:N_entradas
    for j = 1:Tini
       du{i}(j) = 0;
       u{i}(j) = 0;
    end
end

%% Monta as matrizes offline que definem o problema MIMO DMC

for k=1:N_saidas 
    for j=1:N_entradas
        for i=1:Nu(j)
            Gm{k,j}(i:N(k),i) = g{k,j}(N1(k):N2(k)-i+1);
        end
    end
end
G = cell2mat(Gm);
H = 2*(G'*Qy*G+Qu);
H=(H+H')/2; 

%%%%Simulao
for k=Tini:Tsim
    %%%Roda a planta
        ysim{i} = 0;
    for i = 1:N_saidas
        ysim{i} = -fliplr(Atio{i,i}(2:dimA+1))*y{i}(k-dimA:k-1)';
        for j = 1:N_entradas
            ysim{i} = ysim{i} + fliplr(B{i,j}(1:dimB))*du{j}(k-dimB:k-1)';
        end
        for j = 1:N_perturbacoes
            ysim{i} = ysim{i} + fliplr(Bp{i,j}(1:dimB))*(pert{j}(k)-pert{j}(k-1))';
        end
        y{i}(k) = ysim{i}(1);
    end
    
    %%%Roda o MPC
    for i = 1:N_saidas
        for j = 1:N_entradas
            fma{i} = fma{i} + g{i,j}*du{j}(k-1);
        end
        if usar_feed_forward
            for j = 1:N_perturbacoes
                fma{i} = fma{i} + gp{i,j}*(pert{j}(k)-pert{j}(k-1));
            end
        end
        e = y{i}(k)-fma{i}(1);
        fma{i} = [fma{i}(2:Nrp(i)); fma{i}(Nrp(i))];
        f_aux{i,1} = fma{i}(N1(i):N2(i))+ ones(N(i),1)*e;
    end

    f = cell2mat(f_aux);
    pos = 1;
    if (flag_ref_futura)
        for i = 1:N_saidas
            if (flag_filtro_ref)
                refm(pos:pos+N(i)-1) = ref_fil{i}(k+N1(i):k+N1(i)+N(i)-1);
            else
                refm(pos:pos+N(i)-1) = ref{i}(k+N1(i):k+N1(i)+N(i)-1);
            end
            pos = pos+N(i);
        end
    else
        for i = 1:N_saidas
            if (flag_filtro_ref)
                refm(pos:pos+N(i)-1) = ones(N(i),1)*ref_fil{i}(k);
            else
                refm(pos:pos+N(i)-1) = ones(N(i),1)*ref{i}(k);    
            end
            pos = pos+N(i);
        end
    end
    b = (2*(f-refm')'*Qy*G)';

    if flag_restricao_u
        r_p=[];
        r_m= [];
        for i = 1:N_entradas
            r_p = [r_p; repelem(u_lim(i)-u{i}(k-1),1,Nu(i))'];
            r_m = [r_m; repelem(-u_lim(N_entradas+i)+u{i}(k-1),1,Nu(i))'];
        end
        r_con = [r_p; r_m];
        r = [r_inc; r_con]
    else
        r = r_inc;
    end


    %% Resolve o problema de otimizao
    tic;
    options = optimset('Display', 'off');
    [delta_u] = quadprog(H,b,R,r,[],[],[],[],[],options);
    t_solver = toc;

    %% Atualiza as variveis de controle
    pos = 1;
    for i = 1:N_entradas
        du{i}(k) = delta_u(pos);
        u{i}(k) = u{i}(k-1) + du{i}(k);
        pos = pos+Nu(i);
    end
end

%%%Gera grficos
t = Ts*(Tini:Tsim);
tamletra=20;
tamnum=20;
figure
plot(t,ref{1}(Tini:Tsim),'--',t,y{1}(Tini:Tsim),'-',t,ref{2}(Tini:Tsim),'--',t,y{2}(Tini:Tsim),'-','LineWidth',2)
legend('ref_1','Y_1','ref_2','Y_2','Location','SouthEast')
axis([Tini (Tsim-Tini)*Ts -0.1 0.8]);
 %   ylabel('Sadas e Referncias do Sistemas', 'FontSize', tamletra);
xlabel('Tempo (s)', 'FontSize', tamletra);
ylabel('Controladas', 'FontSize', tamletra);
set(gca,'FontSize',tamnum);
grid on

print -depsc results_con_ex_52_2.eps

figure
plot(t,u{1}(Tini:Tsim),'-',t,u{2}(Tini:Tsim),'-','LineWidth',2)
%title(['y1' ' e Referncia do Sistema'], 'FontSize', tamletra)
legend('U_1','U_2','Location','SouthEast')
axis([Tini (Tsim-Tini)*Ts -0.1 0.5]);
 %   ylabel('Sadas e Referncias do Sistemas', 'FontSize', tamletra);
xlabel('Tempo (s)', 'FontSize', tamletra);
ylabel('Manipuladas', 'FontSize', tamletra);
set(gca,'FontSize',tamnum);
grid on
print -depsc results_man_ex_52_2.eps
