clear all
close all
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%Caso 3_1 - muda a ponderao de apenas 1 ao de controle
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
N_entradas = 3; % nmero de entradas da planta - manipuladas
N_saidas = 3; %nmero de saidas da planta - controladas
N_perturbacoes = 0;

s = tf('s');
% Definir os modelos que relacionam {saida x,entrada y};
G{1,1} = 4.05/(50*s+1)*exp(-27*s);
G{1,2} = 1.77/(60*s+1)*exp(-28*s);
G{1,3} = 5.88/(50*s+1)*exp(-27*s);
G{2,1} = 5.39/(50*s+1)*exp(-18*s);
G{2,2} = 5.72/(60*s+1)*exp(-14*s);
G{2,3} = 6.9/(40*s+1)*exp(-15*s);
G{3,1} = 4.38/(33*s+1)*exp(-20*s);
G{3,2} = 4.42/(44*s+1)*exp(-22*s);
G{3,3} = 7.2/(19*s+1);

Ts = 1.0; %perodo de amostragem
Nrp = [300 300 220]; %horizonte de regime permanente - dimenso 1 x N_saidas

Nu = [5 5 5]; %horizontes de controle - dimenso 1 x N_entradas
N1 = [28 15 1]; % inicio dos horizontes de predio - incluir atraso - dimenso 1 x N_saidas
N2 = [88 75 45]; % fim dos horizontes de predio - dimenso 1 x N_saidas
N = N2-N1+1; %horizonte de predio - no editar

pondY = [1/N(1) 1/N(2) 1/N(3)]; %ponderao nos erros - dimenso 1 x N_saidas
pondU = [1/Nu(1) 1/Nu(2) 1/Nu(3)]; %ponderao nas aes de controle - dimenso 1 x N_entradas       

%restrio no incremeto de controle
du_lim = [10000 10000 10000 -10000 -10000 -10000]; %limites por ao de controle ex:2 [sup_1 sup_2 inf_1 inf_2]
R = [eye(sum(Nu));-eye(sum(Nu))];
r_inc = [repelem(du_lim(1:N_entradas),1,Nu)';repelem(-1*ones(1,N_entradas).*du_lim(N_entradas+1:end),1,Nu)'];
flag_restricao_u = 0;

Tsim =800; %tempo de simulao
degraus_ref{1} = [0 1]; %Amplitudes dos degraus no ensaio para cada sada
degraus_ref{2} = [0 0.5];
degraus_ref{3} = [0 0.2];
degraus_t{1} = [10 Tsim-10];
degraus_t{2} = [280 Tsim-280];
degraus_t{3} = [600 Tsim-600];

degraus_pert{1} = [0 0];
degraus_pert_t  = [0 Tsim-0];

usar_feed_forward = false;

%%%%%%%%%
for i = 1:N_saidas
    for j = 1:N_entradas
        g{i,j} = step(G{i,j},Ts:Ts:Nrp(i)*Ts);
    end
    for j = 1:N_perturbacoes
        gp{i,j} = step(Gp{i,j},0:Ts:(Nrp(i)-1)*Ts);
    end
end

Qy = diag(repelem(pondY,1,N)); %montagem da matriz de ponderao do erro
Qu = diag(repelem(pondU,1,Nu)); %montagem da matriz de ponderao da ao de controle

 
%% PARAMETROS DA SIMULAO
Tini = 3;
Tsim = Tsim + Tini;
for i = 1:N_saidas
    ref{i} = repelem(degraus_ref{i},1,[Tini N(i)+N1(i)]+degraus_t{i});%vetor de referncias - [valores dos degraus] - [repetio do respectivo valor]  
end 
for i = 1:N_perturbacoes
    pert{i} = repelem(degraus_pert{i},1,[Tini N(i)+N1(i)]+degraus_pert_t);%vetor de perturbaes - [valores dos degraus] - [repetio do respectivo valor]
end

 %% RODA SIMULAO
 %inicializa variaveis
for i = 1:N_saidas
     y{i}(1) = 0;
     fma{i} = zeros(Nrp(i),1);
     ysim{i} = zeros(Nrp(i),1);
end
for i = 1:N_entradas
    for j = 1:Tini
       du{i}(j) = 0;
       u{i}(j) = 0;
    end
end

%% Monta as matrizes offline que definem o problema MIMO DMC

for k=1:N_saidas 
    for j=1:N_entradas
        for i=1:Nu(j)
            Gm{k,j}(i:N(k),i) = g{k,j}(N1(k):N2(k)-i+1);
        end
    end
end
G = cell2mat(Gm);
H = 2*(G'*Qy*G+Qu);


%%%%Simulao
for k=Tini:Tsim
    %%%Roda a planta
    for i = 1:N_saidas
        for j = 1:N_entradas
            ysim{i} = ysim{i} + g{i,j}*du{j}(k-1);
        end
        for j = 1:N_perturbacoes
            ysim{i} = ysim{i} + gp{i,j}*(pert{j}(k)-pert{j}(k-1));
        end
        y{i}(k) = ysim{i}(1);
        ysim{i}(1:end-1) = ysim{i}(2:end);
    end
    
    %%%Roda o MPC
    for i = 1:N_saidas
        for j = 1:N_entradas
            fma{i} = fma{i} + g{i,j}*du{j}(k-1);
        end
        if usar_feed_forward
            for j = 1:N_perturbacoes
                fma{i} = fma{i} + gp{i,j}*(pert{j}(k)-pert{j}(k-1));
            end
        end
        e = y{i}(k)-fma{i}(1);
        fma{i} = [fma{i}(2:Nrp(i)); fma{i}(Nrp(i))];
        f_aux{i,1} = fma{i}(N1(i):N2(i))+ ones(N(i),1)*e;
    end

    f = cell2mat(f_aux);
    pos = 1;
    for i = 1:N_saidas
        refm(pos:pos+N(i)-1) = ref{i}(k+N1(i):k+N1(i)+N(i)-1);
        pos = pos+N(i);
    end
    b = (2*(f-refm')'*Qy*G)';

    if flag_restricao_u
        r_p=[];
        r_m= [];
        for i = 1:N_entradas
            r_p = [r_p; repelem(u_lim(i)-u{i}(k-1),1,Nu(i))'];
            r_m = [r_m; repelem(-u_lim(N_entradas+i)+u{i}(k-1),1,Nu(i))'];
        end
        r_con = [r_p; r_m];
        r = [r_inc; r_con]
    else
        r = r_inc;
    end


    %% Resolve o problema de otimizao
    tic;
    [delta_u] = quadprog(H,b,R,r);
    t_solver = toc;

    %% Atualiza as variveis de controle
    pos = 1;
    for i = 1:N_entradas
        du{i}(k) = delta_u(pos);
        u{i}(k) = u{i}(k-1) + du{i}(k);
        pos = pos+Nu(i);
    end
end


%%%Gera grficos
t = Ts*(Tini:Tsim);
tamletra=20;
tamnum=20;
figure
plot(t,ref{1}(Tini:Tsim),'--',t,y{1}(Tini:Tsim),'-',t,ref{2}(Tini:Tsim),'--',t,y{2}(Tini:Tsim),'-',t,ref{3}(Tini:Tsim),'--',t,y{3}(Tini:Tsim),'-','LineWidth',2)
legend('ref_1','Y_1','ref_2','Y_2','ref_3','Y_3','Location','SouthEast')
axis([Tini (Tsim-Tini)*Ts -0.3 1.2]);
 %   ylabel('Sadas e Referncias do Sistemas', 'FontSize', tamletra);
xlabel('Tempo (s)', 'FontSize', tamletra);
ylabel('Controladas', 'FontSize', tamletra);
set(gca,'FontSize',tamnum);
grid on

print -depsc results_con_ex_55_3_1.eps

figure
plot(t,u{1}(Tini:Tsim),'-',t,u{2}(Tini:Tsim),t,u{3}(Tini:Tsim),'-','LineWidth',2)
%title(['y1' ' e Referncia do Sistema'], 'FontSize', tamletra)
legend('U_1','U_2','U_3','Location','SouthEast')
axis([Tini (Tsim-Tini)*Ts -1.3 1.5]);
 %   ylabel('Sadas e Referncias do Sistemas', 'FontSize', tamletra);
xlabel('Tempo (s)', 'FontSize', tamletra);
ylabel('Manipuladas', 'FontSize', tamletra);
set(gca,'FontSize',tamnum);
grid on
print -depsc results_man_ex_55_3_1.eps
