clear all
close all
clc
addpath('../../Bibliotecas/')
run('../../Bibliotecas/parametrosFiguras.m')
%% CSTR instável
global Ac dH p cp ER k0 V
%%% constantes
Ac = 0.05;
dH = -50e6;
p = 1e6;
cp = 1;
ER = 1000;
k0 = 4.872997584;
V = 0.05;

Ts = 3;
%%% pontos de operação
%%%% entradas
%%%%% manipuladas
qo0 = 5e-3;
Caf0 = 5;
Qh0 = 0.75;

%%%%% perturbações
qi0 = 5e-3;
Ti0 = 350;

%%%% saídas %% PO do artigo, estão incorretos
h0 = 1;
Ca0 = 1;
T0 = 400;

%%% teste ponto de operação
x0 = [h0;Ca0;T0];
u0 = [qo0;Caf0;Qh0];
q0 = [qi0;Ti0];

fR1 = @(T) k0*exp(-ER/T);
fcstrpo= @(y) [(q0(1)-u0(1))/Ac;
                q0(1)*(u0(2)-y(2))/V-y(2)*fR1(y(3));
                q0(1)*(q0(2)-y(3))/V-(dH/p/cp)*fR1(y(3))*y(2)-u0(3)/V];

x0f = fsolve(fcstrpo,x0);
[x0,x0f]
            
            
            
%%
%%% linearização no ponto de operação
%%% ordem dos estados [Ca,T,Tc]
As = [0,0,0;
      0,-qi0/V-k0*exp(-ER/T0),-Ca0*ER/T0^2*k0*exp(-ER/T0);
      0,-dH/p/cp*k0*exp(-ER/T0),-dH/p/cp*Ca0*ER/T0^2*k0*exp(-ER/T0)];
eig(As)
%%% ordem das entradas [F,Te,Caf,Fc,Tce]
Bs = [-1/Ac,0,0,1/Ac,0;
      0,qi0/V,0,(Caf0-Ca0)/V,0;
      0,0,-1/V,(Ti0-T0)/V,qi0/V];
 
Gss = ss(As,Bs,eye(3),[]); 
G = minreal(tf(Gss(:,1:3)));
Gq = minreal(tf(Gss(:,4:5)));

Gssz = c2d(Gss,Ts,'zoh');
%%% teste
xs0 = modeloCSTR(x0,u0,q0,2000)
[xs0,x0,x0f]

%%% steps
xs0 = modeloCSTR(x0,[u0(1);u0(2);u0(3)*1.1],[q0(1);q0(2)],5000)
[xs0]


%%

m = 3; % número de entradas da planta - manipuladas
n = 3; % número de saidas da planta - controladas
mq = 0; % número de perturbações da planta

%%% discretização do processo
Gz = c2d(G,Ts,'zoh');

%% PARAMETROS DE AJUSTE DO CONTROLADOR

Nu = [5 5 5]; %horizontes de controle - dimensão 1 x m
N1 = [1 1 1]; % inicio dos horizontes de predição - incluir atraso - dimensão 1 x n
N2 = [30 30 30]; % fim dos horizontes de predição - dimensão 1 x n
Ny = N2-N1+1; %horizonte de predição - não editar

Nss = [100 100 100];
Nf = [30 30 30];

delta = [20 2.5 1/400]; %ponderação nos erros - dimensão 1 x n
lambda = 1.005*[1 1 1]; %ponderação nas ações de controle - dimensão 1 x m       
betaf = [0.8 0.8 0.8]; %% polos dos filtros dos erros de predição por saída

%% definição do cenário de simulação
nit =floor(300/Ts); %tempo de simulação
nin = max(max(Nf),max(Nss))+10; %número da iteração inicial
nit = nit+nin;


refs = repmat(x0,[1,nit]);
perts = repmat(q0,[1,nit]);

refs(1,nin+floor(5/Ts):end) = x0(1)*1.05;
% refs(2,nin+floor(5/Ts):end) = x0(2)*1.05;
% refs(2,nin+floor(60/Ts):end) = h0(2)*0.8;
% 
% refs(1,nin+floor(120/Ts):end) = [8.7664];
% refs(2,nin+floor(120/Ts):end) = [9.3579];
% refs(3,nin+floor(120/Ts):end) = [1.2273];
% refs(4,nin+floor(120/Ts):end) = [0.9785];

% perts(1,nin+floor(100/Ts):end) = q0(1)*1.1;
% perts(2,nin+floor(200/Ts):end) = q0(2)*0.95;


saidas = repmat(x0,[1,nit]); % vetor da saída
entradas = repmat(u0,[1,nit]); % vetor do sinal de controle
du = zeros(m,nit); % vetor dos incrementos de controle



%% Controlador GDMC - MIMO
modDegrauU = {};
%%%%%%%%%
for i = 1:n
    for j = 1:m
        modDegrauU{i,j} = step(Gz(i,j),Ts:Ts:round(1.5*Nss(i))*Ts);
    end
end

Qy = diag(repelem(delta,1,Ny)); %montagem da matriz de ponderação do erro
Qu = diag(repelem(lambda,1,Nu)); %montagem da matriz de ponderação da ação de controle

 

%% Monta as matrizes offline que definem o problema MIMO DMC
for k=1:n 
    for j=1:m
        for i=1:Nu(j)
            Gm{k,j}(i:Ny(k),i) = modDegrauU{k,j}(N1(k):N2(k)-i+1);
        end
    end
end
G = cell2mat(Gm);

Kdmc = (G'*Qy*G+Qu)\G'*Qy; % utilizado apenas no caso irrestrito
Kdmc1 = [];
for i=1:m
    Kdmc1 = [Kdmc1;Kdmc(sum(Nu(1:i-1))+1,:)];
end

%%% matrizes do problema de otimização
Hoti = 2*(G'*Qy*G+Qu);
foti1 = -2*G'*Qy;

Aoti = [];
for i=1:m
    Aoti = blkdiag(Aoti,tril(ones(Nu(i))));
end
Aoti = [Aoti;-Aoti];

%% Cálculo do filtro do erro de predição
% formato do filtro (az^2+bz)/(z-betaf)^2

z = tf('z',Ts);
nf = 2;



%%% para a saída 1 é especial por conta de ser integrador.
for l=1:1
    F{l} = tf(0,1,Ts);
    
    for i=N1(l):N2(l)
        %%% monta sistema para obtenção dos parâmetros do filtro
        %%% primeira equação (z^i - F_i(z)) = 0 -> para z=pz
        %%% segunda equação F_i(1) = 0 -> força ganho unitário
        indf = i-N1(1)+1;
        Af = [1 1;
              -2*betaf(l),(-1-betaf(l))];
        bf = [(1-betaf(l))^2;
              i*(1-betaf(l))^3];
        X = Af\bf;
        F{l}(indf,1) = (X(1)*z^2+X(2)*z)/(z-betaf(l))^2;
        %%% teste da condição
%         zero(z^i - F{l}(indf,1))

        %%% armazena coeficientes gtil
        for k=1:m
            modDegrauUF{l,k,i} = filter(F{l}(indf,1).num{1},F{l}(indf,1).den{1},modDegrauU{l,k});    
        end
    end
end



%%% para as duas outras saídas basta remover o polo instavel
polos = unique(eig(expm(As*Ts)));
I = find(abs(polos)>1);
pz = polos(I); % obtem o polo indesejado (instável) de malha aberta
% pz = 1.1

for l=2:n
    F{l} = tf(0,1,Ts);
    
    for i=N1(l):N2(l)
        %%% monta sistema para obtenção dos parâmetros do filtro
        %%% primeira equação (z^i - F_i(z)) = 0 -> para z=pz
        %%% segunda equação F_i(1) = 0 -> força ganho unitário
        indf = i-N1(1)+1;
        Af = [1 1;
              pz^2, pz];
        bf = [(1-betaf(l))^2;
              pz^i*(pz-betaf(l))^2];
        X = Af\bf;
        F{l}(indf,1) = (X(1)*z^2+X(2)*z)/(z-betaf(l))^2;
        %%% teste da condição
%         zero(z^i - F{l}(indf,1))

        %%% armazena coeficientes gtil
        for k=1:m
            modDegrauUF{l,k,i} = filter(F{l}(indf,1).num{1},F{l}(indf,1).den{1},modDegrauU{l,k});    
        end
    end
end

%%% armazena coeficientes gtil
for l=1:n
    for k=1:m
        %%% calcula a matriz H para o cálculo da resposta livre no caso GDMC
        H1 = [];
        H2 = [];

        for j=N1(1):N2(1)
            H1 = [H1;modDegrauU{l,k}(j+1:j+Nf(l))'];
            H2 = [H2;modDegrauUF{l,k,j}(1:Nf(l))'];

        end
        H{l,k} = H1-H2;

    end
end


%% inicializacao dos estados e variaveis da simulação
for l=1:n
    yfilt{l} = zeros(Ny(l),nit);
end

for k = nin:nit
    %% -- simulador, não alterar
    %%% Simulador MIMO que utiliza estados
    saidas(:,k) = modeloCSTR(saidas(:,k-1),entradas(:,k-1),perts(:,k-1),Ts);
    
    %% -- Controlador GDMC 
    %%% referencias
    R = [];
    for i=1:n
        R = [R;
             repelem(refs(i,k),Ny(i),1)];
    end 
    
    %%% calculo da resposta livre
    f = [];
    for l=1:n
        for i=1:Ny(1)
            yfilt{l}(i,k) = -F{l}(i,1).den{1}(2:end)*yfilt{l}(i,k-1:-1:k-nf)' + F{l}(i,1).num{1}*saidas(l,k:-1:k-nf)';
        end
        
        ftemp = 0;
        for ll=1:m
            ftemp = ftemp + H{l,ll}*du(ll,k-1:-1:k-Nf(ll))';
        end
        f = [f; ftemp + yfilt{l}(:,k)];
    end
        
    
    %%% montagem das matrizes de restrição
%     boti = repelem(umax'-entradas(:,k-1),Nu');
%     boti = [boti;
%             repelem(-umin'+entradas(:,k-1),Nu')];
%     
%     foti = foti1*(R-f);
%     X = quadprog(Hoti,foti,Aoti,boti);
%     for i=1:m
%         du(i,k) = X(sum(Nu(1:i-1))+1,1);
%     end
    %% Resolve o problema de otimização
    du(:,k) = Kdmc1*(R-f);
    entradas(:,k) = entradas(:,k-1)+du(:,k);

end

%%
figure
subplot(3,2,1)
plot(saidas(1,:))
hold on
plot(refs(1,:),':')

subplot(3,2,3)
plot(saidas(2,:))
hold on
plot(refs(2,:),':')

subplot(3,2,5)
plot(saidas(3,:))
hold on
plot(refs(3,:),':')


subplot(3,2,2)
plot(entradas(1,:))

subplot(3,2,4)
plot(entradas(2,:))

subplot(3,2,6)
plot(entradas(3,:))

% 
% plotyy(1:nit,saidas(1:2,:)',1:nit,saidas(3,:))
% hold on
% hold on
% plot(refs',':')
% subplot(2,1,2)
% plot(entradas')





%% modelo do sistema 4 tanques
%%% x = [h,Ca,T]
%%% u = [qo,Caf,Qh]
%%% q = [qi,Ti]
function h = modeloCSTR(x,u,q,Ts)
    
    global Ac dH p cp ER k0 V
    
    opts = odeset('NonNegative',[1 2 3]);
    fR1 = @(T) k0*exp(-ER/T);
    
    fcstr= @(t,y) [(q(1)-u(1))/Ac;
                    q(1)*(u(2)-y(2))/V-y(2)*fR1(y(3));
                    q(1)*(q(2)-y(3))/V-(dH/p/cp)*fR1(y(3))*y(2)-u(3)/V];
                       
    [t,xf] = ode45(fcstr,[0 Ts],x,opts);
    h = xf(end,:)';
end
