clear all
close all
clc
addpath('../../Bibliotecas/')
run('../../Bibliotecas/parametrosFiguras.m')
%% CSTR instável

%%% constantes
global E k0 dH V A U Vj cp cpc p pc R Ts
E = 92973;
k0 = 4.25e12;
dH = 6e4;
V = 18.925;
A = 37.16
U = 68;
Vj = 5;
cp = 3.14;
cpc = 4.18;
p = 900;
pc = 1000;
R = 8.314;

Ts = .5; % em minutos
Ts = 5;

%%% pontos de operação
%%%% entradas
%%%%% manipuladas
F0 = 0.21; % do artigo
% F0 = 0.1486;
Caf0 = 10;
Fc0 = 3.21;

%%%%% perturbações
Te0 = 294;
Tce0 = 294;

%%%% saídas %% PO do artigo, estão incorretos
Ca0 = 2.678;
% Ca0 = 3.2
T0 = 320.33;
Tc0 = 298;

%%% teste ponto de operação
% Ts = 100
fR1 = @(Ca,T) k0*Ca*(10-Ca)*exp(-E/R/T);
u = [F0;Caf0;Fc0];
q = [Te0;Tce0];
fcstr= @(x) [u(1)/V*(u(2)-x(1))-fR1(x(1),x(2));
               u(1)/V*(q(1)-x(2))+dH/p/cp*fR1(x(1),x(2))-(U*A/V/p/cp)*(x(2)-x(3));
               u(3)/Vj*(q(2)-x(3))+(U*A/Vj/pc/cpc)*(x(2)-x(3))];
           
x0 = fsolve(fcstr,[Ca0;T0;Tc0]);
[x0,[Ca0;T0;Tc0]]
% Ca0 = x0(1);
% T0 = x0(2);
% Tc0 = x0(3);

%%% limites máximos
umin = [0,8,0];
umax = [.5,12,20];



%%
%%% linearização no ponto de operação
%%% ordem dos estados [Ca,T,Tc]
As = [-F0/V-k0*(10-2*Ca0)*exp(-E/R/T0),-(E/R/T0^2)*k0*Ca0*(10-Ca0)*exp(-E/R/T0),0;
     (dH/p/cp)*k0*(10-2*Ca0)*exp(-E/R/T0),-F0/V-(U*A/V/p/cp)+(dH/p/cp)*(E/R/T0^2)*k0*Ca0*(10-Ca0)*exp(-E/R/T0),(U*A/V/p/cp),
     0,(U*A/Vj/pc/cpc),-Fc0/Vj-(U*A/Vj/pc/cpc)];
eig(As)
%%% ordem das entradas [F,Te,Caf,Fc,Tce]
Bs = [(Caf0-Ca0)/V,0,F0/V,0,0;
     (Te0-T0)/V,F0/V,0,0,0;
     0,0,0,(Tce0-Tc0)/Vj,Fc0/Vj];
 
Gss = ss(As,Bs,eye(3),[]); 
G = tf(Gss(:,[1,3,4]));
Gq = tf(Gss(:,[2,5]));

Gssz = c2d(Gss,Ts,'zoh');
%%% teste
x0 = modeloCSTR([Ca0;T0;Tc0],[F0;Caf0;Fc0],[Te0;Tce0])
[x0,[Ca0;T0;Tc0]]

%%

m = 3; % número de entradas da planta - manipuladas
n = 3; % número de saidas da planta - controladas
mq = 0; % número de perturbações da planta

%%% discretização do processo
Gz = c2d(G,Ts,'zoh');

%% PARAMETROS DE AJUSTE DO CONTROLADOR

Nu = [5 5 5]; %horizontes de controle - dimensão 1 x m
N1 = [1 1 1]; % inicio dos horizontes de predição - incluir atraso - dimensão 1 x n
N2 = [50 50 50]; % fim dos horizontes de predição - dimensão 1 x n
Ny = N2-N1+1; %horizonte de predição - não editar

Nss = [500 500 500];
Nf = [230 230 230];

delta = [1 1 1]./Ny; %ponderação nos erros - dimensão 1 x n
lambda = [1 1 1]./Nu; %ponderação nas ações de controle - dimensão 1 x m       
betaf = [0.8 0.8 0.8]; %% polos dos filtros dos erros de predição por saída

%% definição do cenário de simulação
nit =floor(300/Ts); %tempo de simulação
nin = 20; %número da iteração inicial
nit = nit+nin;


refs = repmat([Ca0;T0;Tc0],[1,nit]);
perts = repmat([Te0;Tce0],nit);

% refs(1,nin+floor(5/Ts):end) = h0(1)*1.2;
% refs(2,nin+floor(60/Ts):end) = h0(2)*0.8;
% 
% refs(1,nin+floor(120/Ts):end) = [8.7664];
% refs(2,nin+floor(120/Ts):end) = [9.3579];
% refs(3,nin+floor(120/Ts):end) = [1.2273];
% refs(4,nin+floor(120/Ts):end) = [0.9785];

% perts(1,nin+130:end) = 0.2;

saidas = repmat([Ca0;T0;Tc0],[1,nit]); % vetor da saída
entradas = repmat([F0;Caf0;Fc0],[1,nit]); % vetor do sinal de controle
du = zeros(m,nit); % vetor dos incrementos de controle



%% Controlador GDMC - MIMO
modDegrauU = {};
%%%%%%%%%
for i = 1:n
    for j = 1:m
        modDegrauU{i,j} = step(Gz(i,j),Ts:Ts:2*Nss(i)*Ts);
    end
end

Qy = diag(repelem(delta,1,Ny)); %montagem da matriz de ponderação do erro
Qu = diag(repelem(lambda,1,Nu)); %montagem da matriz de ponderação da ação de controle

 

%% Monta as matrizes offline que definem o problema MIMO DMC
for k=1:n 
    for j=1:m
        for i=1:Nu(j)
            Gm{k,j}(i:Ny(k),i) = modDegrauU{k,j}(N1(k):N2(k)-i+1);
        end
    end
end
G = cell2mat(Gm);


%%% matrizes do problema de otimização
Hoti = 2*(G'*Qy*G+Qu);
foti1 = -2*G'*Qy;

Aoti = [];
for i=1:m
    Aoti = blkdiag(Aoti,tril(ones(Nu(i))));
end
Aoti = [Aoti;-Aoti];

%% Cálculo do filtro do erro de predição
% formato do filtro (az^2+bz)/(z-betaf)^2

z = tf('z',Ts);
nf = 2;

polos = unique(eig(expm(As*Ts)));
I = find(abs(polos)>1);
pz = polos(I); % obtem o polo indesejado (instável) de malha aberta
% pz = 1.1

for l=1:n
    F{l} = tf(0,1,Ts);
    
    for i=N1(l):N2(l)
        %%% monta sistema para obtenção dos parâmetros do filtro
        %%% primeira equação (z^i - F_i(z)) = 0 -> para z=pz
        %%% segunda equação F_i(1) = 0 -> força ganho unitário
        indf = i-N1(1)+1;
        Af = [1 1;
              pz^2, pz];
        bf = [(1-betaf(l))^2;
              pz^i*(pz-betaf(l))^2];
        X = Af\bf;
        F{l}(indf,1) = (X(1)*z^2+X(2)*z)/(z-betaf(l))^2;
        %%% teste da condição
%         zero(z^i - F{l}(indf,1))

        %%% armazena coeficientes gtil
        for k=1:m
            modDegrauUF{l,k,i} = filter(F{l}(indf,1).num{1},F{l}(indf,1).den{1},modDegrauU{l,k});    
        end
    end
end

%%% armazena coeficientes gtil
for l=1:n
    for k=1:m
        %%% calcula a matriz H para o cálculo da resposta livre no caso GDMC
        H1 = [];
        H2 = [];

        for j=N1(1):N2(1)
            H1 = [H1;modDegrauU{l,k}(j+1:j+Nf(l))'];
            H2 = [H2;modDegrauUF{l,k,j}(1:Nf(l))'];

        end
        H{l,k} = H1-H2;

    end
end


%% inicializacao dos estados e variaveis da simulação
% saidas = repmat(h0,[1,nit]); % vetor da saída
entradas = repmat(v0,[1,nit]); % vetor do sinal de controle
du = zeros(m,nit); % vetor dos incrementos de controle

for k = nin:nit
    %% -- simulador, não alterar
    %%% Simulador MIMO que utiliza estados
    saidas(:,k) = modelo4tanques(saidas(:,k-1),entradas(:,k-1));
    
    %% -- Controlador GPC 
    %%% referencias
    R = [];
    for i=1:n
        R = [R;
             repelem(refs(i,k),Ny(i),1)];
    end 
    
    %%% calculo da resposta livre
    du1 = [];
    for i=1:m
        du1 = [du1;du(i,k-1:-1:k-nH(i))'];
    end
    
    y1 = [];
    for i=1:n
        y1 = [y1;saidas(i,k:-1:k-nF(i)+1)'];
    end
    
    f = F*y1 + H*du1;
    
    %%% montagem das matrizes de restrição
    boti = repelem(umax'-entradas(:,k-1),Nu');
    boti = [boti;
            repelem(-umin'+entradas(:,k-1),Nu')];
    
    foti = foti1*(R-f);
    X = quadprog(Hoti,foti,Aoti,boti);
    for i=1:m
        du(i,k) = X(sum(Nu(1:i-1))+1,1);
    end
    %% Resolve o problema de otimização
%     du(:,k) = Kgpc1*(R-f);
    entradas(:,k) = entradas(:,k-1)+du(:,k);

end

figure
subplot(2,1,1)
plot(saidas')
hold on
plot(refs',':')
subplot(2,1,2)
plot(entradas')





%% modelo do sistema 4 tanques
%%% x = [Ca, T,Tc]
%%% u = [F, Caf, Fc]
%%% q = [Te,Tce]
function h = modeloCSTR(x,u,q)
    
    global E k0 dH V A U Vj cp cpc p pc R Ts
    
    opts = odeset('NonNegative',[1 2 3]);
    fR1 = @(Ca,T) k0*Ca*(10-Ca)*exp(-E/R/T);
    
    fcstr= @(t,y) [u(1)/V*(u(2)-y(1))-fR1(y(1),y(2));
                   u(1)/V*(q(1)-y(2))+dH/p/cp*fR1(y(1),y(2))-(U*A/V/p/cp)*(y(2)-y(3));
                   u(3)/Vj*(q(2)-y(3))+(U*A/Vj/pc/cpc)*(y(2)-y(3))];
                       
    [t,x] = ode45(fcstr,[0 Ts],x,opts);
    h = x(end,:)';
end
