%%% simulação GPC e DMC para o caso sem restrições, sem filtro de
%%% referência e com ação antecipativa
clear all
close all
clc



%% Sintonia do GPC e DMC

N1 = 1; %% horizonte de predição inicial
N2 = 100; %% horizonte predição
Ny = N2-N1+1;
Nu = 30; %% horizonte de controle
Nq = 0; %% horizonte de perturbação

lambda = 1; % ponderação do incremento de controle
delta = 1;  % ponderação dos erros futuros


%% Modelo FT
z = tf('z',-1)
Gz = 0.1972/(z-0.9868)
Gzq = -0.1403/(z-0.9838)


A = conv(Gz.den{1},Gzq.den{1})
B = Gz.num{1}(2)*Gzq.den{1}
Bq = [0 Gzq.num{1}(2)*Gz.den{1}]

d=0;
dq = 0;

nb = size(B,2)-1; % ordem do polinômio B
na = size(A,2)-1; % ordem do polinômio A
nbq = size(Bq,2)-1; % ordem do polinômio Bq

%% montagem das matrizes
Btil = conv(B,[zeros(1,d) 1]); % incorporação do atraso no numerador B
Bqtil = conv(Bq,[zeros(1,dq) 1]); % incorporação do atraso no numerador Bq


G = zeros(Ny,Ny); % matriz dinâmica G
H = zeros(Ny,nb+d);
Gq = zeros(Ny,Ny-1); 
Hq = zeros(Ny,nbq+dq);

[E,F] = diofantina(conv(A,[1 -1]),N1,N2);

for i=N1:N2
    EjB = conv(E(i-N1+1,1:i),Btil);
    EjBq = conv(E(i-N1+1,1:i),Bq);
    
    G(i-N1+1,1:i) = EjB(i:-1:1);    
    H(i-N1+1,:) = EjB(i+1:end);

    Gq(i-N1+1,1:i) = EjBq(i:-1:1);    
    Hq(i-N1+1,:) = EjBq(i+1:end);
    
end
G = G(:,1:Nu);
Gq = Gq(:,1:Nq);


G,F,H,Gq,Hq

%% obtenção do ganho irrestrito
Qu = lambda*eye(Nu); % matriz de ponderação dos incrementos de controle
Qe = delta*eye(Ny);  % matriz de ponderaão dos erros futuros

Kgpc = (G'*Qe*G+Qu)\G'*Qe;

Kgpc1 = Kgpc(1,:);

%% vetores de simulação
nin = 5; % iteração inicial (para inicialização correta)
nit = nin+200; % número de iterações da simulação

refs = h0+zeros(1,nit+Ny); % vetor das referências
perts = as0+zeros(1,nit+Ny); % vetor da perturbação

refs(nin+10:end) = h0*1.1;
perts(1,nin+100:end) = as0*1.3;


saidasGPC = h0+zeros(1,nit); % vetor da saída
entradasGPC = ae0+zeros(1,nit); % vetor do sinal de controle
duGPC = zeros(1,nit); % vetor dos incrementos de controle


for i=nin+1:nit
    %%% simulação do processo
    saidasGPC(1,i) = simModelo(entradasGPC(1,i-1),perts(1,i-1),saidasGPC(1,i-1));
              
    %%% vetor de referências
    R = refs(1,i)*ones(N2-N1+1,1);
    
    %%% cálculo da resposta livre;
    f = F*saidasGPC(1,i:-1:i-na)';
    
    %%% adição da parcela dos incrementos de controle passados na
    %%% resposta livre
    if(~isempty(H))
        f = f+ H*duGPC(1,i-1:-1:i-nb-d)';
    end
    
    %%% adição da parcela do incremento da perturbação passada na
    %%% resposta livre
    deltaQ = perts(1,i:-1:i-nbq-dq+1)'-perts(1,i-1:-1:i-nbq-dq)';
    if(~isempty(Hq))
        f = f+ Hq*deltaQ;
    end
    
    %%% cálculo do incremento de controle ótimo
    duGPC(1,i) = Kgpc1*(R-f);
    
    %%% cálculo do sinal de controle ótimo
    entradasGPC(1,i) = duGPC(1,i)+entradasGPC(1,i-1);
    
              
end


%%
% %% Geração das figuras
% tamletra=14;
% tamnum=14;
% 
% figure
% h=subplot(3,1,1)
% plot(saidasGPC(1,nin+1:nit),'LineWidth',2)
% hold on
% plot(saidasDMC(1,nin+1:nit),'LineWidth',2)
% plot(refs(1,nin+1:nit),'--','LineWidth',2)
% % ylim([0 2])
% hl = legend('GPC','DMC','Referência','Location','NorthEast')
% hl.Position = [0.6785 0.7018 0.2368 0.1136];
% ylabel('Saída','FontSize', tamletra)
% set(h, 'FontSize', tamnum);
% grid on
% 
% h = subplot(3,1,2)
% plot(entradasGPC(1,nin+1:nit),'LineWidth',2)
% hold on
% plot(entradasDMC(1,nin+1:nit),'LineWidth',2)
% 
% ylabel('Entrada','FontSize', tamletra)
% grid on
% set(h, 'FontSize', tamnum);
% 
% h = subplot(3,1,3)
% plot(duGPC(1,nin+1:nit),'LineWidth',2)
% hold on
% plot(duDMC(1,nin+1:nit),'LineWidth',2)
% 
% ylabel('\Delta u','FontSize', tamletra)
% xlabel('Tempo (amostras)','FontSize', tamletra)
% grid on
% set(h, 'FontSize', tamnum);

% h=subplot(3,1,3)
% plot(t,erro(1:nit-N2-1),'LineWidth',2)
% title('Erro', 'FontSize', tamletra);
% xlabel('tempo', 'FontSize', tamletra);
% set(h, 'FontSize', tamnum);

% print('simDMC_GPC_ff','-depsc')





