%%% Simulação comparativa do PNMPC, NEPSAC e xxx
clear all, close all, clc


%% parametros CSTR completo
global ER V k0 k1 k2 k3 Ts
%%% condicoes iniciais
Ca0 = 0.1 ; %[mol/l]
T0 = 438.5444; %[K]

f0 = 100; %[l/min] process flow rate
fc0 = 103.41; % [l/min] coolant flow rate

CAf0 = 1; %[mol/l] feed concentration
Tcf0 = 350.0030; %[K] inlet coolant temperature
Tf0 = 350; %[K] feed temperature

Ts = 0.1; % [min]

%%% parametros
ha = 7e5; %[cal/(min K)] heat transfer term
ER = 1e4; %[K] activation energy term
rho = 1e3; %[g/l] liquid densities rho=rho_c
rhoc = rho;
V = 100; %[l] CSTR volume
k0 = 7.2e10; %[1/min] reaction rate constant
dH = -2e5; %[cal/mol] heat of reaction
Cp = 1; %[cal/(g K)] specific heats
Cpc = 1;

k1 = -dH*k0/(rho*Cp);
k2 = rhoc*Cpc/(rho*Cp*V);
k3 = ha/(rhoc*Cpc);


%% restrições
umax = [150, 110];
umin = [95, 60];


%% definicao das variaveis de simulacao
%%% Modelo CSTR listado em DU et al 
n = 2; % numero de saidas
m = 2; % numero de entradas
q = 3; % numero de perturbacoes

na = 1; % numero de amostras de y atrasadas usadas no calculo da saida
nb = 1; % numero de amostras de u atrasadas usadas no calculo da saida
nq = 1; % numero de amostras de q atrasadas usadas no calculo da saida


%% configuracaoes de simulacao
nin = 2; %%% iteracao inicial
nit = nin+500; %%% iteracao final

N1 = [1,1]
Ny = [100, 100]; 
Nu = [30, 30]; % vetor contendo os horizontes de controle
Qy = [10000, 1]; % vetor contendo as ponderacoes de saida
Qu = [((1e-03)^2)*30, (4.3^2)*30]; % vetor contendo as ponderacoes de controle
Qu = [3, 3];

C = 1; % polinômio C do modelo da perturbação
D = [1 -1]; % polinômio D do modelo da perturbação

nc = size(C,2);
nd = size(D,2);


%% inicializacao de algumas variaveis
saida = zeros(n,nit);
saida(1,1:nin) = Ca0;
saida(2,1:nin) = T0;

entrada = zeros(m,nit);
entrada(1,:) = f0;
entrada(2,:) = fc0;

refs = zeros(n,nit+max(Ny));
refs(1,:) = Ca0;
refs(2,:) = T0;
refs(1,nin+10:end) = 0.05;
refs(2,nin+150:end) = T0+5;
refs(1,nin+250:end) = Ca0;
refs(2,nin+250:end) = T0;


perts = zeros(q,nit);
perts(1,:) = CAf0;
perts(2,:) = Tf0;
perts(3,:) = Tcf0;

% perts(1,nin+50:nit) = CAf0*1.1;
% perts(2,nin+200:nit) = T0+5;
% perts(3,nin+400:nit) = Tcf0*0.85;


eta = zeros(n,nit+max(N2));
e = zeros(n,nit+max(N2));
saidaModelo = zeros(n,nit);
du = zeros(m,nit);


%% simulação 

for k=nin:nit
    %% modelo do CSTR
    saida(:,k) = modeloCSTR(saida(1,k-1),saida(2,k-1),entrada(1,k-1),entrada(2,k-1),perts(1,k-1),perts(2,k-1),perts(3,k-1));

    %% controlador PNMPC
    %%% cálculo da resposta do modelo nominal
    saidaModelo(k) = modeloCSTRnominal(saida(k-1),entrada(k-1),q(k-1));
    
    %%% calculo dos etas futuros
    eta(:,k) = saida(:,k) - saidaModelo(:,k); % cálculo do eta atual
    
    e(k) = -C(2:end)*e(k-1:-1:k-nc+1)'+ D*eta(k:-1:k-nd+1)';
    
    for j=1:max(N2)
        eta(:,k+j) = -D(2:end)*eta(:,k+j-1:-1:k+j-nd+1)' + C*e(:,k+j:-1:k+j-nc+1)';
    end
    
    
    %%% cálculo da resposta livre corrigida
    N2max = max(N2);
    yc = zeros(n,N2max);
    yc(:,1) = modeloCSTRnominal(saida(1,k),saida(2,k),entrada(1,k-1),entrada(2,k-1),Caf0,Tf0,Tcf0) + eta(:,k+1);
    for i=2:N2max
        yc(:,i) = modeloCSTRnominal(yc(1,i-1),yc(2,i-1),entrada(1,k-1),entrada(2,k-1),Caf0,Tf0,Tcf0) + eta(:,k+i);
    end
    
    %%% passando para o formato vetorial correto
    ycv = zeros(sum(N2-N1)+1,1);
    Nyi = N2-N1;
    for i=1:n
        ycv(sum(Nyi(1:i-1))+1:sum(Ny(1:i)),1) = yc(i,N1(i):N2(i))';
    end

    
    %%% calculo de G
    % calculo de y0
    y0 = zeros(n,N2max);
    y0(:,1) = modeloCSTRnominal(saida(1,k),saida(2,k),entrada(1,k-1),entrada(2,k-1),Caf0,Tf0,Tcf0);
    for i=2:N2max
        y0(:,i) = modeloCSTRnominal(yo(1,i-1),yo(2,i-1),entrada(1,k-1),entrada(2,k-1),Caf0,Tf0,Tcf0);
    end
    
    y0v = zeros(sum(N2-N1)+1,1);
    Nyi = N2-N1;
    for i=1:n
        y0v(sum(Nyi(1:i-1))+1:sum(Ny(1:i)),1) = y0(i,N1(i):N2(i))';
    end

    %%% calculo de yi para obter G
    G = zeros(sum(Nyi),sum(Nu));
    for i=1:m
        
        if(entrada(i,k-1) == 0)
            epsilon = 0.001;
        else
            epsilon = entrada(i,k-1)/1000;
        end

        y = zeros(N2,Nu);
        for j=1:Nu
            u = entrada(k-1)*ones(N2,1);
            u(j:end) = u(j:end)+epsilon;

            y(1,j) = modeloTanque(saida(k),u(1),q(k-1));
            for i=2:N2
                y(i,j) = modeloTanque(y(i-1,j),u(i),q(k-1));
            end
        end


        G = (y-repmat(y0,1,Nu))./epsilon;
    


    
    f = yc(N1:N2);
    
    
    %%% referências futuras
    rfut = [];
    for i=1:n
        rfut = [rfut;
                ref(i,k)*ones(N2(i)-N1(i)+1,1)];
    end

    Hoti = 2*(G'*Qe*G+Qu);
    foti = -2*G'*Qe*(rfut-f);

    X = quadprog(Hoti,foti,[],[]);
    
    for i=1:m
        du(i,k) = X(sum(Nu(1:i-1))+1);
    end
        
    entrada(:,k) = entrada(:,k-1)+du(:,k); 
end

%%
figure
subplot(2,1,1)
plot(saida(1,:))

subplot(2,1,2)
plot(saida(2,:))


%%
%%% Modelo CSTR MIMO completo livro Bequette Process Dynamics
function [x] = modeloCSTR(Ca0,T0,f,fc,caf,tf,tcf)
global ER V k0 k1 k2 k3 Ts

    %%% definição do modelo com as entradas e perturbações constantes. Isto
    %%% simula o sustentador de ordem zero.
    %%% y = [Ca;T];
    fcstr = @(t,y) [f/V*(caf-y(1))-k0*y(1)*exp(-ER/y(2)); % derivada de Ca
                    f/V*(tf-y(2)) + k1*y(1)*exp(-ER/y(2)) + k2*fc*(1-exp(-k3/fc))*(tcf-y(2))]; % derivada de T

    [t,x] = ode45(fcstr,[0 Ts],[Ca0;T0]);
    x = x(end,:)';
                    
end

%% Modelo CSTR nominal utilizado pelos controladores
function [y] = modeloCSTRnominal(Ca,T,f,fc,caf,tf,tcf)
global ER V k0 k1 k2 k3 Ts

y(1) = Ca+ Ts*(f/V*(caf-Ca) - k0*Ca*exp(-ER/T));
y(2) = T+ Ts*(f/V*(tf-T) + k1*Ca*exp(-ER/T) + k2*fc*(1-exp(-k3/fc))*(tcf-T));

end
