%%% Simulação comparativa do PNMPC, NEPSAC e xxx
clear all, 
close all, 
clc

import casadi.*
%% parametros CSTR completo
global ER V k0 k1 k2 k3 Ts opts
%%% condicoes iniciais
Ca0 = 0.1 ; %[mol/l]
T0 = 438.5444; %[K]

f0 = 100; %[l/min] process flow rate
fc0 = 103.41; % [l/min] coolant flow rate

CAf0 = 1; %[mol/l] feed concentration
Tcf0 = 350.0030; %[K] inlet coolant temperature
Tf0 = 350; %[K] feed temperature

Ts = 0.1; % [min]

%%% parametros
ha = 7e5; %[cal/(min K)] heat transfer term
ER = 1e4; %[K] activation energy term
rho = 1e3; %[g/l] liquid densities rho=rho_c
rhoc = rho;
V = 100; %[l] CSTR volume
k0 = 7.2e10; %[1/min] reaction rate constant
dH = -2e5; %[cal/mol] heat of reaction
Cp = 1; %[cal/(g K)] specific heats
Cpc = 1;

k1 = -dH*k0/(rho*Cp);
k2 = rhoc*Cpc/(rho*Cp*V);
k3 = ha/(rhoc*Cpc);


%%% opções para o ODE45 que simulará o sistema contínuo do CSTR
% opts = odeset('RelTol',1e-2,'AbsTol',1e-4);
% opts = odeset('RelTol',1e-3,'AbsTol',1e-6);
opts = odeset('NonNegative',[1 2]);

%% restrições
umax = [150, 130];
umin = [75, 60];

dumax = [10,10];
dumin = -dumax;



%% definicao das variaveis de simulacao
%%% Modelo CSTR listado em DU et al 
n = 2; % numero de saidas
m = 2; % numero de entradas
q = 3; % numero de perturbacoes

na = 1; % numero de amostras de y atrasadas usadas no calculo da saida
nb = 1; % numero de amostras de u atrasadas usadas no calculo da saida
nq = 1; % numero de amostras de q atrasadas usadas no calculo da saida


%% configuracaoes de simulacao
nin = 2; %%% iteracao inicial
nit = nin+1050; %%% iteracao final
nit = nin+250;

N1 = [1,1]
N2 = [50, 50]; 
N2max = max(N2);
Nyi = N2-N1+1;
    
Nu = [30, 30]; % vetor contendo os horizontes de controle

delta = [10000, 1]; % vetor contendo as ponderacoes de saida
% lambda = [((1e-03)^2)*30, (4.3^2)*30]; % vetor contendo as ponderacoes de controle

% delta = [1/Ca0^2,1/50^2];
% lambda = [1/f0^2,1/fc0^2];

lambda = [.5, .5];

alpha = 0.85;

C = 1; % polinômio C do modelo da perturbação
D = [1 -1]; % polinômio D do modelo da perturbação

nc = size(C,2);
nd = size(D,2);

Qe = [];
for i=1:n
    Qe = blkdiag(Qe,delta(i)*eye(N2(i)-N1(i)+1));
end

Qu = [];
for i=1:m
    Qu = blkdiag(Qu,lambda(i)*eye(Nu(i)));
end

%%% matrizes para restrições
LBdu = [];
UBdu = [];
brest0 = [];
Arest = [];
for i=1:2
    LBdu = [LBdu;ones(Nu(i),1)*dumin(i)];
    UBdu = [UBdu;ones(Nu(i),1)*dumax(i)];
    
    T = tril(ones(Nu(i)));
    Arest = blkdiag(Arest,[T;-T]);
    
    brest0 = [brest0;ones(Nu(i),1)*umax(i);ones(Nu(i),1)*-umin(i)];
end



%% inicializacao de algumas variaveis
saida = zeros(n,nit);
saida(1,1:nin) = Ca0;
saida(2,1:nin) = T0;

entrada = zeros(m,nit);
entrada(1,:) = f0;
entrada(2,:) = fc0;

refs = zeros(n,nit+max(N2));
refs(1,:) = Ca0;
refs(2,:) = T0;
refs(1,nin+10:end) = 0.08;
refs(2,nin+150:end) = T0+5;
refs(1,nin+300:end) = 0.1;
refs(2,nin+450:end) = T0;


perts = zeros(q,nit);
perts(1,:) = CAf0;
perts(2,:) = Tf0;
perts(3,:) = Tcf0;

perts(1,nin+600:nit) = CAf0*1.03;
perts(2,nin+900:nit) = Tf0+3;
perts(3,nin+750:nit) = Tcf0-3;


eta = zeros(n,nit+max(N2));
e = zeros(n,nit+max(N2));
saidaModelo = zeros(n,nit);
du = zeros(m,nit);

rfant = saida(:,1);


%% simulação 

for k=nin:nit
    %% modelo do CSTR
    saida(:,k) = modeloCSTR(saida(:,k-1),entrada(:,k-1),perts(:,k-1));

    %% controlador PNMPC
    %%% cálculo da resposta do modelo nominal
    tic
    saidaModelo(:,k) = modeloCSTRnominal(saida(:,k-1),entrada(:,k-1),[CAf0;Tf0;Tcf0]);
    x1(k) = toc;
    tic
    %%% calculo dos etas futuros
    eta(:,k) = saida(:,k) - saidaModelo(:,k); % cálculo do eta atual
    
    e(k) = -C(2:end)*e(k-1:-1:k-nc+1)'+ D*eta(k:-1:k-nd+1)';
    
    for j=1:max(N2)
        eta(:,k+j) = -D(2:end)*eta(:,k+j-1:-1:k+j-nd+1)' + C*e(:,k+j:-1:k+j-nc+1)';
    end
    
    
    %%% cálculo da resposta livre corrigida
    yc = zeros(n,N2max);
    yc(:,1) = modeloCSTRnominal(saida(:,k),entrada(:,k-1),[CAf0;Tf0;Tcf0]) + eta(:,k+1);
    for i=2:N2max
        yc(:,i) = modeloCSTRnominal(yc(:,i-1),entrada(:,k-1),[CAf0;Tf0;Tcf0]) + eta(:,k+i);
    end
    
    %%% passando para o formato vetorial correto
    f = zeros(sum(N2-N1)+1,1);
    for i=1:n
        f(sum(Nyi(1:i-1))+1:sum(Nyi(1:i)),1) = yc(i,N1(i):N2(i))';
    end

    
    %%% calculo de G
    % calculo de y0
    y0 = zeros(n,N2max);
    y0(:,1) = modeloCSTRnominal(saida(:,k),entrada(:,k-1),[CAf0;Tf0;Tcf0]);
    for i=2:N2max
        y0(:,i) = modeloCSTRnominal(y0(:,i-1),entrada(:,k-1),[CAf0;Tf0;Tcf0]);
    end
    
    %%% passando para o formato vetorial correto
    y0v = zeros(sum(N2-N1)+1,1);
    for i=1:n
        y0v(sum(Nyi(1:i-1))+1:sum(Nyi(1:i)),1) = y0(i,N1(i):N2(i))';
    end

    %%% calculo de yi para obter G
    G = zeros(sum(Nyi),sum(Nu));
    for i=1:m
        u0 = entrada(:,k-1);
        
        if(entrada(i,k-1) == 0)
            epsilon = 0.001;
        else
            epsilon = entrada(i,k-1)/1000;
        end

        for j=1:Nu
            yi = zeros(n,N2max);

            ui = repmat(u0,[1 N2max]);
            ui(i,j:end) = ui(i,j:end)+epsilon;

            yi(:,j) = modeloCSTRnominal(saida(:,k),ui(:,1),[CAf0;Tf0;Tcf0]);
            for jj=2:N2max
                yi(:,jj) = modeloCSTRnominal(yi(:,jj-1),ui(:,jj),[CAf0;Tf0;Tcf0]);
            end
            
            %%% passando para o formato vetorial correto
            yiv = zeros(sum(N2-N1)+1,1);
            for jj=1:n
                yiv(sum(Nyi(1:jj-1))+1:sum(Nyi(1:jj)),1) = yi(jj,N1(jj):N2(jj))';
            end
            
            G(:,sum(Nu(1:i-1))+j) = (yiv-y0v)/epsilon;
            
        end
    end

    %%% referências futuras
    rf = alpha*rfant + (1-alpha)*refs(:,k);
    rfant = rf;
    rfut = [];
    for i=1:n
        rfut = [rfut;
                rf(i,1)*ones(N2(i)-N1(i)+1,1)];
    end

    Hoti = 2*(G'*Qe*G+Qu);
    foti = -2*G'*Qe*(rfut-f);
    
    temp = [];
    for i=1:2
        temp = [temp;entrada(i,k-1)*[-ones(Nu(i),1);ones(Nu(i),1)]];
    end
    
    brest = brest0+temp;
    X = quadprog(Hoti,foti,Arest,brest,[],[],LBdu,UBdu);
    x2(k) = toc;
    for i=1:m
        du(i,k) = X(sum(Nu(1:i-1))+1);
    end
        
    entrada(:,k) = entrada(:,k-1)+du(:,k); 

end

saidaPNMPC = saida;
entradaPNMPC = entrada;
duPNMPC = du;

%%
figure
subplot(4,1,1)
plot(saida(1,:))
hold on
plot(refs(1,1:nit))

subplot(4,1,2)
plot(saida(2,:))
hold on
plot(refs(2,1:nit))

subplot(4,1,3)
plot(entrada(1,:))

subplot(4,1,4)
plot(entrada(2,:))

% figure,plot(x2)

%% NEPSAC - inicialização e configuração

T = [];
for i=1:m
    T = blkdiag(T,tril(ones(Nu(i))));
end
T1 = inv(T);


saida = zeros(n,nit);
saida(1,1:nin) = Ca0;
saida(2,1:nin) = T0;

entrada = zeros(m,nit);
entrada(1,:) = f0;
entrada(2,:) = fc0;

eta = zeros(n,nit+max(N2));
e = zeros(n,nit+max(N2));
saidaModelo = zeros(n,nit);
du = zeros(m,nit);

rfant = saida(:,1);

%% simulação 

for k=nin:nit
    %% modelo do CSTR
    saida(:,k) = modeloCSTR(saida(:,k-1),entrada(:,k-1),perts(:,k-1));

    %% controlador NEPSAC
    %%% cálculo da resposta do modelo nominal
    tic
    saidaModelo(:,k) = modeloCSTRnominal(saida(:,k-1),entrada(:,k-1),[CAf0;Tf0;Tcf0]);
    x1n(k) = toc;
    tic
    %%% calculo dos etas futuros
    eta(:,k) = saida(:,k) - saidaModelo(:,k); % cálculo do eta atual
    
    e(k) = -C(2:end)*e(k-1:-1:k-nc+1)'+ D*eta(k:-1:k-nd+1)';
    
    for j=1:max(N2)
        eta(:,k+j) = -D(2:end)*eta(:,k+j-1:-1:k+j-nd+1)' + C*e(:,k+j:-1:k+j-nc+1)';
    end
    
    %%% referências futuras
    rf = alpha*rfant + (1-alpha)*refs(:,k);
    rfant = rf;
    rfut = [];
    for i=1:n
        rfut = [rfut;
                rf(i,1)*ones(N2(i)-N1(i)+1,1)];
    end
    
    ub = repmat(entrada(:,k-1),[1 N2max]); % u base inicial
    dunorm = inf; % inicilizar norma de delta u como infinito
    
    cont = 0;
    u0 = entrada(:,k-1)/1000;
    while (dunorm > 5 && cont <= 100) % iterar até obter a resposta dentro do erro especificado
        %%% Obtenção da resposta base
        yb = zeros(n,N2max); % resposta base
        yb(:,1) =  modeloCSTRnominal(saida(:,k),entrada(:,k-1),[CAf0;Tf0;Tcf0]) + eta(:,k+1); % yb(k+1|k)
        for i=2:N2max
            yb(:,i) = modeloCSTRnominal(yb(:,i-1),entrada(:,k-1),[CAf0;Tf0;Tcf0]) + eta(:,k+i);
        end
        
        %%% passando para o formato vetorial correto
        ybv = zeros(sum(N2-N1)+1,1);
        for i=1:n
            ybv(sum(Nyi(1:i-1))+1:sum(Nyi(1:i)),1) = yb(i,N1(i):N2(i))';
        end
        
        
        %%% cálculo de Ge
        Ge = zeros(sum(Nyi),sum(Nu));

        for jj=1:m
            for j=1:Nu(jj)
                yot = zeros(n,N2max);
                udelta = ub;
                if(j<Nu(jj))
                    udelta(jj,j) = udelta(jj,j)+u0(jj,1);
                else
                    udelta(jj,j:end) = udelta(jj,j:end)+u0(jj,1);
                end

                yot(:,1) = modeloCSTRnominal(saida(:,k),udelta(:,1),[CAf0;Tf0;Tcf0]) + eta(:,k+1);
                for i=2:N2max
                    yot(:,i) = modeloCSTRnominal(yot(:,i-1),udelta(:,i),[CAf0;Tf0;Tcf0]) + eta(:,k+i);
                end
                
                %%% passando para o formato vetorial correto
                yotv = zeros(sum(N2-N1)+1,1);
                for i=1:n
                    yotv(sum(Nyi(1:i-1))+1:sum(Nyi(1:i)),1) = yot(i,N1(i):N2(i))';
                end

                Ge(:,sum(Nu(1:jj-1))+j) = (yotv - ybv)/u0(jj);
                
            end
        end
        
        
        %%% passando para o formato vetorial correto
        ubv = zeros(sum(Nu),1);
        for i=1:m
            ubv(sum(Nu(1:i-1))+1:sum(Nu(1:i)),1) = ub(i,1:Nu(i))';
        end
        
        utemp = zeros(sum(Nu),1);
        for i=1:m
            utemp(sum(Nu(1:i-1))+1:sum(Nu(1:i)),1) = -entrada(i,k-1)*ones(Nu(i),1);
        end
        
        
        Hoti = 2*(Ge'*Qe*Ge+T1'*Qu*T1);
        foti = 2*(Ge'*Qe*(ybv-rfut)+T1'*Qu*T1*(utemp+ubv));
        
        X = quadprog(Hoti,foti,[],[]);
        
        for i=1:m
            ind = sum(Nu(1:i-1))+1:sum(Nu(1:i)); 
            ub(i,1:Nu(i)) =  ub(i,1:Nu(i)) + X(ind)';
            ub(i,Nu(i)+1:end) = ub(i,Nu(i));
        end
        
        dunorm = norm(X);
        cont = cont+1;
    end
    temp(k) = cont;
    
    entrada(:,k) = ub(:,1);
    du(:,k) = entrada(:,k) - entrada(:,k-1);

    x2n(k) = toc; 
end

saidaNEPSAC = saida;
entradaNEPSAC = entrada;
duNEPSAC = du;

%%
figure
subplot(4,1,1)
plot(saida(1,:))
hold on
plot(refs(1,1:nit))

subplot(4,1,2)
plot(saida(2,:))
hold on
plot(refs(2,1:nit))

subplot(4,1,3)
plot(entrada(1,:))

subplot(4,1,4)
plot(entrada(2,:))

% figure,plot(x2n)
% figure,plot(x1n)

%% Geração das figuras
tamletra=14;
tamnum=14;
nin = 2;

t = ((nin:600)-nin)*Ts;
ind = nin:600;

t = ((nin:nit)-nin)*Ts;
ind = nin:nit;

hf= figure
h=subplot(4,1,1)
plot(t,saidaPNMPC(1,ind),'LineWidth',2)
hold on
plot(t,saidaNEPSAC(1,ind),'-.','LineWidth',2)
plot(t,refs(1,ind),':','LineWidth',2)

% ylim([0 100])
hl = legend('PNMPC','NEPSAC','Referência','Location','SouthEast')

ylabel('C_a(t)','FontSize', tamletra)
xlim([0 105])
set(h, 'FontSize', tamnum);
grid on

h=subplot(4,1,2)
plot(t,saidaPNMPC(2,ind),'LineWidth',2)
hold on
plot(t,saidaNEPSAC(2,ind),'-.','LineWidth',2)
plot(t,refs(2,ind),':','LineWidth',2)

ylabel('T(t)','FontSize', tamletra)
xlim([0 105])
set(h, 'FontSize', tamnum);
grid on

h=subplot(4,1,3)
plot(t,entradaPNMPC(1,ind),'LineWidth',2)
hold on
plot(t,entradaNEPSAC(1,ind),'-.','LineWidth',2)

ylabel('F(t)','FontSize', tamletra)
xlim([0 105])
set(h, 'FontSize', tamnum);
grid on

h=subplot(4,1,4)
plot(t,entradaPNMPC(2,ind),'LineWidth',2)
hold on
plot(t,entradaNEPSAC(2,ind),'-.','LineWidth',2)

ylabel('F_c(t)','FontSize', tamletra)
xlabel('Tempo [min]')
xlim([0 105])
set(h, 'FontSize', tamnum);
grid on

hf.Position = [680 383 998 595];
hl.Position = [0.7834 0.7304 0.1333 0.1143];

print('casoCSTR_comparativo','-depsc')


figure
semilogy(x2(1,ind),'LineWidth',2)
hold on
semilogy(x2n(1,ind),'-.','LineWidth',2)

xlim([0 nit])
hl = legend('PNMPC','NEPSAC','Location','NorthWest')

ylabel('Tempo [s]','FontSize', tamletra)
xlabel('Amostras')

set(gca, 'FontSize', tamnum);
grid on

% hl.Position = [0.7834 0.7304 0.1333 0.1143];

print('casoCSTR_comparativo_tempos','-depsc')

%%
modeloCSTRnominal([Ca0;T0],[f0;fc0],[CAf0;Tf0;Tcf0])
 
modeloCSTR([Ca0;T0],[f0;fc0],[CAf0;Tf0;Tcf0])


%% calculo PO
% perts(1,nin+600:nit) = CAf0*1.03;
% perts(2,nin+650:nit) = Tf0+3;
% perts(3,nin+900:nit) = Tcf0-3;

q0 = [CAf0*1.03;Tf0+3;Tcf0]

f1 = @(y) [y(3)/V*(q0(1)-y(1))-k0*y(1)*exp(-ER/y(2)); % derivada de Ca
           y(3)/V*(q0(2)-y(2)) + k1*y(1)*exp(-ER/y(2)) + k2*y(4)*(1-exp(-k3/y(4)))*(q0(3)-y(2))]; % derivada de T

f1 = @(y) [y(1)/V*(q0(1)-Ca0)-k0*Ca0*exp(-ER/T0); % derivada de Ca
           y(1)/V*(q0(2)-T0) + k1*Ca0*exp(-ER/T0) + k2*y(2)*(1-exp(-k3/y(2)))*(q0(3)-T0)]; % derivada de T
       
       
[X,fval] = fsolve(f1,[f0;fc0])

modeloCSTRnominal([Ca0;T0],X,q0)

%%
%%% Modelo CSTR MIMO completo livro Bequette Process Dynamics
function [x] = modeloCSTR(y,u,q)
    global ER V k0 k1 k2 k3 Ts opts
    Ca = y(1);
    T = y(2);

    f = u(1);
    fc = u(2);

    caf = q(1);
    tf = q(2);
    tcf = q(3);
    %%% definição do modelo com as entradas e perturbações constantes. Isto
    %%% simula o sustentador de ordem zero.
    %%% y = [Ca;T];
    fcstr = @(t,y) [f/V*(caf-y(1))-k0*y(1)*exp(-ER/y(2)); % derivada de Ca
                    f/V*(tf-y(2)) + k1*y(1)*exp(-ER/y(2)) + k2*fc*(1-exp(-k3/fc))*(tcf-y(2))]; % derivada de T

    [t,x] = ode45(fcstr,[0 Ts],[Ca;T],opts);
    x = x(end,:)';
                    
end

%% Modelo CSTR nominal utilizado pelos controladores
function [yo] = modeloCSTRnominal(y,u,q)
    global ER V k0 k1 k2 k3 Ts 
    Ca = y(1);
    T = y(2);

    f = u(1);
    fc = u(2);

    caf = q(1);
    tf = q(2);
    tcf = q(3);

    yo(1,1) = Ca+ Ts*(f/V*(caf-Ca) - k0*Ca*exp(-ER/T));
    yo(2,1) = T+ Ts*(f/V*(tf-T) + k1*Ca*exp(-ER/T) + k2*fc*(1-exp(-k3/fc))*(tcf-T));

    %%% saturação dos estados
    if(yo(1,1)<0)
        yo(1,1)= 0;
    end

    if(yo(2,1)<0)
        yo(2,1) = 0;
    end


end
