%%% Simulação comparativa do PNMPC, NEPSAC e xxx
clear all, 
% close all, 
clc

import casadi.*
%% parametros CSTR completo
global ER V k0 k1 k2 k3 Ts opts
%%% condicoes iniciais
Ca0 = 0.1 ; %[mol/l]
T0 = 438.5444; %[K]

f0 = 100; %[l/min] process flow rate
fc0 = 103.41; % [l/min] coolant flow rate

CAf0 = 1; %[mol/l] feed concentration
Tcf0 = 350.0030; %[K] inlet coolant temperature
Tf0 = 350; %[K] feed temperature

Ts = 0.1; % [min]

%%% parametros
ha = 7e5; %[cal/(min K)] heat transfer term
ER = 1e4; %[K] activation energy term
rho = 1e3; %[g/l] liquid densities rho=rho_c
rhoc = rho;
V = 100; %[l] CSTR volume
k0 = 7.2e10; %[1/min] reaction rate constant
dH = -2e5; %[cal/mol] heat of reaction
Cp = 1; %[cal/(g K)] specific heats
Cpc = 1;

k1 = -dH*k0/(rho*Cp);
k2 = rhoc*Cpc/(rho*Cp*V);
k3 = ha/(rhoc*Cpc);


%% restrições
umax = [150, 130];
umin = [75, 60];

dumax = [10,10];
dumin = -dumax;

%% definicao das variaveis de simulacao
%%% Modelo CSTR listado em DU et al 
n = 2; % numero de saidas
m = 2; % numero de entradas
q = 3; % numero de perturbacoes

na = 1; % numero de amostras de y atrasadas usadas no calculo da saida
nb = 1; % numero de amostras de u atrasadas usadas no calculo da saida
nq = 1; % numero de amostras de q atrasadas usadas no calculo da saida


%% configuracaoes de simulacao
nin = 2; %%% iteracao inicial
nit = nin+1050; %%% iteracao final
% nit = nin+250;

N1 = [1,1]
N2 = [50, 50]; 
N2max = max(N2);
Nyi = N2-N1+1;
    
Nu = [30, 30]; % vetor contendo os horizontes de controle

delta = [10000, 1]; % vetor contendo as ponderacoes de saida
% lambda = [((1e-03)^2)*30, (4.3^2)*30]; % vetor contendo as ponderacoes de controle

% delta = [1/Ca0^2,1/50^2];
% lambda = [1/f0^2,1/fc0^2];

lambda = [.5, .5];
% lambda = [0,0]

alpha = 0.85;

C = 1; % polinômio C do modelo da perturbação
D = [1 -1]; % polinômio D do modelo da perturbação

nc = size(C,2);
nd = size(D,2);

%% configurações CASADI
Ca = MX.sym('Ca',1);
T = MX.sym('T',1);
x = [Ca; T];

f = MX.sym('f',1);
fc = MX.sym('fc',1);
u = [f;fc];

caf = MX.sym('caf',1);
tf = MX.sym('tf',1);
tcf = MX.sym('tcf',1);
qs = [caf;tf;tcf];

etaca = MX.sym('etaca',1);
etat = MX.sym('etat',1);
eta1 = [etaca;etat];

modeloNominal = [Ca+ Ts*(f/V*(caf-Ca) - k0*Ca*exp(-ER/T))+etaca;
                 T+ Ts*(f/V*(tf-T) + k1*Ca*exp(-ER/T) + k2*fc*(1-exp(-k3/fc))*(tcf-T))+etat];

fModelo = Function('F',{x,u,qs,eta1},{modeloNominal});

U1 = MX.sym('f',Nu(1),1);
U2 = MX.sym('fc',Nu(2),1);
U = [U1{:};U2{:}];

eta1 = MX.sym('eta1',max(N2),1);
eta2 = MX.sym('eta1',max(N2),1);

X0 = MX.sym('X0',2,1);
U0 = MX.sym('U0',2,1);
X = [];
for k=1:max(N2)
%     uk = MX.sym('uk',2,1');
    uk = [];
    if(k>Nu(1))
        uk = [uk;U1(Nu(1))];
    else
        uk = [uk;U1(k)];
    end
    if(k>Nu(2))
        uk = [uk;U2(Nu(2))];
    else
        uk = [uk;U2(k)];
    end    
    
    if(k==1)
        X = [X,fModelo(X0,uk,qs,[eta1(k);eta2(k)])];
    else
        X = [X,fModelo(X(:,k-1),uk,qs,[eta1(k);eta2(k)])];
    end
end

J = 0;
R = MX.sym('R',2,1);
for i=1:2
    for j=N1(i):N2(i)
        J = J+delta(i)*(R(i)-X(i,j))^2;
    end
end

lbu = [];
ubu = [];
lbdu = [];
ubdu = [];
g = {};

for i=1:2
    indi = sum(Nu(1:i-1))+1;
    indf = sum(Nu(1:i));
    T = tril(ones(Nu(i)));
    
    du = inv(T)*(U(indi:indf)-U0(i)*ones(Nu(i),1));
    
    for j=1:Nu(i)
        J = J+ lambda(i)*du(j)^2;
    end
    
    lbu = [lbu;umin(i)*ones(Nu(i),1)];
    ubu = [ubu;umax(i)*ones(Nu(i),1)];
    lbdu = [lbdu;dumin(i)*ones(Nu(i),1)];
    ubdu = [ubdu;dumax(i)*ones(Nu(i),1)];
    
    g = {g{:},du};
    
end
    

% Create an NLP solver
prob = struct('f', J, 'x', vertcat(U{:}), 'g', vertcat(g{:}),'p',vertcat(R{:},X0{:},U0{:},eta1{:},eta2{:},qs{:}));
solver = nlpsol('solver', 'ipopt', prob);


%% inicializacao de algumas variaveis
saida = zeros(n,nit);
saida(1,1:nin) = Ca0;
saida(2,1:nin) = T0;

entrada = zeros(m,nit);
entrada(1,:) = f0;
entrada(2,:) = fc0;

refs = zeros(n,nit+max(N2));
refs(1,:) = Ca0;
refs(2,:) = T0;
refs(1,nin+10:end) = 0.08;
refs(2,nin+150:end) = T0+5;
refs(1,nin+300:end) = 0.1;
refs(2,nin+450:end) = T0;


perts = zeros(q,nit);
perts(1,:) = CAf0;
perts(2,:) = Tf0;
perts(3,:) = Tcf0;

perts(1,nin+600:nit) = CAf0*1.03;
perts(2,nin+900:nit) = Tf0+3;
perts(3,nin+750:nit) = Tcf0-3;


eta = zeros(n,nit+max(N2));
e = zeros(n,nit+max(N2));
saidaModelo = zeros(n,nit);
du = zeros(m,nit);

rfant = saida(:,1);

u_opt = [];
for i=1:2
    u_opt = [u_opt;entrada(i,1)*ones(Nu(i),1)];
end

%% simulação 

for k=nin:nit
    %% modelo do CSTR
    saida(:,k) = modeloCSTR(saida(:,k-1),entrada(:,k-1),perts(:,k-1));

    %% controlador NMPC com CASADI
    %%% cálculo da resposta do modelo nominal
    tic
    saidaModelo(:,k) = modeloCSTRnominal(saida(:,k-1),entrada(:,k-1),[CAf0;Tf0;Tcf0]);
    x1(k) = toc;
    tic
    %%% calculo dos etas futuros
    eta(:,k) = saida(:,k) - saidaModelo(:,k); % cálculo do eta atual
    
    e(k) = -C(2:end)*e(k-1:-1:k-nc+1)'+ D*eta(k:-1:k-nd+1)';
    
    for j=1:max(N2)
        eta(:,k+j) = -D(2:end)*eta(:,k+j-1:-1:k+j-nd+1)' + C*e(:,k+j:-1:k+j-nc+1)';
    end

    %%% referências futuras
    rf = alpha*rfant + (1-alpha)*refs(:,k);
    rfant = rf;


    p0 = vertcat(rf,saida(:,k),entrada(:,k-1),eta(1,k+1:k+max(N2))',eta(2,k+1:k+max(N2))',perts(:,1));
    
    % Solve the NLP
    sol = solver('x0', u_opt,'p',p0, 'lbx', lbu, 'ubx', ubu, ...
                 'lbg', lbdu, 'ubg', ubdu);
    u_opt = full(sol.x);


    for i=1:2
        entrada(i,k) = u_opt(sum(Nu(1:i-1))+1);
    end
end

saidaPNMPC = saida;
entradaPNMPC = entrada;
duPNMPC = du;

%%
figure
subplot(4,1,1)
plot(saida(1,:))
hold on
plot(refs(1,1:nit))

subplot(4,1,2)
plot(saida(2,:))
hold on
plot(refs(2,1:nit))

subplot(4,1,3)
plot(entrada(1,:))

subplot(4,1,4)
plot(entrada(2,:))




%%
%%% Modelo CSTR MIMO completo livro Bequette Process Dynamics
function [x] = modeloCSTR(y,u,q)
    global ER V k0 k1 k2 k3 Ts opts
    Ca = y(1);
    T = y(2);

    f = u(1);
    fc = u(2);

    caf = q(1);
    tf = q(2);
    tcf = q(3);
    %%% definição do modelo com as entradas e perturbações constantes. Isto
    %%% simula o sustentador de ordem zero.
    %%% y = [Ca;T];
    fcstr = @(t,y) [f/V*(caf-y(1))-k0*y(1)*exp(-ER/y(2)); % derivada de Ca
                    f/V*(tf-y(2)) + k1*y(1)*exp(-ER/y(2)) + k2*fc*(1-exp(-k3/fc))*(tcf-y(2))]; % derivada de T

    [t,x] = ode45(fcstr,[0 Ts],[Ca;T],opts);
    x = x(end,:)';
                    
end

%% Modelo CSTR nominal utilizado pelos controladores
function [yo] = modeloCSTRnominal(y,u,q)
    global ER V k0 k1 k2 k3 Ts 
    Ca = y(1);
    T = y(2);

    f = u(1);
    fc = u(2);

    caf = q(1);
    tf = q(2);
    tcf = q(3);

    yo(1,1) = Ca+ Ts*(f/V*(caf-Ca) - k0*Ca*exp(-ER/T));
    yo(2,1) = T+ Ts*(f/V*(tf-T) + k1*Ca*exp(-ER/T) + k2*fc*(1-exp(-k3/fc))*(tcf-T));

    %%% saturação dos estados
    if(yo(1,1)<0)
        yo(1,1)= 0;
    end

    if(yo(2,1)<0)
        yo(2,1) = 0;
    end


end
