%%% Exemplo do tanque não linear comparando GPC linear, GPC com 
%%% resposta livre não linear e o NEPSAC

clear all
close all
clc

%% parametros comuns aos controladores 
N1 = 1;
N2 = 3;
Ny = N2-N1+1;
Nu = 2;
Nq = 0;

lambda = 0; % ponderação do incremento de controle na função custo.
delta = 1; % ponderação do erro na função custo

Qe = delta*eye(N2-N1+1);
Qu = lambda*eye(Nu);

C = 1; % polinômio C do modelo da perturbação
D = [1 -1]; % polinômio D do modelo da perturbação

nc = size(C,2);
nd = size(D,2);

umax = 100; % valor máximo da abertura da válvula
umin = 0; % valor mínimo da abertura da válvula

refFutura = false; % usar referencia futura?

%% parametros gerais da simulação
%%% condições iniciais
y0 = 0;
u0 = -0.25;
q0 = 0;

% y0 = 0.5;
% u0 =-0.5666;
% q0 = 0;


%%% parâmetros do modelo
dn = 0; % atraso discreto nominal

nin = dn+5;
nit = nin+600; % numero de iterações

%%% definição da referência
ref = y0*ones(1,nit);
ref(20:end) = 0.8;
ref(100:end) = -0.1;


t = linspace(-2,2,400);
fo = 6;
f1 = 8;
refchirp = chirp(t,fo,1,f1,'quadratic',[],'convex');
figure
plot(refchirp)

ampr = linspace(.3,2.1,200);
ampr = [ampr ampr(end:-1:1)]
ref(nit-400+1:end) = ampr.*refchirp+1;
ref(end:end+N2) = ref(end);


%%% definição da perturbação
q = q0*ones(1,nit);
q(150:end) = 0.2;


alpha = 0;

figure
plot(ref(nit-400+1:end))

%%


%% Configuração do GPC linear
%%% Modelo da planta linearizado no ponto de operação
a1 = (2.5*y0*(1+y0^2+y0^2)-2.5*y0*y0*2*y0)/(1+y0^2+y0^2)^2-0.3*0.5*sin(0.5*(y0+y0));
a2 = a1;
b1 = 1.2;

A = [1 -a1 -a2];
B = b1; 

d=dn;

nb = size(B,2)-1; % ordem do polinômio B
na = size(A,2)-1; % ordem do polinômio A


%%% Obtenção das matrizes do GPC linear
Btil = conv(B,[zeros(1,d) 1]); % incorporação do atraso no numerador B

G = zeros(Ny,Ny); % matriz dinâmica G
H = zeros(Ny,nb+d);

[E,F] = diofantina(conv(A,[1 -1]),N1,N2);

for i=N1:N2
    EjB = conv(E(i-N1+1,1:i),Btil);
    
    G(i-N1+1,1:i) = EjB(i:-1:1);    
    H(i-N1+1,:) = EjB(i+1:end);

    
end
G = G(:,1:Nu);

Hoti = 2*(G'*Qe*G+Qu);
foti = -2*G'*Qe;


%% simulação do GPC linear
saida= y0*ones(1,nit);
entrada = u0*ones(1,nit);
du = zeros(1,nit);

rfant = y0;


for k=nin:nit
    %%% simulacao do modelo
    saida(k) = modeloSim(saida(:,k-1:-1:k-2),entrada(:,k-1-dn),q(:,k-1));
    
    %% controlador GPC
    % calculo das referencias
    
    if(refFutura)
        rfut = ref(k+N1:k+N2)';
    else
        rf = alpha*rfant + (1-alpha)*ref(k);
        rfant = rf;
        rfut = rf*ones(N2-N1+1,1);
    end

    %%% cálculo da resposta livre;
    f = F*saida(1,k:-1:k-na)';
    
    if(~isempty(H))
        f = f+ H*du(1,k-1:-1:k-nb-d)'; % parcela dos incrementos de controle
    end
    
   
    %%% cálculo do incremento de controle ótimo
    duOti = quadprog(Hoti,foti*(rfut-f),[],[]);
    du(k) = duOti(1);
    
    %%% cálculo do sinal de controle ótimo
    entrada(k) = du(k)+entrada(k-1);
 

end

saidaGPC = saida;
entradaGPC = entrada;
duGPC = du;


%% plotagem das figuras
t = (2:nit)-2;
figure
subplot(2,1,1)
plot(t,ref(2:nit),t,saidaGPC(2:nit))
% ylim([0 100])

subplot(2,1,2)
plot(t,entradaGPC(2:nit))


%% simulação do GPC com resposta livre não linear
saida= y0*ones(1,nit);
entrada = u0*ones(1,nit);
du = zeros(1,nit);

rfant = y0;

saidaModelo = y0*ones(1,nit);

eta = zeros(1,nit+N2);
e = zeros(1,nit+N2);


for k=nin:nit
    %%% simulacao do CSTR
    saida(k) = modeloSim(saida(:,k-1:-1:k-2),entrada(:,k-1-dn),q(:,k-1));

    %% controlador GPC com modelo não linear na resposta livre
    % calculo das referencias
    if(refFutura)
        rfut = ref(k+N1:k+N2)';
    else
        rf = alpha*rfant + (1-alpha)*ref(k);
        rfant = rf;
        rfut = rf*ones(N2-N1+1,1);
    end

    
    %%% cálculo da resposta livre;
    saidaModelo(k) = modeloNominal(saida(k-1:-1:k-2),entrada(k-1-dn));
    
    %%% calculo dos etas futuros
    eta(k) = saida(k) - saidaModelo(k); % cálculo do eta atual
    
    e(k) = -C(2:end)*e(k-1:-1:k-nc+1)'+ D*eta(k:-1:k-nd+1)';
    
    for j=1:N2
        eta(k+j) = -D(2:end)*eta(k+j-1:-1:k+j-nd+1)' + C*e(k+j:-1:k+j-nc+1)';
    end

    yb = zeros(N2,1); % resposta base
                        % yb=[y(k+1);y(k+2);...;y(k+N2)];
    ub = [entrada(k-dn:k-1), entrada(k-1)*ones(1,N2)]; % u base inicial

    yb(1) = modeloNominal(saida(k:-1:k-1),ub(1)) + eta(k+1); % yb(k+1|k)
    yb(2) = modeloNominal([yb(1),saida(k)],ub(2)) + eta(k+2); % yb(k+2|k)
    for i=3:N2
        yb(i) = modeloNominal(yb(i-1:-1:i-2),ub(i)) + eta(k+i); %yb(k+i|k)
    end
    
    f = yb(N1:N2);
    
    %%% cálculo do incremento de controle ótimo
    duOti = quadprog(Hoti,foti*(rfut-f),[],[]);
    du(k) = duOti(1);
    
    %%% cálculo do sinal de controle ótimo
    entrada(k) = du(k)+entrada(k-1);
 

end
saidaGPCnlin = saida;
entradaGPCnlin = entrada;
duGPCnlin = du;

%% plotagem das figuras
t = (2:nit)-2;
figure
subplot(2,1,1)
plot(t,ref(2:nit),t,saida(2:nit))
% ylim([0 100])

subplot(2,1,2)
plot(t,entrada(2:nit))


%% Configuração do NEPSAC
T = tril(ones(Nu));
T1 = inv(T);

%% simulação do NEPSAC
saida= y0*ones(1,nit);
entrada = u0*ones(1,nit);
du = zeros(1,nit);

rfant = y0;

saidaModelo = y0*ones(1,nit);

eta = zeros(1,nit+N2);
e = zeros(1,nit+N2);


for k=nin:nit
    %%% simulacao do CSTR
    saida(k) = modeloSim(saida(:,k-1:-1:k-2),entrada(:,k-1-dn),q(:,k-1));
    
    %% controlador NEPSAC
    %%% cálculo da resposta do modelo nominal
    saidaModelo(k) = modeloNominal(saida(k-1:-1:k-2),entrada(k-1-dn));
    
    %%% calculo dos etas futuros
    eta(k) = saida(k) - saidaModelo(k); % cálculo do eta atual
    
    e(k) = -C(2:end)*e(k-1:-1:k-nc+1)'+ D*eta(k:-1:k-nd+1)';
    
    for j=1:N2
        eta(k+j) = -D(2:end)*eta(k+j-1:-1:k+j-nd+1)' + C*e(k+j:-1:k+j-nc+1)';
    end
    
    ub = [entrada(k-dn:k-1), entrada(k-1)*ones(1,N2)]; % u base inicial
                          %ub = [u(k-d) u(k-d+1) ... u(k) u(k+1)... u(k+N2-1)
    dunorm = inf; % inicilizar norma de delta u como infinito
   
    % calculo das referencias
    if(refFutura)
        rfut = ref(k+N1:k+N2)';
    else
        rf = alpha*rfant + (1-alpha)*ref(k);
        rfant = rf;
        rfut = rf*ones(N2-N1+1,1);
    end

    
    cont = 0;
    while (dunorm > 1e-7 && cont <= 100) % iterar até obter a resposta dentro do erro especificado
        %%% Obtenção da resposta base
        yb = zeros(N2,1); % resposta base
                            % yb=[y(k+1);y(k+2);...;y(k+N2)];

        yb(1) = modeloNominal(saida(k:-1:k-1),ub(1)) + eta(k+1); % yb(k+1|k)
        yb(2) = modeloNominal([yb(1),saida(k)],ub(2)) + eta(k+2); % yb(k+2|k)
        for i=3:N2
            yb(i) = modeloNominal(yb(i-1:-1:i-2),ub(i)) + eta(k+i); %yb(k+i|k)
        end

        
        %%% cálculo do controle ótimo

        %%% cálculo de Ge
        Ge = zeros(N2,Nu);
        u00 = entrada(k-1)/1000; 
%         u00 = 1;
        for j=1:Nu
            yot = zeros(N2,1);
            udelta = ub;
            if(j<Nu)
                udelta(j+dn) = udelta(j+dn)+u00;
            else
                udelta(j+dn:end) = udelta(j+dn:end)+u00;
            end

            yot(1) = modeloNominal(saida(k:-1:k-1),udelta(1)) + eta(k+1);
            yot(2) = modeloNominal([yot(1),saida(k)],udelta(2)) + eta(k+2);
            for i=3:N2
                yot(i) = modeloNominal(yot(i-1:-1:i-2),udelta(i)) + eta(k+i);
            end
                        
            Ge(:,j) = (yot - yb)/(u00);
        end
        Ge = Ge(N1:N2,:);
        
        
        
        Hoti = 2*(Ge'*Qe*Ge+T1'*Qu*T1);
        foti = 2*(Ge'*Qe*(yb(N1:end)-rfut)+T1'*Qu*T1*(-entrada(k-1)*ones(Nu,1)+ub(dn+1:dn+Nu)'));
        
        X = quadprog(Hoti,foti,[],[]);
        
        ub(dn+1:dn+Nu) = ub(dn+1:dn+Nu)+X';
        ub(dn+1+Nu:end) = ub(dn+Nu);
        
        dunorm = norm(X);
        cont = cont+1;
    end
    temp(1,k) = cont;
    temp(2,k) = dunorm;
    
    entrada(k) = ub(dn+2);
 

end

%% plotagem das figuras
t = (2:nit)-2;
figure
subplot(2,1,1)
plot(t,ref(2:nit),t,saida(2:nit))
% ylim([0 100])

subplot(2,1,2)
plot(t,entrada(2:nit))

%%
%% Geração das figuras
tamletra=14;
tamnum=14;
nin = 2;

figure
h=subplot(2,1,1)
plot(saidaGPC(1,nin+1:nit),'-.','LineWidth',2)
hold on
plot(saidaGPCnlin(1,nin+1:nit),'LineWidth',2)
plot(saida(1,nin+1:nit),'--','LineWidth',2)
plot(ref(1,nin+1:nit),':','LineWidth',2)
% ylim([0 100])
hl = legend('GPC linear','GPC não linear','NEPSAC','Referência','Location','SouthEast')
hl.Position = [0.0792 0.4793 0.2964 0.2131];
ylabel('Saída','FontSize', tamletra)
set(h, 'FontSize', tamnum);
grid on

h = subplot(2,1,2)
plot(entradaGPC(1,nin+1:nit),'-.','LineWidth',2)
hold on
plot(entradaGPCnlin(1,nin+1:nit),'LineWidth',2)
plot(entrada(1,nin+1:nit),'--','LineWidth',2)

ylabel('Entrada','FontSize', tamletra)
% ylim([20 60])
grid on
set(h, 'FontSize', tamnum);
% ylim([-5 5])

% h=subplot(3,1,3)
% plot(t,erro(1:nit-N2-1),'LineWidth',2)
% title('Erro', 'FontSize', tamletra);
% xlabel('tempo', 'FontSize', tamletra);
% set(h, 'FontSize', tamnum);



% print('tanqueCompleto_comparativo_l10','-depsc')


%%
y00 = 0.5
x0 = (2.5*y00*y00)/(1+y00^2+y00^2)+0.3*cos(0.5*(y00+y00))
u00 = -x0/1.2 
x = modeloSim([y00;y00],u00,0)

%%
%%% Modelo CSTR isotérmico SISO livro Bequette - Process Dynamics
function [x] = modeloSim(y,u,q)
    yk1 = y(1); %y(k-1)
    yk2 = y(2); %y(k-2)
    
    uq = u+q;

    
    x = (2.5*yk1*yk2)/(1+yk1^2+yk2^2)+1.2*uq+0.3*cos(0.5*(yk1+yk2));
    
%     if(x>5)
%         x = 5;
%     elseif(x<-5)
%         x = -5;
%     end

end

%% Modelo CSTR nominal (Modelo Volterra) utilizado pelos controladores
function [ret] = modeloNominal(y,u)

    ret = modeloSim(y,u,0);
    
end