%%% Exemplo do tanque não linear comparando GPC linear, GPC com 
%%% resposta livre não linear e o NEPSAC

clear all
close all
clc

addpath('../../Bibliotecas/')
run('../../Bibliotecas/parametrosFiguras.m')
%% parametros comuns aos controladores 
N1 = 1;
N2 = 40;
Ny = N2-N1+1;
Nu = 5;
Nq = 0;

lambda = 5; % ponderação do incremento de controle na função custo.
delta = 1; % ponderação do erro na função custo

Qe = delta*eye(N2-N1+1);
Qu = lambda*eye(Nu);

C = 1; % polinômio C do modelo da perturbação
D = [1 -1]; % polinômio D do modelo da perturbação

nc = size(C,2);
nd = size(D,2);

umax = 100; % valor máximo da abertura da válvula
umin = 0; % valor mínimo da abertura da válvula

%% parametros gerais da simulação
%%% condições iniciais
Ca0 = 1.235;
Cb0 = 0.9;
Caf0 = 5.1;
FV0 = 0.3138;

%%% parâmetros do modelo
global k1 k2 k3 Ts opts

k1 = 0.8433;
k2 = k1;
k3 = 0.1123;

Ts = 0.18;

opts = odeset('NonNegative',[1 2]);

% Ln = 0.72; %atraso nominal
% dn = 4; % atraso discreto nominal

Ln = 0;
dn = 0;

nin = dn+5;
nit = 200; % numero de iterações

%%% definição da referência
ref = Cb0*ones(1,nit);
ref(20:end) = 0.6;
ref(80:end) = 0.8;
% ref(100:end) = 1.1;

%%% definição da perturbação
q = Caf0*ones(1,nit);
q(150:end) = 4.6;

alpha = 0;



%% Configuração do GPC linear
%%% Modelo da planta linearizado no ponto de operação
s = tf('s');

Ca0g = 1.032;
Cb0g = 0.8;
Caf0g = Caf0;
FV0g = 0.2432;


CaU = (Caf0g-Ca0g)/(s+(k1+2*k3*Ca0g+FV0g))
CbCa = k1/(s+(k2+FV0g))
CbU = -Cb0g/(s+(k2+FV0g))

Gs = minreal(CbCa*CaU + CbU,1e-3)

Gz = c2d(Gs,Ts,'zoh')

A = Gz.den{1};
B = Gz.num{1}(2:end); 

d=dn;
dq = d;

nb = size(B,2)-1; % ordem do polinômio B
na = size(A,2)-1; % ordem do polinômio A


%%% Obtenção das matrizes do GPC linear
Btil = conv(B,[zeros(1,d) 1]); % incorporação do atraso no numerador B

G = zeros(Ny,Ny); % matriz dinâmica G
H = zeros(Ny,nb+d);

[E,F] = diofantina(conv(A,[1 -1]),N1,N2);

for i=N1:N2
    EjB = conv(E(i-N1+1,1:i),Btil);
    
    G(i-N1+1,1:i) = EjB(i:-1:1);    
    H(i-N1+1,:) = EjB(i+1:end);

    
end
G = G(:,1:Nu);

Hoti = 2*(G'*Qe*G+Qu);
foti = -2*G'*Qe;


%% simulação do GPC linear
estados = repmat([Ca0;Cb0],[1 nit]);
saida= estados(2,:);

entrada = FV0*ones(1,nit);
du = zeros(1,nit);

rfant = Cb0;


for k=nin:nit
    %%% simulacao do CSTR
    estados(:,k) = modeloCSTR(estados(:,k-1),entrada(:,k-1-dn),q(:,k-1));
    saida(k) = estados(2,k);
    
    %% controlador GPC
    % calculo das referencias
    rf = alpha*rfant + (1-alpha)*ref(k);
    rfant = rf;
    rfut = rf*ones(N2-N1+1,1);

    %%% cálculo da resposta livre;
    f = F*saida(1,k:-1:k-na)';
    
    if(~isempty(H))
        f = f+ H*du(1,k-1:-1:k-nb-d)'; % parcela dos incrementos de controle
    end
    
   
    %%% cálculo do incremento de controle ótimo
    duOti = quadprog(Hoti,foti*(rfut-f),[],[]);
    du(k) = duOti(1);
    
    %%% cálculo do sinal de controle ótimo
    entrada(k) = du(k)+entrada(k-1);
 

end

saidaGPC = saida;
entradaGPC = entrada;
duGPC = du;


%% plotagem das figuras
t = (2:nit)-2;
figure
subplot(2,1,1)
plot(t,ref(2:nit),t,saidaGPC(2:nit))
% ylim([0 100])

subplot(2,1,2)
plot(t,entradaGPC(2:nit))


%% simulação do GPC com resposta livre não linear
estados = repmat([Ca0;Cb0],[1 nit]);
saida= estados(2,:);

saidaModelo = repmat([Ca0;Cb0],[1,nit]);
entrada = FV0*ones(1,nit);
du = zeros(1,nit);

eta = zeros(2,nit+N2);
e = zeros(2,nit+N2);
rfant = Cb0;



for k=nin:nit
    %%% simulacao do CSTR
    estados(:,k) = modeloCSTR(estados(:,k-1),entrada(:,k-1-dn),q(:,k-1));
    saida(k) = estados(2,k);

    %% controlador GPC com modelo não linear na resposta livre
    % calculo das referencias
    rf = alpha*rfant + (1-alpha)*ref(k);
    rfant = rf;
    rfut = rf*ones(N2-N1+1,1);
    
    %%% cálculo da resposta livre;
    saidaModelo(:,k) = modeloCSTRnominal(estados(:,k-1),entrada(k-1-dn),Caf0);
    
    %%% calculo dos etas futuros
    eta(:,k) = estados(:,k) - saidaModelo(:,k); % cálculo do eta atual
    
    e(1,k) = -C(2:end)*e(1,k-1:-1:k-nc+1)'+ D*eta(1,k:-1:k-nd+1)';
    e(2,k) = -C(2:end)*e(2,k-1:-1:k-nc+1)'+ D*eta(2,k:-1:k-nd+1)';
    
    for j=1:N2
        eta(1,k+j) = -D(2:end)*eta(1,k+j-1:-1:k+j-nd+1)' + C*e(1,k+j:-1:k+j-nc+1)';
        eta(2,k+j) = -D(2:end)*eta(2,k+j-1:-1:k+j-nd+1)' + C*e(2,k+j:-1:k+j-nc+1)';
    end

    yb = zeros(2,N2); % resposta base
                        % yb=[y(k+1);y(k+2);...;y(k+N2)];
    ub = [entrada(k-dn:k-1), entrada(k-1)*ones(1,N2)]; % u base inicial

    yb(:,1) = modeloCSTRnominal(estados(:,k),ub(1),Caf0) + eta(:,k+1); % yb(k+1|k)
    for i=2:N2
        yb(:,i) = modeloCSTRnominal(yb(:,i-1),ub(i),Caf0) + eta(:,k+i); %yb(k+i|k)
    end
    
    f = yb(2,N1:N2)';
    
    %%% cálculo do incremento de controle ótimo
    duOti = quadprog(Hoti,foti*(rfut-f),[],[]);
    du(k) = duOti(1);
    
    %%% cálculo do sinal de controle ótimo
    entrada(k) = du(k)+entrada(k-1);
 

end
saidaGPCnlin = saida;
entradaGPCnlin = entrada;
duGPCnlin = du;

%% plotagem das figuras
t = (2:nit)-2;
figure
subplot(2,1,1)
plot(t,ref(2:nit),t,saida(2:nit))
% ylim([0 100])

subplot(2,1,2)
plot(t,entrada(2:nit))


%% Configuração do NEPSAC
T = tril(ones(Nu));
T1 = inv(T);

%% simulação do NEPSAC
estados = repmat([Ca0;Cb0],[1 nit]);
saida= estados(2,:);

saidaModelo = repmat([Ca0;Cb0],[1,nit]);
entrada = FV0*ones(1,nit);
du = zeros(1,nit);

eta = zeros(2,nit+N2);
e = zeros(2,nit+N2);
rfant = Cb0;


for k=nin:nit
    %%% simulacao do CSTR
    estados(:,k) = modeloCSTR(estados(:,k-1),entrada(:,k-1-dn),q(:,k-1));
    saida(k) = estados(2,k);
    
    %% controlador NEPSAC
    %%% cálculo da resposta do modelo nominal
    saidaModelo(:,k) = modeloCSTRnominal(estados(:,k-1),entrada(k-1-dn),Caf0);
    
    %%% calculo dos etas futuros
    eta(:,k) = estados(:,k) - saidaModelo(:,k); % cálculo do eta atual
    
    e(1,k) = -C(2:end)*e(1,k-1:-1:k-nc+1)'+ D*eta(1,k:-1:k-nd+1)';
    e(2,k) = -C(2:end)*e(2,k-1:-1:k-nc+1)'+ D*eta(2,k:-1:k-nd+1)';
    
    for j=1:N2
        eta(1,k+j) = -D(2:end)*eta(1,k+j-1:-1:k+j-nd+1)' + C*e(1,k+j:-1:k+j-nc+1)';
        eta(2,k+j) = -D(2:end)*eta(2,k+j-1:-1:k+j-nd+1)' + C*e(2,k+j:-1:k+j-nc+1)';
    end
    
    
    ub = [entrada(k-dn:k-1), entrada(k-1)*ones(1,N2)]; % u base inicial
                          %ub = [u(k-d-1) u(k-d) ... u(k) u(k+1)... u(k+N2-1)
    dunorm = inf; % inicilizar norma de delta u como infinito
   
    % calculo das referencias
    rf = alpha*rfant + (1-alpha)*ref(k);
    rfant = rf;
    rfut = rf*ones(N2-N1+1,1);
    
    cont = 0;
    while (dunorm > 1e-4 && cont <= 100) % iterar até obter a resposta dentro do erro especificado
        %%% Obtenção da resposta base
        yb = zeros(2,N2); % resposta base
                        % yb=[y(k+1);y(k+2);...;y(k+N2)];

        yb(:,1) = modeloCSTRnominal(estados(:,k),ub(1),Caf0) + eta(:,k+1); % yb(k+1|k)
        for i=2:N2
            yb(:,i) = modeloCSTRnominal(yb(:,i-1),ub(i),Caf0) + eta(:,k+i); %yb(k+i|k)
        end


        %%% cálculo do controle ótimo
        udelta = ub;
        
        %%% cálculo de Ge
        Ge = zeros(N2,Nu);
        for j=1:Nu
            yot = zeros(2,N2);
            udelta = ub;
            if(j<Nu)
                udelta(j+dn) = udelta(j+dn)+entrada(k-1)/1000;
            else
                udelta(j+dn:end) = udelta(j+dn:end)+entrada(k-1)/1000;
            end

            yot(:,1) = modeloCSTRnominal(estados(:,k),udelta(1),Caf0) + eta(:,k+1); % yb(k+1|k)
            for i=2:N2
                yot(:,i) = modeloCSTRnominal(yot(:,i-1),udelta(i),Caf0) + eta(:,k+1);
            end
                        
            Ge(:,j) = (yot(2,:)' - yb(2,:)')/(entrada(k-1)/1000);
        end
        Ge = Ge(N1:N2,:);
        
        
        
        Hoti = 2*(Ge'*Qe*Ge+T1'*Qu*T1);
        foti = 2*(Ge'*Qe*(yb(2,N1:end)'-rfut)+T1'*Qu*T1*(-entrada(k-1)*ones(Nu,1)+ub(dn+1:dn+Nu)'));
        
        X = quadprog(Hoti,foti,[],[]);
        
        ub(dn+1:dn+Nu) = ub(dn+1:dn+Nu)+X';
        ub(dn+Nu:end) = ub(dn+Nu);
        
        dunorm = norm(X);
        cont = cont+1;
    end
    temp(1,k) = cont;
    temp(2,k) = dunorm;
    
    entrada(k) = ub(dn+1);
 

end

%% plotagem das figuras
t = (2:nit)-2;
figure
subplot(2,1,1)
plot(t,ref(2:nit),t,saida(2:nit))
% ylim([0 100])

subplot(2,1,2)
plot(t,entrada(2:nit))

%%
%% Geração das figuras
cores = gray(5);
cores = cores(1:end-1,:);

t = ((nin+1:nit)-nin)*Ts;

hf = figure
h=subplot(2,1,1)
plot(t,saidaGPC(1,nin+1:nit),'LineWidth',tamlinha,'Color',cores(1,:))
hold on
plot(t,saidaGPCnlin(1,nin+1:nit),'-.','LineWidth',tamlinha,'Color',cores(2,:))
plot(t,saida(1,nin+1:nit),'--','LineWidth',tamlinha,'Color',cores(3,:))
plot(t,ref(1,nin+1:nit),':','LineWidth',tamlinha,'Color',cores(4,:))
xlim([0 t(end)])
ylim([0.57 0.95])
hl = legend('GPC linear','GPC não linear','NEPSAC','Referência','Location','SouthEast')
ylabel(['Controlada',newline,'(mol/L)'],'FontSize', tamletra)
set(h, 'FontSize', tamletra);
grid on

h = subplot(2,1,2)
plot(t,entradaGPC(1,nin+1:nit),'LineWidth',tamlinha,'Color',cores(1,:))
hold on
plot(t,entradaGPCnlin(1,nin+1:nit),'-.','LineWidth',tamlinha,'Color',cores(2,:))
plot(t,entrada(1,nin+1:nit),'--','LineWidth',tamlinha,'Color',cores(3,:))

ylabel(['Manipulada',newline,'(L/min)'],'FontSize', tamletra)
xlabel('Tempo (min)','FontSize', tamletra)
xlim([0 t(end)])
grid on
set(h, 'FontSize', tamletra);
ylim([0 0.35])

% h=subplot(3,1,3)
% plot(t,erro(1:nit-N2-1),'LineWidth',tamlinha,'Color',cores(1,:))
% title('Erro', 'FontSize', tamletra);
% xlabel('tempo', 'FontSize', tamletra);
% set(h, 'FontSize', tamletra);



hf.Position = tamfigura;
hl.Position = [0.6488 0.4766 0.2536 0.2371];
print('nepsac_cstr_comparativo_nlin','-depsc')


%%

% x = modeloCSTRnominal([Cb0;Cb0],[FV0;FV0])

%%
%%% Modelo CSTR isotérmico SISO livro Bequette - Process Dynamics
function [x] = modeloCSTR(y,u,q)
    global k1 k2 k3 Ts opts
    Ca = y(1);
    Cb = y(2);

    f = u(1);

    caf = q(1);
    %%% definição do modelo com as entradas e perturbações constantes. Isto
    %%% simula o sustentador de ordem zero.
    %%% y = [Ca;T];
    fcstr = @(t,y) [-k1*y(1)-k3*y(1)^2+(caf-y(1))*u(1); % derivada de Ca
                    k1*y(1)-k2*y(2)-y(2)*u(1)]; % derivada de Cb

    [t,x] = ode45(fcstr,[0 Ts],[Ca;Cb],opts);
    x = x(end,:)';
                    
end



%% Modelo CSTR nominal (Modelo Volterra) utilizado pelos controladores
function [ret] = modeloCSTRnominal(x,u,caf)
    global k1 k2 k3 Ts
    ca = x(1);
    cb = x(2);
    ret = zeros(2,1);
    
    ret(1) = ca +Ts*( -k1*ca - k3*ca^2+(caf-ca)*u);
    ret(2) = cb + Ts*(k1*ca-k2*cb-cb*u);

    if(ret(1)<0)
        ret(1)=0;
    end
    if(ret(2)<0)
        ret(2)=0;
    end
    
    
end