%%% Exemplo duplo integrador do capítulo GPC SISO
%%% cálculo do ganho irrestrito
%%% cálculo das funções de transferência equivalentes.

clear all
% close all
clc



%% parametros de sintonia do controlador GPC
N1 = 1;
N2 = 30;
Ny = N2-N1+1;
Nu = 5;
Nq = 0;

lambda = 1; % ponderação do incremento de controle na função custo.
delta = 1; % ponderação do erro na função custo


C = 1; % polinômio C do modelo da perturbação
D = [1 -1]; % polinômio D do modelo da perturbação

nc = size(C,2);
nd = size(D,2);

umax = 1; % valor máximo da abertura da válvula
umin = 0; % valor mínimo da abertura da válvula

T = tril(ones(Nu));
T1 = inv(T);

Qe = delta*eye(N2-N1+1);
Qu = lambda*eye(Nu);

%% parametros da simulação
nit = 100; % numero de iterações

h0 = 50;
u0 = 62.6418;
q0 = 20;

%%% parâmetros do modelo
global K1;
global K2;
global g;
K1 = 2;
K2 = 0.2;
g = 9.81;


ref = h0*ones(1,nit);
ref(10:end) = h0*1.2;
ref(40:end) = h0*0.4;
ref(80:end) = 80;

saida= h0*ones(1,nit);
saidaModelo = h0*ones(1,nit);
entrada = u0*ones(1,nit);
du = zeros(1,nit);
pert = q0*ones(1,nit);

pert(50:end) = 17;


%% Modelo da planta
A = [1 (-1+K2*q0*g/sqrt(2*g*h0))];
B = [K1]; 
Bq = [0 -K2*sqrt(2*g*h0)];

d=0;
dq = d;

nb = size(B,2)-1; % ordem do polinômio B
na = size(A,2)-1; % ordem do polinômio A
nbq = size(Bq,2)-1; % ordem do polinômio Bq


%% Obtenção das matrizes do GPC
Btil = conv(B,[zeros(1,d) 1]); % incorporação do atraso no numerador B
Bqtil = conv(Bq,[zeros(1,dq) 1]); % incorporação do atraso no numerador Bq


G = zeros(Ny,Ny); % matriz dinâmica G
H = zeros(Ny,nb+d);
Gq = zeros(Ny,Ny-1); 
Hq = zeros(Ny,nbq+dq);

[E,F] = diofantina(conv(A,[1 -1]),N1,N2);

for i=N1:N2
    EjB = conv(E(i-N1+1,1:i),Btil);
    EjBq = conv(E(i-N1+1,1:i),Bq);
    
    G(i-N1+1,1:i) = EjB(i:-1:1);    
    H(i-N1+1,:) = EjB(i+1:end);

    Gq(i-N1+1,1:i) = EjBq(i:-1:1);    
    Hq(i-N1+1,:) = EjBq(i+1:end);
    
end
G = G(:,1:Nu);
Gq = Gq(:,1:Nq);


G,F,H,Gq,Hq

Hoti = 2*(G'*Qe*G+Qu);
foti = -2*G'*Qe;


%% simulação 
for k=2:nit
    %%% simulacao do tanque
    if(k<60)
        saida(k) = modeloTanque(saida(k-1),entrada(k-1),pert(k-1));
    else %% inserção de perturbação de entrada
        saida(k) = modeloTanque(saida(k-1),entrada(k-1)+5,pert(k-1));
    end

    %% controlador GPC
    R = ref(k)*ones(N2-N1+1,1);
    %%% cálculo da resposta livre;
    f = F*saida(1,k:-1:k-na)';
    
    if(~isempty(H))
        f = f+ H*du(1,k-1:-1:k-nb-d)'; % parcela dos incrementos de controle
    end
    
    %%% adição da parcela do incremento da perturbação passada na
    %%% resposta livre
%     if(~isempty(Hq))
%         deltaQ = pert(1,k:-1:k-nbq-dq+1)'-pert(1,k-1:-1:k-nbq-dq)';
%         f = f+ Hq*deltaQ;
%     end

    
    %%% cálculo do incremento de controle ótimo
    duOti = quadprog(Hoti,foti*(R-f),[],[]);
    du(1,k) = duOti(1);
    
    %%% cálculo do sinal de controle ótimo
    entrada(1,k) = du(1,k)+entrada(1,k-1);
 

end

%% plotagem das figuras
t = (2:nit)-2;
figure
subplot(2,1,1)
plot(t,ref(2:nit),t,saida(2:nit))
ylim([0 100])

subplot(2,1,2)
plot(t,entrada(2:nit))


%%
% %% Geração das figuras
% tamletra=14;
% tamnum=14;
% 
% figure
% h=subplot(3,1,1)
% plot(saidas1(1,nin+1:nit),'LineWidth',2)
% hold on
% plot(saidas2(1,nin+1:nit),'LineWidth',2)
% plot(refs(1,nin+1:nit),'--','LineWidth',2)
% % ylim([0 2])
% hl = legend('GPC1','GPC2','Referência','Location','SouthEast')
% hl.Position = [0.6870    0.6663    0.2368    0.1662];
% ylabel('Saída','FontSize', tamletra)
% set(h, 'FontSize', tamnum);
% grid on
% 
% h = subplot(3,1,2)
% plot(entradas1(1,nin+1:nit),'LineWidth',2)
% hold on
% plot(entradas2(1,nin+1:nit),'LineWidth',2)
% 
% ylabel('Entrada','FontSize', tamletra)
% grid on
% set(h, 'FontSize', tamnum);
% % ylim([-5 5])
% 
% h = subplot(3,1,3)
% plot(du1(1,nin+1:nit),'LineWidth',2)
% hold on
% plot(du2(1,nin+1:nit),'LineWidth',2)
% 
% ylabel('\Delta u','FontSize', tamletra)
% xlabel('Tempo (amostras)','FontSize', tamletra)
% grid on
% % ylim([-5 5])
% set(h, 'FontSize', tamnum);
% 
% % h=subplot(3,1,3)
% % plot(t,erro(1:nit-N2-1),'LineWidth',2)
% % title('Erro', 'FontSize', tamletra);
% % xlabel('tempo', 'FontSize', tamletra);
% % set(h, 'FontSize', tamnum);
% 
% 
% 
% % print('exemploMotorDC_ff','-depsc')




%% função do modelo do sistema de tanque
function hk = modeloTanque(hk1,ak1,qk1)
    global K1;
    global K2;
    global g;

    %%% modelo discreto do tanque
    hk = hk1 + K1*ak1 -K2*qk1*sqrt(2*g*hk1);
    
    if(hk<0)
        hk = 0;
    end
end

