clear all, close all, clc



%% parametros de sintonia do controlador NEPSAC
N1 = 1;
N2 = 30;
Nu = 5;

lambda = 1; % ponderação do incremento de controle na função custo.
delta = 1; % ponderação do erro na função custo


C = 1; % polinômio C do modelo da perturbação
D = [1 -1]; % polinômio D do modelo da perturbação

nc = size(C,2);
nd = size(D,2);

umax = 1; % valor máximo da abertura da válvula
umin = 0; % valor mínimo da abertura da válvula

T = tril(ones(Nu));
T1 = inv(T);

Qy = delta*eye(N2-N1+1);
Qu = lambda*eye(Nu);

%% parametros da simulação
nit = 100; % numero de iterações

h0 = 0;
u0 = 0;
q0 = 0;

ref = zeros(1,nit);
ref(10:end) = 1;
ref(40:end) = 2;

saida= ones(1,nit);
saidaModelo = zeros(1,nit);
entrada = zeros(1,nit);
pert = zeros(1,nit);

pert(50:end) = 0.5;

etas = zeros(1,nit+N2);
e = zeros(1,nit+N2);


saida(1) = h0;
entrada(1) = u0;
pert(1) = q0;


for k=2:nit
    %%% simulacao do tanque
    if(k<60)
        saida(k) = modeloTanque(saida(k-1),entrada(k-1),pert(k-1));
    else
        saida(k) = modeloTanque(saida(k-1),entrada(k-1)+0.1,pert(k-1));
    end

    %% controlador NEPSAC
    %%% cálculo da resposta do modelo nominal
    saidaModelo(k) = modeloTanque(saidaModelo(k-1),entrada(k-1),pert(k-1));
    
    %%% calculo dos etas futuros
    eta(k) = saida(k) - saidaModelo(k); % cálculo do eta atual
    
    e(k) = -C(2:end)*e(k-1:-1:k-nc+1)'+ D*eta(k:-1:k-nd+1)';
    
    for j=1:N2
        eta(k+j) = -D(2:end)*eta(k+j-1:-1:k+j-nd+1)' + C*e(k+j:-1:k+j-nc+1)';
    end
    
    
    ub = entrada(k-1)*ones(N2,1); % u base inicial
    dunorm = inf; % inicilizar norma de delta u como infinito
    
    rfut = ref(k)*ones(N2-N1+1,1);
    
    cont = 0;
    while dunorm > 1e-3 && cont <= 20 % iterar até obter a resposta dentro do erro especificado
        %%% Obtenção da resposta base
        yb = zeros(N2,1); % resposta base

        yb(1) = modeloTanque(saida(k),ub(1),pert(k-1)); % yb(k+1|k)
        for i=2:N2
            yb(i) = modeloTanque(yb(i-1),ub(i),pert(k-1));
        end

        yb2 = yb;
        yb = yb + eta(k+1:k+N2)';


        %%% cálculo do controle ótimo
        udelta = ub;
        
        %%% cálculo de Ge
        Ge = zeros(N2,Nu);
        for j=1:Nu
            yot = zeros(N2,1);
            udelta = ub;
            if(j<Nu)
                udelta(j) = udelta(j)+1;
            else
                udelta(j:end) = udelta(j:end)+1;
            end
            
            yot(1) = modeloTanque(saida(k),udelta(1),pert(k-1));
            for i=2:N2
                yot(i) = modeloTanque(yot(i-1),udelta(i),pert(k-1));
            end
            
            yot = yot + eta(k+1:k+N2)';
            
            Ge(:,j) = yot - yb;
        end
        Ge = Ge(N1:N2,:);
        
        
        
        Hoti = 2*(Ge'*Qy*Ge+T1'*Qu*T1);
        foti = 2*(Ge'*Qy*(yb-rfut)+T1'*Qu*T1*(-entrada(k-1)*ones(Nu,1)+ub(1:Nu)));
        
        X = quadprog(Hoti,foti,[],[]);
        
        ub(1:Nu) = ub(1:Nu)+X;
        ub(Nu+1:end) = ub(Nu);
        
        dunorm = norm(X);
        cont = cont+1;
    end
    temp(k) = cont;
    
    entrada(k) = ub(1);
 

end

%% plotagem das figuras
t = (2:nit)-2;
figure
subplot(2,1,1)
plot(t,ref(2:nit),t,saida(2:nit))

subplot(2,1,2)
plot(t,entrada(2:nit))




%% função do modelo do sistema de tanque
function hk = modeloTanque(hk1,ak1,qk1)
    
    %%% parâmetros do modelo
    K1 = 2;
    K2 = 0.2;
    g = 9.81;

    %%% modelo discreto do tanque
    hk = hk1 + K1*ak1 -K2*qk1*sqrt(2*g*hk1);
    
    if(hk<0)
        hk = 0
    end
end


