clear all
close all
clc

run('../Bibliotecas/parametrosFiguras.m')

%%
import casadi.*

% Define parâmetros do problema

K1=2;             
K2=.1;  
g=9.8;                    % Aceleração da gravidade  [m/s^2]
tsim=100;                 % Tempo de simulação  [s]
ts=1;                     % Período de amostragem [s]
nit=round(tsim/ts);       % número de iterações

ns=1;                 % Numero de estados
m=1;                  % Numero de entradas
n=1;                  % Numero de saídas
N=20;                 % Horizonte de predição
Nu=5;                 % Horizonte de controle

Qx = eye(n)            % Ponderação estado
Qu = 0.01             % Ponderação controle

xs = 50*ones(n,nit);      % Estados
us = 31.9438*ones(m,nit);  % Controle
qs = 50*ones(n,nit);      % Perturbação
xr= 50*ones(n,nit);       % Referência
J = zeros(1,nit);         % Objetivo

Umax = 100;        % Valor máximo da entrada
Umin = 0;           % Valor mínimo da entrada

Xmax = 100;         % Valor máximo do estado
Xmin = 0;           % Valor mínimo do estado


% valor da referência ao longo da simulação
xr(1,round(tsim*0.1/ts):end) = 45;

% Valor da perturbação ao longo da simulação
qs(1,round(nit/2):end) = 45;

% Laço de simulação em i
for k=2:nit
    %% simulação do modelo
    xs(k)= xs(k-1) + K1*qs(k-1) - K2*us(k-1)*sqrt(2*g*xs(k-1));   
    
    %%% Controlador NMPC com CasADi
    %% Variáveis de minimização NLP
    opti = casadi.Opti();
    X = opti.variable(n,N+1);
    U = opti.variable(n,Nu);
    obj = 0;

    % Inicialização X e U com os valores atuais (warm start)
    opti.set_initial(U,us(k-1));
    opti.set_initial(X,xs(k-1));
    % Montando as restrições e o objetivo
    for j=1:N
        obj = obj + (xr(:,k)-X(:,j+1))'*Qx*(xr(:,k)-X(:,j+1));
        % Restrição dinâmica do modelo para j<=Nu
        if j<=Nu
            % Caso j+1<Nu soma parcela de controle e adiciona restrição
            obj = obj +  U(:,j)'*Qu*U(:,j);            
            opti.subject_to(X(:,j+1) == X(:,j) + K1*qs(k) - K2*U(:,j)*sqrt(2*g*max(X(:,j),1e-3)));
            % Restrições de operação
            opti.subject_to(Xmin<=X(:,j+1)<=Xmax);
            opti.subject_to(Umin<=U(:,j)<=Umax);
        else
            % Se j+1>=Nu, sinal de controle é fixo 
            opti.subject_to(X(:,j+1) == X(:,j) + K1*qs(k) - K2*U(:,Nu)*sqrt(2*g*max(X(:,j),1e-3)));
            % Restrições de operação
            opti.subject_to(Xmin<=X(:,j+1)<=Xmax);
        end
    end
        

    % Declarando o objetivo do problema: minimizar obj
    opti.minimize(obj);
    % Utilizando o solver ipopt
    opti.solver('ipopt');
    % Solucionando o problema não linear
    sol = opti.solve();
    % Vetor com os valores do objetivo
    J(k) = sol.value(obj);
    % Vetor com os valores da entrada
    us(:,k) = sol.value(U(:,1));
end

%% 
cores = gray(4);
cores = cores(1:end-1,:);

hf = figure
h=subplot(2,1,1)
plot(xs(1,2:nit)','LineWidth',tamlinha,'Color',cores(1,:))
hold on
plot(xr(:,2:nit)',':','LineWidth',tamlinha,'Color',cores(2,:))
% ylim([0 2])
ylabel('Controladas (%)','FontSize', tamletra)
hl0 = legend('Controlada','Referência','Location','NorthEast')
ylim([43 52])
set(h, 'FontSize', tamletra);
grid on

h = subplot(2,1,2)
plot(us(1,2:nit)','LineWidth',tamlinha,'Color',cores(1,:))

ylabel('Manipuladas (%)','FontSize', tamletra)

grid on
set(h, 'FontSize', tamletra);
ylim([28 35])


xlabel('Tempo (s)','FontSize', tamletra)
grid on
% ylim([-5 5])
set(h, 'FontSize', tamletra);

% h=subplot(3,1,3)
% plot(t,erro(1:nit-N2-1),'LineWidth',tamlinha,'Color',cores(1,:))
% title('Erro', 'FontSize', tamletra);
% xlabel('tempo', 'FontSize', tamletra);
% set(h, 'FontSize', tamletra);



hf.Position = tamfigura;
print('exemplo_casadi_tanque','-depsc')
        